(defpackage bst
  (:documentation "Binary search tree package

Types
-----
<bst>

Operations
----------
insert
remove-key
successor
predecessor
find-key
minimum
maximum 
sorted?
sorted-iterator
test-bst
pprint-bst
")
  (:export
   <bst>

   insert
   remove-key
   successor
   predecessor
   minimum
   maximum
   sorted?
   sorted-iterator
   find-key
   test-bst
   pprint-bst)
  (:use
   cl
   tree
   utils))


(in-package bst)

(defclass <bst> ()
  ((root :accessor root)
   (comparator :initarg :comparator :reader comparator :initform #'default-comparator)
   (key-fn :initarg :key-fn :reader key-fn :initform #'identity))
  (:documentation "Class <bst>

Represents a binary search tree.

Initargs
:key-fn - takes in an element, returns a key.
:comparator - used to compare keys.  Function of two arguments that returns either '<, '=, or '>.

The insert and remove operations for the root <bst> class do not do any rebalancing and so can lead to unbalanced trees in the worst case."))

(defmethod initialize-instance :after ((b <bst>) &rest args)
  (declare (ignore args))
  (setf (root b) (make-instance '<node> :node-label nil)))

(defun find-key (b k &key (comparator (comparator b)))
  "find-key BST KEY &key (COMPARATOR (COMPARATOR B))

If an element with this key is in the tree, return 1) The element 2) t
If not, return nil and nil."
  ;; Note third return value for internal use

  (let ((node (root b)))
    (loop
       (let ((v (node-label node)))
	 (if v
	     (dsbind
	      (key . val) v
	      (ecase (funcall comparator k key)
		(= (return (values val t node)))
		(< (setf node (left-child node)))
		(> (setf node (right-child node)))))
	     (return (values nil nil node)))))))

(defun insert (b e &key (comparator (comparator b)))
  "insert BST ELEMENT &key (COMPARATOR (COMPARATOR B))

If an element with the same key as ELEMENT is in the tree, do nothing and return nil and the existing element.  Otherwise, add to the tree and return a non-nil value."

  (let ((k (funcall (key-fn b) e)))
    (mvbind 
	(val found n)(find-key b k :comparator comparator)
      (if found
	  (values nil val)
	  (progn    
	    (repeat 2 (add-new-child n nil nil))
	    (setf (node-label n) (cons k e))
	    n)))))

(defun remove-key (b k &key (comparator (comparator b)))
  "remove-key BST KEY &key (COMPARATOR (comparator BST))

If an element with this key exists in the tree, return T and the element.  Otherwise, return nil."
  (mvbind (val found n) (find-key b k :comparator comparator)
	  (when found
	    (let ((lc (left-child n))
		  (rc (right-child n)))
	      (prog1 (values t val)
		(let ((swap
		       (if (and (leaf? lc) (leaf? rc))
			   n
			   (if (and (not (leaf? lc))
				    (or (leaf? rc) (> (random 1.0) .5)))
			       (max-node lc)
			       (min-node rc)))))
		  (setf (node-label n) (node-label swap))
		  (let ((lc (left-child swap))
			(rc (right-child swap))
			(par (unless (root? swap) (parent-node swap))))
		    (assert (or (leaf? lc) (leaf? rc)))
		    (remove-subtree lc)
		    (remove-subtree rc)
		    (cond ((not (leaf? lc))
			   (remove-subtree swap)
			   (when par
			     (add-subtree par nil lc 0)))
			  ((not (leaf? rc))
			   (remove-subtree swap)
			   (when par
			     (add-subtree par nil rc 1)))
			  (t (setf (node-label swap) nil))))))))))

(defun predecessor (node)
  "Return predecessor node of NODE in bst, or nil if it's the minimum."
  (or (max-node (left-child node))
      (loop
	 (if (root? node)
	     (return nil)
	     (let ((par (parent-node node)))
	       (if (eq node (right-child par))
		   (return par)
		   (setf node par)))))))

(defun successor (node)
  "Return successor node of NODE in containing bst, or nil if it's the maximum."
  (or (min-node (right-child node))
      (loop
	 (if (root? node) 
	     (return nil)
	     (let ((par (parent-node node)))
	       (if (eq node (left-child par))
		   (return par)
		   (setf node par)))))))
	     
	 

(defun max-node (node)
  (unless (leaf? node)
    (let ((rc (right-child node)))
      (if (leaf? rc)
	  node
	  (max-node rc)))))

(defun min-node (node)
  (unless (leaf? node)
    (let ((lc (left-child node)))
      (if (leaf? lc)
	  node
	  (min-node lc)))))

(defun maximum (bst)
  "maximum BST.  Return the element with the highest key value and non-nil, or nil and nil if the tree is empty."
  (let ((n (max-node (root bst))))
    (values (awhen n (cdr (node-label it))) n)))

(defun minimum (bst)
  (let ((n (min-node (root bst))))
    (values (awhen n (cdr (node-label it))) nil)))

(defun sorted-iterator (bst)
  "sorted-iterator BST.  Return an iterator (function of zero arguments that returns 1) next-element 2) done? on each call) that goes over the tree in sorted order."
  (let ((n (min-node (root bst))))
    #'(lambda ()
	(if n
	    (prog1 (values (cdr (node-label n)) nil)
	      (let ((rc (right-child n)))
		(setf n 
		      (if (leaf? rc)
			  (let ((node n))
			    (loop
			       (if (root? node) 
				   (return nil)
				   (let ((par (parent-node node)))
				     (if (eq node (left-child par))
					 (return par)
					 (setf node par))))))
			  (let ((node rc))
			    (loop
			       (let ((lc (left-child node)))
				 (if (leaf? lc)
				     (return node)
				     (setf node lc)))))))))
	    (values nil t)))))
	


(defun left-child (n)
  (elt (children n) 0))

(defun right-child (n)
  (elt (children n) 1))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; debugging
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun pprint-bst (b)
  (pprint-tree (root b)))

(defun sorted? (b)
  (let ((iter (sorted-iterator b))
	(prev-init (gensym))
	(comp (comparator b))
	(f (key-fn b))
	prev)
    (setf prev prev-init)
    (loop
       (mvbind (next done?) (funcall iter)
	       (unless (or (eq prev prev-init)
			   (eq (funcall comp (funcall f prev) (funcall f next)) '<))
		 (return nil))
	       (when done? (return t))))))
    

(defun test-bst (n m &optional (verbose nil))
  (repeat m
	  (let ((b (make-instance '<bst> :key-fn #'car)))
	    (flet ((check-sorted ()
		     (assert (sorted? b) nil "Not sorted: ~/pprint-bst/" b)))

	      (check-sorted)
	      (repeat 
	       n
	       (if (> (random 1.0)  .5)
		   (let ((e (cons (random 100) (random 1.0))))
		     (when verbose
		       (check-sorted)
		       (format t "~&Inserting ~a" e)
		       )
		     (insert b e)
		     )
		   (let ((k (random 100)))
		     (when verbose
		       (check-sorted)
		       (format t "~&Removing ~a" k)
		       )
		     (remove-key b k)
		     )))
	      (when verbose
		(do-iterator (sorted-iterator b) (x)
		  (print x)))))))