(defpackage condition-variable
  (:documentation "Package for condition variables.  Condition variables are a synchronization tool typically used to make a thread wait for some state that depends on another thread's actions.

Condition variables are always associated with a process lock, and all condition variable operations must be done while a process is holding the lock that is associated with the condition variable, i.e. (eq (process-lock-locker lock) *current-process*) (and this is asserted in each method.

To use a condition variable, use these methods:

make-condition-variable process-lock)
    Creates a condition variable associated with the given process lock.  It is perfectly acceptable to associate multiple condition variables with the same lock.

num-waiting condition-variable)
    Returns the number of processes waiting on the condition variable.

get-waiting condition-variable)
Returns a list of the processes waiting on this condition variable.

reset-cv - remove all waiting processes and reset the table of process ids.

The code in this package currently only works with Allegro Lisp, Version 6.2 or later.")


  (:nicknames cv)
  (:use common-lisp 
	mp
	threads
	utils)
  (:export 
   <condition-variable>
   make-condition-variable
   reset-cv
   wait
   notify-all
   num-waiting
   get-waiting))
   
	
(in-package condition-variable)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; exported functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <condition-variable> ()
  ((lock
    :type     process-lock
    :initarg  :lock
    :reader   cv-lock)
   (waiting-processes :type list :initform nil :accessor waiting-processes)
   (num-waiting
    :type fixnum
    :initform 0
    :accessor cv-num-waiting)
   (name
    :initarg :name
    :reader name
    :initform nil)
   (whostate :initarg :whostate :reader whostate :type string)
   (gate-table :type hash-table :accessor gate-table)
   ))

(defun make-condition-variable (lock &optional (name "cv"))
  (make-instance '<condition-variable>
    :lock lock
    :name name
    :whostate (format nil "Waiting on condition variable ~a" name)
    ))

(defun reset-cv (cv)
  "reset-cv CV.  Used to reset a condition variable.  Not synchronized - should not be called when there are multiple threads running."
  (setf (waiting-processes cv) nil)
  (setf (cv-num-waiting cv) 0)
  (setf (gate-table cv) (make-hash-table :test #'eq)))


(defmethod wait ((cv <condition-variable>))
  "wait CV
    Wait causes the current process to block (wait) on the condition variable.  The process gives up the lock while holding, but is guaranteed to have it again when it returns.  The condition variable has memory requirements proportional to the number of distinct processes that ever call wait on it, because their ids are stored in a table.  Use reset-cv to reset this table."
  
  (let ((lock (cv-lock cv))
	(gate (get-my-gate cv)))
    (assert (eq (process-lock-locker lock) *current-process*) ()
      "Wait called without holding lock:~A" lock)
    (push *current-process* (waiting-processes cv))
    (incf (cv-num-waiting cv))
    (unwind-protect
	(mp::without-interrupts
	  (process-unlock lock)
	  (close-gate gate)
	  (process-wait (whostate cv) #'gate-open-p gate)
	  )
      (process-lock lock))))

(defmethod notify-all ((cv <condition-variable>))
  "notify-all CV
Causes all processes blocked on CV to wake up.  Does not release the lock, so they won't actually be able to run until the calling process releases the lock."
  (assert (eq (process-lock-locker (cv-lock cv)) *current-process*) nil
    "Notify-all called on ~a without holding lock" (name cv))
  (awhile (pop (waiting-processes cv))
	  (decf (cv-num-waiting cv))
	  (open-gate (get-gate cv it))))

(defmethod num-waiting ((cv <condition-variable>))
  (let ((lock (cv-lock cv)))
    (assert (eq (process-lock-locker lock) *current-process*) ()
      "Num-waiting called without holding lock:~A" lock)
    (cv-num-waiting cv)))

(defmethod get-waiting ((cv <condition-variable>))
  (let ((lock (cv-lock cv)))
    (assert (eq (process-lock-locker lock) *current-process*) ()
      "get-waiting called without holding lock:~A" lock)
    (waiting-processes cv)))


(defun get-gate (cv proc)
  (assert (am-holding-lock (cv-lock cv)) nil
    "get-gate called without holding lock for condition variable ~a" (name cv))
  (mvbind (gate present?) (gethash proc (gate-table cv))
	  (if present?
	      gate
	    (setf (gethash proc (gate-table cv)) (make-gate nil)))))

(declaim (inline get-my-gate))
(defun get-my-gate (cv)
  (get-gate cv *current-process*))




