(defpackage difference-updating
  (:use
   cl
   utils
   graph
   prob
   set
   diff-fn
   mdp)
  (:nicknames diff-up)
  (:export
   difference-updating
   bellman-update
   normalize)
  )

(in-package diff-up)


(defun difference-updating (m &key (max-cycle-length 2) (normalization-frequency .5)
				   (step-size .1) (max-num-steps 'infty) (normalization-step-size step-size)
			    &aux (g (state-transition-graph m)))
									      
  "difference-updating MDP &key (MAX-CYCLE-LENGTH 2) (NORMALIZATION-FREQUENCY .5) (STEP-SIZE .1)  (MAX-NUM-STEPS 'infty) (NORMALIZATION-STEP-SIZE STEP-SIZE)

Applies the difference updating algorithm to MDP, and returns the resulting object of type <difference-function>.  This is a randomized, asynchronous, incremental DP algorithm.  On each step, it first picks whether to do a normalization or Bellman update (the first case happening with probability NORMALIZATION-FREQUENCY).  For Bellman updates, it samples a uniform pair of states S,S' from the set of all adjacent pairs in the directed state transition graph, and carries out the Bellman update.  For normalization, it samples uniformly from the set of cycles S0,...,SN,S0 for N<MAX-CYCLE-LENGTH and carries out a normalization.  In either case, the magnitude of the update is proportional to STEP-SIZE.  When MAX-NUM-STEPS is provided, the whole procedure terminates after MAX-NUM-STEPS and the current difference function is returned.  Other termination criteria and update selection schemes may be added later."

  (do ((cycles (mapset 'list #'(lambda (x) (mapset 'vector #'identity x)) (cycles-of-length-upto g max-cycle-length)))
       (adj-pairs (adjacent-pairs g))
       (f (make-instance '<tabular-difference-function> :states (state-set m)))
       (d `((bellman . ,(- 1 normalization-frequency)) (normalization . ,normalization-frequency)))
       (current-steps 0 (incf current-steps)))
      ((my>= current-steps max-num-steps) f)
    
    ;; Body of loop has two cases
    (ecase (sample d)
      
      (bellman
       (let ((p (sample-uniformly adj-pairs)))
	 (dsbind (s0 . s1) p
	   (debug-prompt 0 "~&Bellman update for ~a, ~a" s0 s1)
	   (bellman-update f s0 s1 m step-size))))
      
      (normalization
       (let ((cycle (sample-uniformly cycles)))
	 (debug-prompt 0 "~&Normalization for ~a" cycle)
	 (normalize f cycle normalization-step-size))))
    
    (when-debugging (pprint f))))

	   
	     
    
  
  
  
  

(defun bellman-update (d s0 s1 m step-size)
  "bellman-update DIFF-FN STATE STATE' MDP STEP-SIZE

Do the backup D(s0,s1) -> D(s0,s1) + max_a E [ D(s1,s2) + R(s1,a,s2) | s1,a]"
  
  (update 
   d s0 s1
   (if (eq s0 s1)
       0
     (- 
      (difference d s0 s1 0)
      (reduce-set #'mymax
		  (avail-actions m s0)
		  :initial-value '-infty
		  :key #'(lambda (a)
			   (expectation
			    (trans-dist m s0 a)
			    #'(lambda (s2)
				(+ (reward m s0 a s2)
				   (difference d s0 s2 0))))))))
   step-size))




(defun normalize (d states step-size &aux (n (length states)))
  "normalize D STATES STEP-SIZE

Let STATES be #(S0 S1 ... Sn).   This function tries to make D(S0,S1) + ... + D(S_{n-1},S_n) + D(S_n, S0) closer to 0 by stepsize."
  (let ((inc (/ (+ (sum-over (1- n) #'(lambda (i) (difference d (aref states i) (aref states (1+ i)) 0)))
		   (difference d (aref states (1- n)) (aref states 0) 0))
		n)))
    (flet ((update-inds (i j)
	     (let ((s0 (aref states i))
		   (s1 (aref states j)))
	       (update d s0 s1 (- (difference d s0 s1 0) inc) step-size))))
      
      (dotimes (i (1- n))
	(update-inds i (1+ i)))
      (update-inds (1- n) 0))))







			      

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; internal
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun adjacent-pairs (g &aux (states (nodes g)))
  "Return list of pairs of form (S1 . S2) where S1 and S2 are connected by an edge in the transition graph."
  (let ((l nil))
    
    (do-elements (s1 states l)
      (do-elements (s2 states nil)
	(when (get-edge g s1 s2)
	  (push (cons s1 s2) l))))))

(defun cycles-of-length (g k &aux (states (nodes g)))
  (when (> k 0)
  (mapcar 
   #'(lambda (s) (mapset 'list #'identity s))
   (filter ':list
	   (disjoint-union-list 
	    (mapcar #'permutations (subsets-of-size states k)))
	   #'(lambda (p)
	       (and
		(each (1- k)
		      #'(lambda (i) (get-edge g (item i p) (item (1+ i) p))))
		(get-edge g (item (1- k) p) (item 0 p))))))))

(defun cycles-of-length-upto (g k)
  (disjoint-union-list
   (mapcar #'(lambda (i) (cycles-of-length g i)) (consec 1 k))))
	   