(defpackage depots
  (:documentation "
Numeric version of depots domain from IPC 2002

Constructor
-----------
make-depots-domain

Actions
-------
drive
lift
drop
load
unload

")
  (:use
   cl
   utils
   lookahead
   mapping
   set
   ltk-obs
   ltk
   prob
   prod-set
   )
  (:import-from prop-logic
   prop-args
   prop-symbol
   conjuncts
   disjuncts
   formula
   negation )
  (:export
   make-depots-domain
   drive
   lift
   drop
   load
   unload

   make-depots-hierarchy
   depots-act-complete
   depots-descs-complete
   depots-descs-sound
   ))

(in-package depots)


#|

(make-instance '<prop-abstract-planning-problem>
	  :preprocess t :admissible-heuristic #'ns-act-complete :planning-problem d2 :hierarchy g
	  :guarantee-heuristic #'ns-act-sound :complete-desc-schemas (ns-descs d2)
	  :sound-desc-schemas (ns-descs d2) :subsumption-checker c)

|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; helper macro that needs to come before main function
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro verify-and-set (place init-symbol)
  "PLACE - unevaluated symbol naming a variable
INIT-SYMBOL - evaluated symbol

Expands to code that checks that the first element of place is init-symbol, then sets place to its cdr."
  (let ((x (gensym)))
    `(let ((,x ,place))
       (assert (eq (car ,x) ,init-symbol) nil "~a doesn't start with ~a" ,x ,init-symbol)
       (setf ,place (cdr ,x)))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; STRIPS descriptions of actions in domain
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *action-descs*
  '((drive
     ((?truck . trucks) (?from . locations) (?to . locations))
     (and (at ?truck ?from))
     ((at ?truck ?to))
     ((at ?truck ?from)))
    
    (lift
     ((?hoist . hoists) (?crate . crates) (?surface . surfaces) (?place . locations))
     (and 
      (at ?hoist ?place) (available ?hoist) 
      (at ?crate ?place) (on ?crate ?surface) (clear ?crate))
     ((lifting ?hoist ?crate) (clear ?surface))
     ((at ?crate ?place) (clear ?crate) (available ?hoist) (on ?crate ?surface)))

    (drop
     ((?hoist . hoists) (?crate . crates) (?surface . surfaces) (?place . locations))
     (and 
      (at ?hoist ?place) (lifting ?hoist ?crate)
      (at ?surface ?place) (clear ?surface))
     ((at ?crate ?place) (clear ?crate) (available ?hoist) (on ?crate ?surface))
     ((lifting ?hoist ?crate) (clear ?surface)))

    (load
     ((?hoist . hoists) (?crate . crates) (?truck . trucks) (?place . locations))
     (and 
      (at ?hoist ?place) (at ?truck ?place) (lifting ?hoist ?crate))  ;; TODO: LOAD-limit ??
     ((in ?crate ?truck) (available ?hoist))  ; TODO:  increase load
     ((lifting ?hoist ?crate)))

    (unload
     ((?hoist . hoists) (?crate . crates) (?truck . trucks) (?place . locations))
     (and 
      (at ?hoist ?place) (at ?truck ?place) 
      (available ?hoist) (in ?crate ?truck))
     ((lifting ?hoist ?crate))               ; TODO: decrease load
     ((in ?crate ?truck) (available ?hoist)))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; domain type and state sets
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <depots-domain> (<propositional-domain>)
  ())

#|
(defstruct (depot-comp-set (:conc-name dcs-))
  loc-crates
  loc-misplaced
  truck-locs)
|#

#|(defmethod make-state-set ((d <depots-domain>) s)
  (let ((locations (lookup-type d 'locations))
	(trucks (lookup-type d 'trucks))
	(truck-locs (make-hash-table))
	(crate-locs (make-hash-table))
	(crates (lookup-type d 'crates))
	(chains (chains d))
	(misplaced (make-hash-table)))
    
    (do-elements (p locations)
      (setf (gethash p misplaced) nil))
	
    (do-elements (l locations)
      (do-elements (tr trucks)
	(when (holds s (list 'at tr l))
	  (setf (gethash tr truck-locs) l))
	(do-elements (c crates)
	  (when (holds s (list 'in c tr))
	    (setf (gethash c crate-locs) l))))
      (do-elements (c crates)
	(when (holds s (list 'at c l))
	  (setf (gethash c crate-locs) l))))
    
    
    (dolist (ch chains)
      (let ((l (check-not-null (find (first ch) locations :test #'(lambda (p l) (holds s (list 'at p l)))))))
	(format t "~&chain ~a at ~a" ch l)
	(loop 
	    for c1 in ch
	    for ch2 on (rest ch)
	    unless (holds s (list 'on (first ch2) c1))
	    do (dolist (c ch2)
		 (when (eq (gethash c crate-locs) l)
		   (push c (gethash l misplaced))))
	       (return))))
    
    (make-depot-comp-set :truck-locs truck-locs :loc-crates crate-locs
			 :loc-misplaced misplaced)))|#



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; parse a description in the ipc format for depots
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun make-depots-domain (&key description input-file)
  "make-driver-log-domain &key DESCRIPTION INPUT-FILE

DESCRIPTION - a description of the domain as in the IPC problem files
INPUT-FILE - filename

Exactly one of DESCRIPTION and INPUT-FILE must be given."
  
  (assert (xor description input-file))
  (when input-file
    (setf description
      (with-open-file (f input-file :direction :input)
	(read f))))

  (dbind (prob dom objects init goal) (rest description)

    ;; Initial verification
    (verify-and-set prob 'problem)
    (verify-and-set dom ':domain)
    (verify-and-set objects ':objects)
    (verify-and-set init ':init)
    (verify-and-set goal ':goal)
    (assert (equal dom '(depot)) nil "~a didn't equal depot" (first dom))
      
      ;; Set of objects in each type
    (let ((object-types (make-hash-table))
	  (crates (make-extendable-array))
	  (pallets (make-extendable-array))
	  (locations (make-extendable-array))
	  (hoists (make-extendable-array))
	  (trucks (make-extendable-array)))
	
	(while objects
	  (let* ((dash-pos (position '- objects))
		 (type (nth (1+ dash-pos) objects)))
	    (multiple-value-bind (dest ourtype)
		(ecase type 
		  (Depot       (values locations 'locations))
		  (Distributor (values locations 'locations))
		  (Truck       (values trucks    'trucks))
		  (Pallet      (values pallets   'pallets))
		  (Crate       (values crates    'crates))
		  (Hoist       (values hoists    'hoists)))
	      (loop for i from 1 to dash-pos 
		 for obj in objects 
		 do (vector-push-extend obj dest) (setf (Gethash obj object-types) ourtype))
	      (setf objects (nthcdr (+ 2 dash-pos) objects)))))
	
	(make-instance 
	 '<depots-domain>
	 :types (p2alist 'crates crates 'pallets pallets
			 'surfaces (disjoint-union crates pallets)
			 'trucks trucks 'locations locations 'hoists hoists
			 'locatable (disjoint-union trucks hoists crates pallets))
	 :object-types object-types
	 :fluents '((at locatable locations)
		    (on crates surfaces)
		    (in crates trucks)
		    (lifting hoists crates)
		    (available hoists)
		    (clear surfaces))
	 :nonfluents '()
	 :goal (first goal)
	 :reward-fn (constantly -1)
	 :action-descs *action-descs*
	 :pprint-state-fn #'pprint-depots-state
	 :init-state init
	 ))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; helpers
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; goal chains
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#|

(defun dest-places (d)
  (let ((on-chains (chains d))
	(s (init-state d))
	(h (make-hash-table))
	(locs (lookup-type d 'locations)))
    (flet ((pallet-loc (p)
	     (check-not-null (find-element locs #'(lambda (l) (holds s (list 'at p l)))))))
      (dolist (c on-chains h)
	(let ((loc (pallet-loc (car c))))
	  (dolist (crate (cdr c))
	    (setf (gethash crate h) loc)))))))


(defun chains (d)
  "return the on-chains in the goal"
  (let ((chains nil))
    (dolist (pred (cdr (prop-logic:formula (goal d))) chains)
      (assert (eq (car pred) 'on))
      (dsbind (o1 o2) (cdr pred)
	(let ((c1 (find o1 chains :key #'first))
	      (c2 (find o2 chains :key #'slast))) 
	  (if c1
	      (if c2
		  (progn
		    (deletef chains c1)
		    (nconc c2 c1))
		(progn (deletef chains c1) (push (cons o2 c1) chains)))
	    (if c2
		(nconc c2 (list o1))
	      (push (list o2 o1) chains))))))))

|#
  


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Miscellaneous
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun pprint-depots-state (str s)
  (let ((props (pds-props s))
	(locs nil)
	(hoists-by-loc (make-hash-table))
	(pallets-by-loc (make-hash-table))
	(trucks-by-loc (make-hash-table))
	(crates-by-loc (make-hash-table)) ;ignored
	(in-truck (make-hash-table))
	(on-block (make-hash-table))
	(in-hoist (make-hash-table))
	(clear-blocks (make-hash-table))
	(available-hoists (make-hash-table)))
    (do-elements (p props)
      (case (first p)
	(at        (adjoinf locs (third p))
	           (push (second p) (gethash (third p) (ecase (pd-object-type (pds-domain s) (second p))
							 (trucks trucks-by-loc)
							 (hoists hoists-by-loc)
							 (pallets pallets-by-loc)
							 (crates crates-by-loc)))))
	(on        (setf (gethash (third p) on-block) (second p)))
	(in        (push (second p) (gethash (third p) in-truck)))
	(lifting   (setf (gethash (second p) in-hoist) (third p)))
	(available (setf (gethash (second p) available-hoists) t))
	(clear     (setf (gethash (second p) clear-blocks) t))))

    (pprint-logical-block (str nil)
      (dolist (loc locs)
	(format str "Location ~A~%---------------------~%" loc)
	(dolist (pallet (gethash loc pallets-by-loc))
	  (format str "Pallet ~10A: " pallet)
	  (do*  ((parent pallet obj) 
		 (obj (gethash parent on-block) (Gethash parent on-block)))
		((null obj) (assert (gethash parent clear-blocks)))
	     (format str "[~8A] " obj))
	  (format str "~%"))
	(dolist (hoist (gethash loc hoists-by-loc))
	  (let ((carrying (gethash hoist in-hoist))
		(empty    (gethash hoist available-hoists)))
	    (assert (xor empty carrying))
	    (format str "Hoist  ~10A: ~A~%" hoist (or carrying "")))); "empty"))))
	(dolist (truck (gethash loc trucks-by-loc))
	  (format str "Truck  ~10A: " truck)
	  (let ((carrying  (gethash truck in-truck)))
	    (if carrying 
		(dolist (obj carrying) (format str "[~6A] " obj))
		(format str "")); "empty"))
	    (format str "~%")))
	(format str "~%")))))
		  
	       
    
    
(defun make-extendable-array ()
  (make-array 0 :adjustable t :fill-pointer 0))

