(defpackage driverlog
  (:documentation "
Hard numeric version of driverlog domain from IPC 2002

Constructor
-----------
make-driver-log-domain

Actions
-------
load
unload
board
disembark
drive
walk

")
  (:use
   cl
   utils
   lookahead
   set
   )
  (:export
   make-driver-log-domain
   load
   unload
   board
   disembark
   drive
   walk))
   


(in-package driverlog)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; helper macro that needs to come before main function
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro verify-and-set (place init-symbol)
  "PLACE - unevaluated symbol naming a variable
INIT-SYMBOL - evaluated symbol

Expands to code that checks that the first element of place is init-symbol, then sets place to its cdr."
  (let ((x (gensym)))
    `(let ((,x ,place))
       (assert (eq (car ,x) ,init-symbol) nil "~a doesn't start with ~a" ,x ,init-symbol)
       (setf ,place (cdr ,x)))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; STRIPS descriptions of actions in domain
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *action-descs*
    '((load 
       ((?obj . packages) (?truck . trucks) (?loc . locations))
       (and (at ?truck ?loc) (at ?obj ?loc))
       ((in ?obj ?truck))
       ((at ?obj ?loc)))
      
      (unload 
       ((?obj . packages) (?truck . trucks) (?loc . locations))
       (and (at ?truck ?loc) (in ?obj ?truck))
       ((at ?obj ?loc))
       ((in ?obj ?truck)))
      
      (board
       ((?d . drivers) (?t . trucks) (?l . locations))
       (and (at ?t ?l) (at ?d ?l) (empty ?t))
       ((driving ?d ?t))
       ((at ?d ?l) (empty ?t)))
      
      (disembark
       ((?d . drivers) (?t . trucks) (?l . locations))
       (and (at ?t ?l) (driving ?d ?t))
       ((at ?d ?l) (empty ?t))
       ((driving ?d ?t)))
      
      (drive
       ((?t . trucks) (?from . locations) (?to . locations) (?d . drivers))
       (and (at ?t ?from) (driving ?d ?t) (link ?from ?to))
       ((at ?t ?to))
       ((at ?t ?from)))
      
      (walk
       ((?d . drivers) (?from . locations) (?to . locations))
       (and (at ?d ?from) (path ?from ?to))
       ((at ?d ?to))
       ((at ?d ?from)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; parse a description in the ipc format for driverlog
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun make-driver-log-domain (&key description input-file)
  "make-driver-log-domain &key DESCRIPTION INPUT-FILE

DESCRIPTION - a description of the domain as in the IPC problem files
INPUT-FILE - filename

Exactly one of DESCRIPTION and INPUT-FILE must be given."
  
  (assert (xor description input-file))
  (when input-file
    (setf description
      (with-open-file (f input-file :direction :input)
	(read f))))

  (dbind (prob dom objects init goal metric) (rest description)
    
    ;; Initial verification
    (verify-and-set prob 'problem)
    (verify-and-set dom ':domain)
    (verify-and-set objects ':objects)
    (verify-and-set init ':init)
    (verify-and-set goal ':goal)
    (assert (equal dom '(driverlog)) nil "~a didn't equal driverlog" (first dom))
    (mvbind (fuel-cost time-cost) (parse-metric metric)
      (print fuel-cost)
      (print time-cost)
      
      ;; Set of objects in each type
      (let ((packages (make-extendable-array))
	    (drivers (make-extendable-array))
	    (trucks (make-extendable-array))
	    (locations (make-extendable-array)))
	
	(while objects
	  (dbind (o dash type) objects
	    (assert (eq dash '-))
	    (vector-push-extend o (ecase type
				    (driver drivers)
				    (obj packages)
				    (truck trucks)
				    (location locations))))
	  (setf objects (cdddr objects)))
	
	(let ((drive-distances (map 'list #'list locations)))
	
	  (dolist (p init)
	    (when (eq (first p) '=)
	      (dbind (lhs rhs) (rest p)
		(ecase (first lhs)
		  (load (assert (= rhs 0) nil "Code assumes init loads are 0 in ~a" p))
		  (fuel-per-minute (assert (= rhs 10) nil "Code assumes init fpms are 10 in ~a" p))
		  (fuel-used (assert (= rhs 0) nil "Code assumes init fuel-used is 0 in ~a" p))
		  (time-to-walk)
		  (time-to-drive (push (cons (third lhs) rhs) (cdr (assoc (second lhs) drive-distances))))))))

	  
	  (make-instance '<propositional-domain>
	    :types (p2alist 'trucks trucks 'packages packages 'drivers drivers 'locations locations
			    'locatable (disjoint-union trucks drivers packages))
	    :fluents '((at locatable locations)
		       (in packages trucks)
		       (driving drivers trucks)
		       (empty trucks))
	    :nonfluents '((link locations locations)
			  (path locations locations))
	    :goal (first goal)
	  
	    :reward-fn (make-reward-fn fuel-cost time-cost drive-distances)
	    
	    :action-descs *action-descs*
	    
	    :pprint-state-fn #'pprint-dlog-state
	    
	    :init-state (remove '= init :key #'car)
	  
	    ))))))



(defun parse-metric (m)
  "Take in something of the form '(:metric minimize ... ) as in the IPC driverlog domain files, check that it is a linear combination of fuel and time costs, and return fuel and time costs."
  (verify-and-set m ':metric)
  (dbind (minim (p (t1 a (x)) (t2 b (y)))) m
    (assert (and (eq minim 'minimize) (eq p '+) (eq t1 '*) (eq t2 '*)
		 (member x '(total-time fuel-used)) (member y '(total-time fuel-used)) (not (eq x y)))
	(m))
    
    (if (eq x 'total-time)
	(values b a)
      (values a b))))



       

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Reward function
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun make-reward-fn (fc tc drive-distances)
  #'(lambda (s a)
      (let ((moved (when (eq (first a) 'drive-truck) (second a))))
	(- (+ tc (* fc (iwhen moved  
			 (* (mapping:evaluate (mapping:evaluate drive-distances (third a)) (fourth a))
			    (fuel-cost s moved)))))))))

(defun fuel-cost (s truck)
  (let ((load (sum-over (pds-props s) 
			#'(lambda (p)
			    (indicator (and (eq (first p) 'in)
					    (eq (third p) truck)))))))
    (+ 10 (* load (1+ load) .5))))


	



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Miscellaneous
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun pprint-dlog-state (str s)
  (let ((props (pds-props s))
	(occupied-locs nil)
	(trucks nil))
    (do-elements (p props)
      (case (first p)
	(at (adjoinf occupied-locs (third p)))
	(empty (adjoinf trucks (second p)))
	(driving (adjoinf trucks (third p)))))
    
    (let ((loc-info (mapcar #'list occupied-locs))
	  (truck-info (mapcar #'(lambda (tr) (list tr 'unknown)) trucks)))
      (do-elements (p props)
	(case (first p)
	  (at (push (second p) (cdr (assoc (third p) loc-info))))
	  (driving (setf (second (assoc (third p) truck-info)) (second p)))
	  (empty (setf (second (assoc (second p) truck-info)) 'empty))
	  (in (push (second p) (cddr (assoc (third p) truck-info))))))
      
      (pprint-logical-block (str nil)
	(format str "Locations~2I")
	(dolist (l loc-info)
	  (format str "~:@_~a: ~{~a ~}" (car l) (cdr l)))
	(format str "~0I~:@_Trucks~2I")
	(dolist (tr truck-info)
	  (format str "~@:_~a: ~a; ~{~a ~}" (first tr) (second tr) (cddr tr)))))))

    
    
(defun make-extendable-array ()
  (make-array 0 :adjustable t :fill-pointer 0))