(defpackage flag-collection-env
  (:documentation "Package flag-collection-env (flags)

Creation
--------
make-flag-collection-env

State access
------------
fce-positions
fce-flag-states
fce-world-map
flag-position
flag-status
flag-statuses
effector-position
num-effectors
num-flags

Other
-----
N, S, E, W
")
  (:export
   make-flag-collection-env
   
   fce-positions
   fce-flag-states
   fce-world-map
   flag-position
   flag-status
   num-effectors
   effector-position
   num-flags
   flag-statuses

   n
   s
   e
   w)
  (:nicknames flags)
  (:use
   cl
   set
   bnet
   utils
   gw
   inst-vars
   mdp-env
   prob)
  )

(in-package flags)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; structure that holds state
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(gendefstruct (flag-collection-env-state (:conc-name fce-))
	      (positions clone as-is)
	      (flag-states clone as-is)
	      (world-map copy leave-out))

(defun num-effectors (s)
  (length (fce-positions s)))

(defun pprint-fce-state (&rest args)
  (bind-pprint-args
   (str s) args
   (let ((positions (fce-positions s))
	 (flag-states (fce-flag-states s))
	 (wm (fce-world-map s)))
     (dbind (nr nc) (array-dimensions wm)
       (pprint-logical-block (str nil)
	 (dotimes (r nr)
	   (unless (zerop r)
	     (format str "~:@_"))
	   (dotimes (c nc)
	     (let ((pos (list r c)))
	       (aif (position pos positions :test #'equal)
		    (format str "~W" it)
		    (let ((v (loc-value wm pos)))
		      (write-char
		       (if (eq v ':flag)
			   (if (second (find pos flag-states :test #'equal :key #'first))
			       #\.
			     #\F)
			 (if (eq v ':goal)
			     #\G
			   #\.))
		       str)))))))))))

(set-pprint-dispatch 'flag-collection-env-state #'pprint-fce-state)

(defun flag-position (i s)
  (first (elt (fce-flag-states s) i)))

(defun flag-status (i s)
  (second (elt (fce-flag-states s) i)))

(defun flag-statuses (s)
  (map 'list #'second (fce-flag-states s)))

(defun num-flags (s)
  (length (fce-flag-states s)))

(defun effector-position (i s)
  (elt (fce-positions s) i))
	   
       


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; positions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun make-pos-desc (i d wm &aux (positions (apply #'direct-product 'list (array-dimensions wm))))
  (make-2tbn-var-desc
   :id `(pos ,i) :domain positions
   :prev-slice-parents `((pos ,i) (action ,i))
   :init-dist (constantly d)
   :trans-dist (make-cpd (prev-pos a)
			 (make-deterministic-dist
			  (result-legal wm prev-pos a)))))


	
	


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; flag status
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun make-flag-desc (j loc n)
  (make-2tbn-var-desc
   :id `(flag-status ,j)
   :domain (direct-product 'list (list loc) '(t nil))
   :prev-slice-parents (list `(flag-status ,j))
   :curr-slice-parents (mapset 'list #'(lambda (i) `(pos ,i)) n)
   :init-dist (constantly (make-deterministic-dist (list loc nil)))
   :trans-dist
   (make-cpd 
    (status &rest positions)
    (make-deterministic-dist
     (list loc
	   (or (second status)
	       (to-boolean (member loc positions :test #'equal))))))))
	     


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; creating the overall environment
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun make-flag-collection-env (a init-pos-dists &aux (n (length init-pos-dists)))
  "make-flag-collection-env WORLD-MAP INIT-POS-DISTS

WORLD-MAP is a 2d-array representing the world.  Each element can be one of ':empty, ':goal, or ':flag
INIT-POS-DISTS is a sequence of probability distributions over locations on the map (represented as two-element lists)

Returns an environment that works as follows.  There are N robots (where N = length (INIT-POS-DISTS)) moving around the map.  There are also a set of flags.  At each step, each robot can move north, south, east, west, or pickup the flag from its current location, if it hasn't already been picked up.  The episode ends when all robots are at goal positions, and all the flags have been picked up.  There is a cost of -1 per timestep."

  
  (let ((flag-locs
	 (make-instance '<filtered-set>
	   :base-set (apply #'direct-product 'list (array-dimensions a))
	   :predicate #'(lambda (loc) (eq (loc-value a loc) ':flag))
	   :store-cached-size t)))
		       
    (let ((2tbn
	   (make-instance '<2tbn>
	     
	     :state-descs 
	     (concatenate
		 'list
	       (mapset 'list #'(lambda (i) (make-pos-desc i (elt init-pos-dists i) a)) n) 
	       (mapset 'list #'(lambda (j) (make-flag-desc j (item j flag-locs) n)) (size flag-locs))
	       (list (make-param-desc 'world-map (clone a))))
	     
	     :action-descs
	     (mapset 'list #'(lambda (i) (make-2tbn-var-desc :id `(action ,i) :domain '(N S E W))) n)
	     
	     :reward-descs
	     (list (make-2tbn-var-desc :id 'cost-of-living :domain 'real-numbers :reward-fn (constantly -1)))
	     
	     :state-acc
	     (make-struct-accessors 
	      (flag-collection-env-state :conc-name "FCE-")
	      ((positions array n)
	       (flag-states array (size flag-locs))
	       world-map))
	     
	     :action-acc
	     (make-list-accessors n)
	     
	     )))
      
      (values (make-2tbn-mdp-env     
	       2tbn
	       :term-pred
	       #'(lambda (s)
		   (let ((positions (fce-positions s))
			 (flag-states (fce-flag-states s)))
		     (and
		      (every #'(lambda (pos) (eq ':goal (loc-value a pos))) positions)
		      (every #'second flag-states)))))))))
			   

