(defpackage forage
  (:documentation "The forage environment package

Operations
----------
create-forage-env

Symbols
-------
N,S,E,W,F

")
  (:export
   create-forage-env
   n
   s
   e
   w
   f)
      
  (:use create-env
	cl
	utils
	gw
	set
	prob
	inst-vars
	mdp-env
	bnet)
  )

(in-package forage)


(defun create-forage-env (wmap food-locs food-energies init-pos init-energy goal-energy)
  "create-forage-env WMAP FOOD-LOCS FOOD-ENERGIES INIT-POS INIT-ENERGY GOAL-ENERGY

WMAP - a 2d array representing the world map, as per the conventions in the grid-world package.  t represents free squares and nil represents obstacles.
FOOD-LOCS - a vector of food locations, each of which is a list (R C)
FOOD-ENERGIES - a vector of positive integers
INIT-POS - a location
INIT-ENERGY, GOAL-ENERGY - fixnums

Returns the corresponding forage environment."
  (make-2tbn-mdp-env 
   (make-forage-2tbn  wmap food-locs food-energies init-pos init-energy)
   :term-pred #'(lambda (s) (let ((e (forage-env-state-energy s))) (or (>= e goal-energy) (<= e 0))))))





(defstruct forage-env-state
  pos
  energy
  loc-status
  food-obtained
  wmap
  food-energies
  food-locs
  loc-food)



(defun make-pos-desc (wmap init-pos)
  (make-2tbn-var-desc :id 'pos :domain (apply #'direct-product 'list (array-dimensions wmap))
		      :prev-slice-parents '(pos action) :init-slice-parents nil
		      :trans-dist (make-deterministic-cpd (prev-pos action)
				      (result-legal wmap prev-pos action))
		      :init-dist (make-deterministic-cpd () init-pos)))

(defun make-food-obtained-desc ()
  (make-2tbn-var-desc :id 'food-obtained :domain (disjoint-union '(()) 'natural-numbers)
		      :prev-slice-parents '(pos action loc-food loc-status)
		      :trans-dist (make-deterministic-cpd (pos action loc-food loc-status)
				      (when (eq action 'f)
					(let ((f (apply #'aref loc-food pos)))
					  (when (and (>= f 0) (aref loc-status f)) f))))
		      :init-dist (make-deterministic-cpd () nil)))

(defun make-energy-desc (init-energy)
  (make-2tbn-var-desc :id 'energy :domain 'natural-numbers
		      :prev-slice-parents '(energy) :curr-slice-parents '(food-obtained food-energies)
		      :trans-dist (make-deterministic-cpd (energy food-obtained food-energies)
				      (+ (1- energy) (iwhen food-obtained (aref food-energies food-obtained))))
		      :init-dist (make-deterministic-cpd () init-energy)))


(defun make-loc-status-desc (n)
  (make-2tbn-var-desc :id 'loc-status :domain (apply #'direct-product 'vector (make-list n :initial-element '(nil t)))
		      :prev-slice-parents '(loc-status) :curr-slice-parents '(food-obtained)
		      :trans-dist (make-deterministic-cpd (loc-status food-obtained)
				      (if food-obtained
					  (let ((new-status (make-array n :element-type 'boolean :initial-contents loc-status)))
					    (setf (aref new-status food-obtained) nil)
					    new-status)
					loc-status))
		      :init-dist (make-deterministic-cpd () (make-array n :element-type 'boolean :initial-element t))))
					



(defun make-forage-2tbn (wmap food-locs food-energies init-pos init-energy)
  (dbind (m n) (array-dimensions wmap)
    (let ((loc-food (make-array (list m n) :element-type 'fixnum :initial-element -1)))
      ;; make inverse index from map to food location id
      (do-elements (l food-locs nil i)
	(setf (apply #'aref loc-food l) i))
      
      (make-instance '<2tbn>
    
	:state-descs
	(list
	 (make-pos-desc wmap init-pos)
	 (make-energy-desc init-energy)
	 (make-loc-status-desc (length food-locs))
	 (make-food-obtained-desc)
	 (make-param-desc 'world-map wmap)
	 (make-param-desc 'food-energies food-energies)
	 (make-param-desc 'food-locs food-locs)
	 (make-param-desc 'loc-food loc-food))
	
	:action-descs
	(make-2tbn-var-desc :id 'action :domain '(N S E W F))
	
	:reward-descs
	(make-2tbn-var-desc :id 'reward :domain '(0 1) 
			    :prev-slice-parents '() :curr-slice-parents '(energy)
			    :reward-fn (make-cpd (ce) 
					 (if (zerop ce) '-infty -1)))
	
	:state-acc (make-struct-accessors 
		    forage-env-state
		    (pos energy loc-status food-obtained wmap food-energies food-locs loc-food))
	
	:action-acc (make-single-object-accessors)
	:reward-acc (make-vec-accessors 1)))))
					  
					   
	
	




(defun pprint-forage-env-state (str s)
  (let ((wm (forage-env-state-wmap s))
	(loc-food (forage-env-state-loc-food s))
	(pos (forage-env-state-pos s))
	(loc-status (forage-env-state-loc-status s)))
    (pprint-logical-block (str nil)
      (dbind (nr nc) (array-dimensions wm)
	(flet ((horiz-wall ()
		 (format str "~v@{~a~:*~}" (+ nc 2) #\X)))
	  (horiz-wall)
	  (dotimes (i nr)
	    (format str "~:@_X")
	    (dotimes (j nc)
	      (write-char 
	       (if (equal pos (list i j))
		   #\A
		 (if (aref wm i j)
		     (let ((f (aref loc-food i j)))
		       (if (>= f 0)
			   (if (aref loc-status f)
			       #\F
			     #\f)
			 #\Space))
		   #\X))
	       str))
	    (write-char #\X str))
	  (pprint-newline :mandatory str)
	  (horiz-wall)
	  (format str "~:@_Energy : ~a" (forage-env-state-energy s)))))))
	  
      

(set-pprint-dispatch 'forage-env-state #'pprint-forage-env-state)
  
  
 

    