(in-package othello)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; sets of locations
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun row (r cmin cmax)
  (make-image-set (1+ (- cmax cmin))
		  #'(lambda (i)
		      (make-loc r (+ cmin i)))))

(defun col (c rmin rmax)
  (make-image-set (1+ (- rmax rmin))
		  #'(lambda (i)
		      (make-loc (+ rmin i) c))))

(defun box (rmin rmax cmin cmax)
  (make-image-set
   (direct-product 'list (1+ (- rmax rmin)) (1+ (- cmax cmin)))
   #'(lambda (x) (make-loc (+ rmin (first x)) (+ cmin (second x))))))
		  

(defun process-board-or-size (b)
  (etypecase b
    ((and (integer 1) (satisfies evenp)) b)
    (othello-board (array-dimension b 0))
    ))



(defun corners (b)
  (let ((n (1- (process-board-or-size b))))
    `((0 0) (0 ,n) (,n 0) (,n ,n))))

(defun precorners (b)
  (let ((n (- (process-board-or-size b) 2)))
    `((1 1) (1 ,n) (,n 1) (,n ,n))))


(defun non-corner-edges (b)
  (let ((n (1- (process-board-or-size b))))
    (disjoint-union
       (row 0 1 (1- n))
       (row n 1 (1- n))
       (col 0 1 (1- n))
       (col n 1 (1- n)))))

(defun non-precorner-internal (b)
  (let ((n (process-board-or-size b)))
    (disjoint-union
       (row 1 2 (- n 3))
       (row (- n 2) 2 (- n 3))
       (box 2 (- n 3) 1 (- n 2)))))

(defun noncorners (b)
  (let ((n (1- (process-board-or-size b))))
    (disjoint-union
     (row 0 1 (1- n))
	    (row n 1 (1- n))
	    (box 1 (1- n) 0 n))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Various squarewise features
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun count-squares (b v &optional (s (all-locations b)))
  "Count squares in S whose value is V."
  (sum-over s 
	    #'(lambda (x)
		(indicator (eq v (loc-contents b x))))))


(defun square-diff (b c s)
  "square-diff BOARD COLOUR SET.  Returns the number of squares in SET occupied by pieces of type COLOUR minus the number of squares in SET occupied by pieces of the opposite colour."
  (let ((opp (opposite-colour c)))
    (sum-over s
	      #'(lambda (l)
		  (let ((contents (loc-contents b l)))
		    (cond
		     ((eq contents c) 1)
		     ((eq contents opp) -1)
		     ((eq contents 'empty) 0)
		     (t (assert nil nil "Unknown location contents ~a in square-diff" contents))))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; features and state abstractors
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun binned-state-abstractor (b c num-corner-bins num-noncorner-bins num-empty-bins &aux (n (process-board-or-size b)))
  "Returns an abstractor that returns a list with three numbers: advantage on corners, binned into NUM-CORNER-BINS, advantage on noncorners, binned into NUM-NONCORNER-BINS, and number of empty squares, binned into NUM-EMPTY-BINS (and flipped around)."
  (check-type num-empty-bins (integer 2) "an integer greater than 1")
  (let ((corners (corners n))
	(noncorners (noncorners n))
	(num-squares (* n n)))
    (let ((num-corners (size corners))
	  (num-noncorners (size noncorners))
	  (min-num-empty (- (- num-squares 4) (* num-empty-bins (/ (- num-squares 4) (1- num-empty-bins))))))
      #'(lambda (b)
	  (list
	   (classify-bin (square-diff b c corners) (- num-corners) num-corners num-corner-bins)
	   (classify-bin (square-diff b c noncorners) (- num-noncorners) num-noncorners num-noncorner-bins)
	   (- (1- num-empty-bins) (classify-bin (count-squares b 'empty) min-num-empty (- num-squares 4) num-empty-bins)))))))

(defun make-q-featurizer (n c num-phases)
  "make-q-featurizer N C NUM-PHASES.
The featurizer works as follows.  For each phase of the game, there are 5 features: a constant, and the square difference for corners, precorners, edges, and internal squares."

  (let ((b2 (make-array (list n n) :element-type 'board-contents))
	(square-sets
	 (list (corners n) (precorners n) (non-corner-edges n) (non-precorner-internal n)))
	(normalizers '(1 1 5 6))
	(max-num-empty (- (* n n) 4)))
    
    
    #'(lambda (b a)
	(make-move-on-board b c a b2)
	(let ((v (make-array (* 5 num-phases) :element-type 'float :initial-element 0.0))
	      (ind (* 5 (- (1- num-phases) (classify-bin (count-squares b2 'empty) 0 max-num-empty num-phases)))))
	  (setf (aref v ind) 1)
	  (loop
	      for i from (1+ ind)
	      for s in square-sets
	      for n in normalizers
	      do (setf (aref v i) (/ (square-diff b2 c s) n))
	      finally (return v))))))
	       
	  