(defpackage othello 
  (:use
   cl
   set
   env
   policy
   utils)
  (:export 
   empty
   black
   white
   human
   pass
   
   <othello-env>
   set-colour
   
   <square-weighted-policy>
   binned-state-abstractor
   make-q-featurizer
   
   othello-io-interface
   make-init-board
   ))

(in-package othello)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Constants
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defparameter *opposite-colours* '((black . white) (white . black)))
(defparameter *directions* '(N NE E SE S SW W NW))
(defparameter *direction-vectors* '((N -1 0) (NE -1 1) (E 0 1) (SE 1 1) (S 1 0) (SW 1 -1) (W 0 -1) (NW -1 -1)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Types
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(deftype board-contents ()
  "Type for contents of a square in an Othello board."
  '(member empty black white))

(deftype colour () "black or white" '(member black white))
  

(defun othello-board-p (b)
  (and (arrayp b)
       (= (array-rank b) 2)
       (= (array-dimension b 0) (array-dimension b 1))
       (evenp (array-dimension b 0))
       (typep (aref b 0 0) 'board-contents)))

(deftype othello-board ()
  "An othello board is an NxN array, where N is even, and the contents of each cell are either 'black, 'white, or 'empty."
  '(satisfies othello-board-p))


(defstruct (loc (:constructor make-loc (r c)) (:type list))
  r c)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; High-level
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun legal? (b c l)
  "legal? BOARD COLOUR MOVE.  Is MOVE a legal move for COLOUR on BOARD?  MOVE may either be a location or the symbol 'pass."
  (if (eq l 'pass)
      (null (legal-locations b c))
    (and (eq (loc-contents b l) 'empty)
	 (some #'(lambda (dir) (> (num-steps-till-same-colour b l dir c) 0)) *directions*))))

(defun make-move-on-board (b c l &optional (new-board t))
  "make-move-on-board BOARD COLOUR MOVE &optional (NEW-BOARD t).  Returns board state resulting from making the given move on BOARD.  If NEW-BOARD is nil, the original board object is modified.  If NEW-BOARD is T, a new board object is created.  If NEW-BOARD is a board object, the result is stored in board."
  (let ((b2 (if (eq new-board t)
		(copy-board b)
	      (if (null new-board) b (copy-board b new-board)))))
    (if (eq l 'pass)
	b2
      (progn
	(set-loc-contents b2 l c)
	(dolist (dir *directions* b2)
	  (let ((l2 (copy-loc l)))
	    (dotimes (i (num-steps-till-same-colour b2 l dir c))
	      (set-loc-contents b2 (move-in-dir l2 dir) c))))))))


(defun game-over? (b)
  "Game is over if neither player can legally move."
  (and (legal? b 'black 'pass) (legal? b 'white 'pass)))

(defun score (b)
  "Returns an association list containing the number of squares of each type."
  (let ((s (pairlis '(black white empty) '(0 0 0))))
    (do-elements (l (all-locations b) s)
      (let ((p (assoc (loc-contents b l) s)))
	(assert s () "Loc value of ~a in ~a was illegal" l b)
	(incf (cdr p))))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Board processing utilities
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun opposite-colour (c)
  (mapping:evaluate *opposite-colours* c))

(defun loc-contents (b l)
  "Return contents of location on board, or 'off-board if the location is not on the board."
  (let ((r (loc-r l))
	(c (loc-c l)))
    (if (and (between2 r 0 (array-dimension b 0))
	     (between2 c 0 (array-dimension b 1)))
	(aref b r c)
      'off-board)))

(defun set-loc-contents (b l c)
  "Set contents of a location on the board.  Returns nothing."
  (assert (not (eq (loc-contents b l) 'off-board)) ()
    "Tried to set location ~a which is not on board ~a" l b)
  (setf (aref b (loc-r l) (loc-c l)) c))

(defun move-in-dir (l d)
  "Modify L to hold the location obtained by moving one step in direction D, and return the new location as well.  Does not worry about legality."
  (dbind (delta-r delta-c) (mapping:evaluate *direction-vectors* d)
	 (incf (loc-r l) delta-r)
	 (incf (loc-c l) delta-c)
	 l))
	 

(defun num-steps-till-same-colour (b l dir c)
  "num-steps-till-same-colour BOARD LOCATION DIRECTION COLOUR.
BOARD is an othello board.
LOCATION is a location on that board.
DIRECTION is a direction.
COLOUR is 'black or 'white.

The magic function that everything depends on.  Suppose there is a positive integer K such that the next K squares in direction DIR from LOCATION are all of the opposite colour to COLOUR, and the one following is of the same colour.  Then returns K.  Otherwise, returns 0."
  (let ((opp (opposite-colour c))
	(l2 (copy-loc l)))
    
    (loop
	for k from 0
	for contents = (loc-contents b (setf l2 (move-in-dir l2 dir)))
		       
	when (eq contents c)
	return k
		
	unless (eq contents opp)
	return 0)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; sets of locations and moves
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun legal-locations (b c)
  (filter ':list (all-locations b)
	  #'(lambda (l) (legal? b c l))))

(defun legal-moves (b c)
  (or (legal-locations b c) '(pass)))

(defun all-locations (b)
  (make-image-set
   (make-instance '<prod-set> :sets (array-dimensions b) :inst-acc (inst-vars:make-vec-accessors 2) :iterate-quickly t)
   #'(lambda (x) (apply #'make-loc x))))





  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; General utilities
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun make-init-board (&optional (n 8))
  (let ((a (make-array (list n n) :element-type 'board-contents :initial-element 'empty)))
    (let ((k (floor n 2)))
      (setf (aref a (1- k) (1- k)) 'black
	    (aref a (1- k) k) 'white
	    (aref a k (1- k)) 'white
	    (aref a k k) 'black)
      a)))


(defun copy-board (b &optional (target nil))
  (let ((a (if target target (make-array (array-dimensions b) :element-type 'board-contents))))
    (dotimes (i (array-total-size b) a)
      (setf (row-major-aref a i) (row-major-aref b i)))))

