(in-package othello)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; class defs
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <othello-policy> (<policy>)
  ((my-colour :type colour :reader my-colour :initarg :my-colour :writer set-colour)))

(defclass <square-weighted-policy> (<othello-policy>)
  ((weighting-function :type function :reader weight-fn :writer set-weight-fn))
  (:documentation "Class <square-weighted-policy>.  Policy for Othello that looks ahead one step and picks the move that maximizes the weights of my squares minus the weights of my opponent's squares.

Initargs
:weighting-function - function from i, j to floats that returns the weight of square i,j.  If not supplied, then the weight matrix is used instead.
:weight-matrix - 2d array containing the weights.  If weight function is not supplied, this must be."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; constructor
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod initialize-instance :after ((p <square-weighted-policy>) &rest args &key (weighting-function nil wf-supp) (weight-matrix nil wm-supp))
  (declare (ignore args))
  (assert (xor wf-supp wm-supp) nil
    "Exactly one of the weighting-function and weight-matrix must be supplied.")
  (set-weight-fn
   (aif weighting-function
       it
     (let ((mat (clone weight-matrix)))
       (dbind (m n) (array-dimensions mat)
	 (assert (= m n) nil "Dimensions of weight matrix must be equal.")
	 #'(lambda (i j)
	     (assert (and (between2 0 i m) (between2 0 j m)) nil
	       "Weight matrix does not contain a weight for row ~a, column ~a." i j)
	     (aref mat i j)))))
   p))
	   
	   

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; making choices
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


	   
(defmethod make-choice ((pol <square-weighted-policy>) b &optional choices)
  (declare (ignore choices))
  (let ((c (my-colour pol))
	(weights (weight-fn pol)))
    (let ((b2 (copy-board b))
	  (opp (opposite-colour c))
	  (moves (legal-moves b c)))
    
      (mvbind (ind v best best-inds)
	  (utils:argmax moves
			:key #'(lambda (m)
				 (make-move-on-board b c m b2)
				 (let ((s
					(sum-over
					 (all-locations b2)
					 #'(lambda (l)
					     (* (funcall weights (loc-r l) (loc-c l))
						(let ((contents (loc-contents b2 l)))
						  (cond
						   ((eql contents c) 1)
						   ((eql contents opp) -1)
						   (t 0))))))))
				   ;(format t "~&Move ~a Score ~a" m s)
				   s)))
	(declare (ignore ind v best))
	(item (prob:sample-uniformly best-inds) moves)))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Example weighting functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun positional-weighting (n)
  #'(lambda (i j)
      (flet ((standardize (x)
	       (if (>= x (/ n 2))
		   (- (1- n) x)
		 x)))
	(let ((i2 (max (standardize i) (standardize j)))
	      (j2 (min (standardize i) (standardize j))))
	  (if (= j2 0)
	      (if (= i2 0) 5
		(if (= i2 1) -3 3))
	    (if (and (= j2 1) (= i2 1)) -4
	      1))))))
		  
	  
	  

      