; maxq0-q-fn.lisp
; the maxq0 q-function, which includes a linear and tabular featurizer, and also supports a subtask-table to determine
; the subtask hierarchy

(in-package maxq0)

(defclass <maxq0-q-function> (calisp:<calisp-approx-q-function>)
  ((linear-fn-approx :accessor linear-fn-approx :initform (make-instance 'fn-approx:<linear-fn-approx> :dim 3))   
   (tabular-fn-approx :accessor tabular-fn-approx :initform (make-instance 'fn-approx:<tabular-fn-approx>))
   (subtask-table :accessor st
    :initarg :st)))

; dummy function, not really used by maxq0 algorithm
(defmethod update ((q-fn <maxq0-q-function>) omega u target eta)
    (dolist (tid (get-thread-ids omega))
      (update-linear-or-tabular q-fn (featurize q-fn omega u tid (get-subtask omega tid)) target eta)))

; updates a composite q-value according to a particular thread's features and its choice
(defmethod update-tid ((q-fn <maxq0-q-function>) omega u target eta tid)
  (update-linear-or-tabular q-fn (featurize q-fn omega u tid (get-subtask omega tid)) target eta))

; updates a primitive subtask's q-value according to a particular thread's features
(defmethod update-v ((q-fn <maxq0-q-function>) omega i target eta tid)
  (update-linear-or-tabular q-fn (featurize q-fn omega nil tid i) target eta))

; standard clone function
(defmethod clone ((q-fn <maxq0-q-function>))
  (make-instance '<maxq0-q-function> :linear-fn-approx (clone (linear-fn-approx q-fn)) :tabular-fn-approx (clone (tabular-fn-approx q-fn)) :st (st q-fn)
))

; evaluates the current joint state and choice; basically loops through and sums up the max-node value of
; each thread, according to the MAXQ0 algorithm described in the paper
(defmethod evaluate ((q-fn <maxq0-q-function>) omega u)
  (let ((sum 0))
    (dolist (tid (get-thread-ids omega))
      (let ((subtask (get-subtask omega tid)))
	(setf sum (+ sum (car (evaluate-max-node q-fn omega tid subtask))))))
    sum))

; featurize simply calls the featurizer on the arguments; note that we need a thread id and a subtask argument
; in order to determine what features to use
(defmethod featurize ((q-fn <maxq0-q-function>) omega u tid i)
  (funcall (q-function::featurizer q-fn) omega u tid i))

; get the list of thread ids from omega
(defun get-thread-ids (omega)
  (let ((list-of-ids nil))
    (maphash (lambda(k v) (setf list-of-ids (append list-of-ids (list k)))) (js-thread-states omega)))
  list-of-ids)

; gets the choice associated with the thread with id tid from the joint choice u
(defun get-choice (u tid)
  (if (not (listp u)) u
    (cdr (assoc-if #'(lambda(x) (equal tid x)) u))))

; evaluate-max-node works as described in the paper 
; Note on implementation: we depend on the subtask tree table to generate the recursive call that evaluates the argmax(a) 
; of children subtasks.  This requires that the featurizer recognize these as the same symbols, since we are generating
; the j's from the hash table 

(defmethod evaluate-max-node-returns-subtask ((q-fn <maxq0-q-function>) omega tid subtask)
  (if (primitive-subtask? subtask (st q-fn))
      (cons (evaluate-linear-or-tabular q-fn (featurize q-fn omega nil tid subtask)) subtask)
      (let ((children-subtasks (get-children-subtasks subtask (st q-fn))))
        (let ((max-so-far (cons 0 nil)))
	  (dolist (j children-subtasks)
	    (let* ((recursive-evaluate (evaluate-max-node-returns-subtask q-fn omega tid j))
	           (j-value (+ (car recursive-evaluate) (evaluate-linear-or-tabular q-fn (featurize q-fn omega (acons tid j '()) tid subtask)))))
	      (if (> j-value (car max-so-far))
              (setf max-so-far (cons j-value j)))))
	  max-so-far))))
	
(defmethod evaluate-max-node ((q-fn <maxq0-q-function>) omega tid subtask)
  (if (primitive-subtask? subtask (st q-fn))
      (evaluate-linear-or-tabular q-fn (featurize q-fn omega nil tid subtask))
      (let ((children-subtasks (get-children-subtasks subtask (st q-fn))))
        (let ((max-so-far 0))
	  (dolist (j children-subtasks)
	    (let* ((recursive-evaluate (evaluate-max-node q-fn omega tid j))
	           (j-value (+ recursive-evaluate (evaluate-linear-or-tabular q-fn (featurize q-fn omega j tid subtask)))))
	      (if (> j-value max-so-far)
              (setf max-so-far j-value))))
	  max-so-far))))
	
; get-subtask simply returns the current label of the thread given as it appears in omega
(defun get-subtask (omega tid)
  (let ((st (js-thread-label omega tid)))
    st))


;; these are references into the subtask tree hash table to find the children of a subtask and to
;; ask whether a subtask is primitive (defined as having no children)

(defun get-children-subtasks (subtask st)
  (gethash subtask st)
)

(defun primitive-subtask? (subtask st)
  (null (gethash subtask st))
)

; update-linear-or-tabular updates either the linear q-function or the tabular q-function, depending on
; whether the car of the features is 'linear or 'tabular
(defmethod update-linear-or-tabular ((q-fn <maxq0-q-function>) features target eta)
    (if (equalp (car features) 'linear)
	(fn-approx:update (linear-fn-approx q-fn) (cdr features) target eta)
        (fn-approx:update (tabular-fn-approx q-fn) (cdr features) target eta)))

; evaluate-linear-or-tabular evaluates either the linear q-function or the tabular q-function, depending on
; whether the car of the features is 'linear or 'tabular
(defmethod evaluate-linear-or-tabular ((q-fn <maxq0-q-function>) features)
    (if (equalp (car features) 'linear)
	(fn-approx:evaluate (linear-fn-approx q-fn) (cdr features))
        (fn-approx:evaluate (tabular-fn-approx q-fn) (cdr features))))