(defpackage maxq0
  (:documentation
   "Package for the MAXQ0 algorithm and associated feature templates for concurrent ALisp programs.
") 
  (:use common-lisp
	utils
	policy
	calisp-obs)
  (:import-from calisp js-thread-states ts-label js-thread-label)
  (:export
        <maxq0>
	<maxq0-q-function>))

(in-package maxq0)

; debugging flag
(setf maxq0_debug_flag nil)

(defclass <maxq0> (<q-learning-algorithm> <calisp-learning-algorithm>)
   ((prev-omega :accessor prev-omega) ; the omega observed by the most recent inform-calisp-step
    (prev-u :accessor prev-u)         ; the joint choice observed by the most recent inform-calisp-step
    (prev-s :accessor prev-s)         ; the next environment state observed by the most recent inform-env-step
    (threads-current-subtasks :accessor threads-current-subtasks) ; what subtasks threads are currently at, updated in inform-calisp-step
    (threads-current-choice-blocks :accessor threads-current-choice-blocks) ; what subtask choice blocks threads are currently in, updated in inform-calisp-step
    (initq :accessor initq)
    (time_elapsed :accessor time_elapsed :initform 0) ; this is used for the eta function, which might vary with time
   (q-function :reader q-fn :type <calisp-q-function> :initarg :q-fn :writer set-q-fn)
   (eta-fn :reader eta-fn :initarg eta-fn :initform #'eta-linear-tabular)
   (reward-decomposer :reader reward-decomposer :initarg :rd)) ; the reward-decomposer function; takes OMEGA STATE ACT REWARD NEXT-STATE as arguments and returns
										   ; an assocation list consisting of thread id keys and reward values (corresponding to the reward 
										   ; assigned to that thread)
  (:documentation "Class <maxq0>.  Subclass of <calisp-learning-algorithm> that uses MAXQ0-learning.

Initargs
:t-lrate learning rate
:q-fn q-function"))

(defmethod reset ((alg <maxq0>)) )

; standard initialize for learning algorithms

(defmethod initialize-instance :after ((alg <maxq0>) &rest args &key q-function)
  (declare (ignore args))
  (setf (initq alg) (clone q-function)))

; simply initializes all the state variables we keep

(defmethod inform-start-episode ((alg <maxq0>) s)
  (setf (prev-s alg) s)
  (slot-makunbound alg 'prev-omega)
  (slot-makunbound alg 'prev-u)
  (slot-makunbound alg 'threads-current-subtasks)
  (setf (threads-current-subtasks alg) (make-hash-table :test 'equal))
  (setf (threads-current-choice-blocks alg) (make-hash-table :test 'equal)))

; in inform-calisp-step, we must store the choice-blocks threads have entered and keep
; track of prior omegas; this is because there is no other way to refer to prior omegas after
; another inform-calisp-step has occured, but maxq0 requires that we wait until the end
; of a subtask step before updating it - this often requires going beyond the immediate omega
; since the subtask might be composite

(defmethod inform-calisp-step ((alg <maxq0>) omega u)
  (d_format t "INFORM CALISP STEP:~%")
  (maphash #'(lambda(k v) (d_format t "~s thread at ~s with joint choice~s~%" 
       k (ts-label v) u)) (js-thread-states omega))
  (setf (prev-omega alg) omega
	(prev-u alg) u)
  (update-current-subtasks alg omega u)
  (update-current-choice-blocks alg omega u))

; Updates the current subtask of the thread according to what was chosen 
; in u; note that this might be obsolete once update current choice blocks
; has been implemented; is here for legacy purposes

(defmethod update-current-subtasks ((alg <maxq0>) omega u)
  (dolist (tchoice u)
    (let ((tid (car tchoice))
	  (tsub (cdr tchoice)))
      (d_format t "updating current subtask of ~s to ~s~%" tid tsub)
      (setf (gethash tid (threads-current-subtasks alg)) tsub))))

; At each calisp step, we update the current choice block of each thread the 
; idea is to store a pair containing a omega and a subtask choice for each 
; thread, which would correspond to its most recent state and action in the 
; MAXQ0 algorithm.  This is necessary because at the end of composite subtask
; a we need to update C(i, s, a), where s was the env state in which we did a
; and i was the higher level subtask we were in (and is presumably returning to
; now that a is done).  Thus, they must be stored until we have the resulting
; state, s' (corresponds to the omega observed at the inform-end-choice-block 
; for that particular thread)

; NOTE: Since threads with no choices (ie idle threads) do not have 
; inform-end-choice-blocks, we don't store their choice blocks, so as to avoid
; memory leakage from not ever popping off choice blocks; we detect such
; threads by noticing that they are not present in the joint choice u

(defmethod update-current-choice-blocks ((alg <maxq0>) omega u)
  (dolist (tchoice u)
    (let ((tid (car tchoice))
	  (tsub (cdr tchoice)))
      (d_format t "PUSHING choice-block for:~%~s at~%~s doing~%~s on to the stack~%" tid (js-thread-label omega tid) tsub)
      (setf (gethash tid (threads-current-choice-blocks alg))
	    (acons omega tsub (gethash tid (threads-current-choice-blocks alg)))))))

; inform-env-step occurs once per environment step; for maxq0, we do the update for primitive
; subtasks at this point because each environment step corresponds to an action done in the environment,
; observing action, reward, and the resulting s'.  We also call the reward distributor at this point
; and interpret its results.  Note that we know the primitive subtask that we just did from storing
; the most recent choice of each thread, which must be a primitive action if it was just before inform-env-step

; note also that we can safely ignore threads that received no rewards because there is no target value
; in that case

(defmethod inform-env-step ((alg <maxq0>) act rew next term)
  (when (slot-boundp alg 'prev-omega)
    (let ((rewards (funcall (reward-decomposer alg) (prev-omega alg) (prev-s alg) act rew next)))
       (d_format t "at this step received rewards: ~s~%" rewards)
       (d_format t "hash table for thread subtasks looks like this:~%")
       (printhash (threads-current-subtasks alg))
       (dolist (trew rewards) ; should be rewards
  	   (let* ((tid (car trew))
	         (drew (cdr trew))
		 (t_subtask (gethash tid (threads-current-subtasks alg))))
	     (d_format t "attempting to gethash on ~s returned ~s~%" tid t_subtask)
	     (update-v (q-fn alg) (prev-omega alg) t_subtask drew (funcall (eta-fn alg) t_subtask (time_elapsed alg)) tid)))))
    (setf (prev-s alg) next))

; Here, a choice block has ended.  Therefore, we should update the composite
; subtask corresponding to the choice block.  This is by referencing to the
; most recent state-action pair we stored earlier in inform-calisp-step and
; making a call to evaluate-max-node.  We then pop off the most recent 
; state-action pair, so as to signal that this choice block has ended and that
; we're returning to the higher level subtask

(defmethod inform-end-choice-block ((alg <maxq0>) tid omega)
  (let* ((t-blocks (gethash tid (threads-current-choice-blocks alg)))
	 (tc-block (first t-blocks))
	 (tc-block-omega (car tc-block))
	 (tc-block-choice (cdr tc-block)))

    ; debug
    (d_format t "POPPING choice-block for thread:~%~s at~%~s doing~%~s off the stack~%" tid (js-thread-label tc-block-omega tid) tc-block-choice)
    
    ; update
    (update-tid (q-fn alg) tc-block-omega tc-block-choice
     (evaluate-max-node (q-fn alg) omega tid (get-subtask tc-block-omega tid))
     (funcall (eta-fn alg) (get-subtask tc-block-omega tid) (time_elapsed alg))
     tid)

    ; pops off current choice block
    (setf (gethash tid (threads-current-choice-blocks alg)) 
	  (cdr (gethash tid (threads-current-choice-blocks alg))))
))
  
; if subtask is any goto then set learning rate as 0.001, otherwise 0.01

(defun eta-linear-tabular (subtask time)
  (if (or (equalp subtask 'goto0)
	  (equalp subtask 'goto1)
	  (equalp subtask 'goto2)
	  (equalp subtask 'goto3) 
	  (equalp subtask 'goto4)
	  (equalp subtask 'goto5) 
	  (equalp subtask 'goto6)
	  (equalp subtask 'goto7) 
	  (equalp subtask 'goto8)
	  (equalp subtask 'goto9) 
	  (equalp subtask 'goto10)
	  (equalp subtask 'goto11) 
	  (equalp subtask 'goto12)
	  (equalp subtask 'goto13)
	  (equalp subtask 'goto14)
	  (equalp subtask 'goto15))
    0.001
    0.01))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; knowledge state 
;; required for learning algorithm interface
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod knowledge-state ((alg <maxq0>) &optional (fresh t))
  (let ((q (q-fn alg)))
    (if fresh (clone q) q)))

(defmethod get-q-fn ((alg <maxq0>) ks)
  ks)


;;;;;;;;;;;;;
;; DEBUG
;;;;;;;;;;;;;

;; debugging function; does the same thing as format when maxq0_debug_flag turned on

(defun d_format (destination control-string &rest args)
  (when (equal maxq0_debug_flag t)
    (funcall #'format destination control-string args)))
    
; debugging function for printing out hash table
(defun printhash (h)
  (maphash #'(lambda(k v) (d_format t "~s maps to ~s~%" k v)) h))
