;; resource-env.lisp
;; Simulates the resource environment, in which peasants attempt to execute several hierarchical actions
;; in pursuing the goal of harvesting a certain amount of gold and wood
;; Adapted from original design by Oregon State MAXQ TL team
;; Works with stratagus maps of size 32x32, where the only player is player 1 (red), and where the only
;; units are, aside from a townhall and peasants for player 1, trees and gold mines
;; Developed to be used with resource-prog.lisp, which defines the partial program for this environment

(defpackage resource-env
  (:use common-lisp
	utils
	stratagus-env
  )
  (:export <resource-env>
           *target-gold*
	   *target-wood*))
	   
	   
(in-package resource-env)

; debugging flag

(setf resource_debug_flag nil)

; the resource environment
; *target-gold* the target amount of gold to collect
; *target-wood* the target amount of wood to collect

(defvar *target-gold* 1000)
(defvar *target-wood* 500)

(defclass <resource-env> (<stratagus-env>) 
  ())

; peasant? just returns whether a particular effector is refers to a peasant

(defun peasant? (id s)
  (= 2 (unit-type (get-unit-state id s))))

; the compute-reward function finds the amount of reward for each peasant
; and returns it as an association list of the received numerical rewards for peasants
; in this step keyed on the peasant ids, to be processed by the reward-decomposer
; this is done as follows:
; 	if peasant was executing a dropoff action and actually succeeded in doing the dropoff 
;      (based on its status args changing), and the resource dropped off had not met the target
;	 before the dropoff, then assign the peasant 99 reward (100, minus one for the step cost)
;	else return -1.0, for the step cost

(defmethod compute-reward ((e <resource-env>) s a next-s)
  (let ((rew-list '()))
    (dolist (peasant-action a)
      (if (peasant? (car peasant-action) s)
        (let ((action-type (cdr peasant-action))
	      (peasant-id (car peasant-action)))
          (if (equalp action-type "7") ; dropoff
	      (let ((peasant-prev-state (get-unit-state peasant-id s))
		    (peasant-next-state (get-unit-state peasant-id next-s)))
    	        (if (or (and 
		         (equalp (unit-status-args peasant-prev-state) '(1))
		         (equalp (unit-status-args peasant-next-state) nil)
		         (< (global-state-gold s) *target-gold*))
		        (and
		         (equalp (unit-status-args peasant-prev-state) '(2))
		         (equalp (unit-status-args peasant-next-state) nil)
		         (< (global-state-wood s) *target-wood*)))
		    (setf rew-list (acons peasant-id 99.0 rew-list))
		    (setf rew-list (acons peasant-id -1.0 rew-list))))
	      (setf rew-list (acons peasant-id -1.0 rew-list))))))
    (when (equal resource_debug_flag t) (format t "Distribute reward: ~s" rew-list))
    rew-list))
		

; the state terminates when we have achieved the target gold and wood

(defmethod is-terminal-state ((e <resource-env>) s) 
      (and (>= (global-state-gold (strat-es-global s)) *target-gold*)
	   (>= (global-state-wood (strat-es-global s)) *target-wood*)))

; reset procedure should be extended to set elapsed-cycles to 0 upon reset

(defmethod reset :after ((e <resource-env>) &optional (reset-to-nonterminal-state t) &aux (s (stratagus-env::stratagus-env-game-socket e))))

