;; The featurizer for the resource collection environment; consists of both the linear and tabular featurizer
;; Adapted from the Oregon State power point presentation for the resource collection environment
;; For more details on the specific features used, please consult their power point presentation

;; ASSUMES: At any time this is called, s consists of exactly one peasant thread
;; that is at a choice point, and a is that particular choice

;; we also needed an extra parameter, i, to describe the current subtask (MaxNode) we're in (corresponds to i in the paper)

(in-package resource-prog)

(defun make-resource-featurizer ()
    (lambda (s a tid i)
	(let ((subtask-label i))
;	  (print "ATTEMPTING TO FEATURIZE SUBTASK: ")
;	  (print subtask-label)
        (cond 
	    ;; depending on what subtask we're in, use the appropriate features (details described in the Oregon State power point
	    ;; presentation
	    ((equal subtask-label 'root) (list 'tabular 'Root (root-tabular-features s a tid)))
          ((equal subtask-label 'get_gold) (list 'tabular 'Get_Gold (get-gold-tabular-features s a tid)))
          ((equal subtask-label 'get_wood) (list 'tabular 'Get_Wood (get-wood-tabular-features s a tid)))
          ((equal subtask-label 'deposit) (list 'tabular 'Deposit (deposit-tabular-features s a tid)))
	    ((equal subtask-label 'harvest_Gold) (list 'tabular 'Harvest_Gold (harvest_gold_value s tid)))
	    ((equal subtask-label 'harvest_Wood) (list 'tabular 'Harvest_Wood (harvest_wood_value s tid)))
	    ((equal subtask-label 'dropoff) (list 'tabular 'Dropoff (dropoff_value s tid)))
	    ((equal subtask-label 'goto0) (cons 'linear (goto_value s tid subtask-label)))
	    ((equal subtask-label 'goto1) (cons 'linear (goto_value s tid subtask-label)))
	    ((equal subtask-label 'goto2) (cons 'linear (goto_value s tid subtask-label)))
	    ((equal subtask-label 'goto3) (cons 'linear (goto_value s tid subtask-label)))
	    ((equal subtask-label 'goto4) (cons 'linear (goto_value s tid subtask-label)))
	    ((equal subtask-label 'goto5) (cons 'linear (goto_value s tid subtask-label)))
	    ((equal subtask-label 'goto6) (cons 'linear (goto_value s tid subtask-label)))
	    ((equal subtask-label 'goto7) (cons 'linear (goto_value s tid subtask-label)))
	    ((equal subtask-label 'goto8) (cons 'linear (goto_value s tid subtask-label)))
	    ((equal subtask-label 'goto9) (cons 'linear (goto_value s tid subtask-label)))
	    ((equal subtask-label 'goto10) (cons 'linear (goto_value s tid subtask-label)))
	    ((equal subtask-label 'goto11) (cons 'linear (goto_value s tid subtask-label)))
	    ((equal subtask-label 'goto12) (cons 'linear (goto_value s tid subtask-label)))
	    ((equal subtask-label 'goto13) (cons 'linear (goto_value s tid subtask-label)))
	    ((equal subtask-label 'goto14) (cons 'linear (goto_value s tid subtask-label)))
	    ((equal subtask-label 'goto15) (cons 'linear (goto_value s tid subtask-label)))
	    (t (assert (= 0 1)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; COMPLETION FUNCTIONS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun root-tabular-features (s a tid)
  (list (agent_resource s tid) (requisite_gold s) (requisite_wood s) a))

(defun get-gold-tabular-features (s a tid)
  (list (agent_resource s tid) (if (equal a 'harvest_gold) 'harvest_gold 'goto) 
	(if (equal a 'harvest_gold) (agent_region_gold s (peasant-xy s tid)) (agent_region_gold s (region-to-coordinate (get-goto-k a))))))

(defun get-wood-tabular-features (s a tid)
   (list(agent_resource s tid) (if (equal a 'harvest_wood) 'harvest_wood 'goto) 
	(if (equal a 'harvest_wood) (agent_region_wood s (peasant-xy s tid)) (agent_region_wood s (region-to-coordinate (get-goto-k a))))))

(defun deposit-tabular-features (s a tid)
   (list (agent_resource s tid) (requisite_gold s) (requisite_wood s) 
     (if (equal a 'dropoff) 'dropoff 'goto) 
     (if (equal a 'dropoff) (agent_region_townhall s (peasant-xy s tid)) (agent_region_townhall s (region-to-coordinate (get-goto-k a))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; VALUE FUNCTIONS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun harvest_gold_value (s tid)
  (agent_region_gold s (peasant-xy s tid)))

(defun harvest_wood_value (s tid)
  (agent_region_wood s (peasant-xy s tid)))

(defun dropoff_value (s tid)
  (list (agent_resource s tid) (requisite_gold s) (requisite_wood s)
    (agent_region_townhall s (peasant-xy s tid))))

(defun goto_value (s tid i)
  (vector (agent_distance_x s tid i) (agent_distance_y s tid i) 1))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; FEATURE ACCESSORS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; features indicating whether we've reached the target amount of gold/wood in the environment

(defun requisite_gold (s)
  (if (<= *target-gold* (global-state-gold (strat-es-global (js-env-state s))))
    1
    0))

(defun requisite_wood (s)
  (if (<= *target-wood* (global-state-wood (strat-es-global (js-env-state s))))
    1
    0))

; given a thread id and joint state, returns a pair: the corresponding thread's peasant's x and y coordinate 
; in the environment

(defun peasant-xy (s tid)
  (let ((peasant-state (get-unit-state (cdr tid) (js-env-state s))))
      (cons (unit-x peasant-state) (unit-y peasant-state))))

;; returns what a peasant is currently carrying

;; NOTE: does not check whether a unit is attacking, repairing, etc.; this means that
;; the status args can only represent what the unit is carrying; probably have to change
;; this if peasant is allowed to attack

(defun agent_resource (s tid)
  (let ((peasant-state (get-unit-state (cdr tid) (js-env-state s))))
    (if (equalp (unit-status-args peasant-state) nil)
        'nothing
        (if (equalp (unit-status-args peasant-state) '(1))
            'gold
            (if (equalp (unit-status-args peasant-state) '(2))
                'wood)))))

;; checks if the 9x9 region around the agent coord provided as (x . y) has a unit of 
;; type ut

(defun agent_region_unit (s coord ut)
  (let ((units-of-type (units-belonging-to-of-type 0 ut (js-env-state s)))
	(peasant-x (car coord))
	(peasant-y (cdr coord))
        (in-region? 0))
    (dolist (i units-of-type)
      (let ((ux (unit-x (get-unit-state i (js-env-state s))))
	    (uy (unit-y (get-unit-state i (js-env-state s)))))
	(if (and (<= ux (+ peasant-x 4)) (>= ux (- peasant-x 4))
		 (<= uy (+ peasant-y 4)) (>= uy (- peasant-y 4)))
	  (setf in-region? 1))))
    in-region?))

(defun agent_region_gold (s coord)
  (agent_region_unit s coord 92)) ; gold mine

(defun agent_region_townhall (s coord)
  (agent_region_unit s coord 74)) ; town hall

; checks the map to see if the 9x9 region around the agent coord has a tree; note that this
; is necessary because trees are not counted as units we can directly access

(defun agent_region_wood (s coord)
  (let ((peasant-x (car coord))
	(peasant-y (cdr coord))
	(in-region? 0))
    (loop for x from (- peasant-x 4) to (+ peasant-x 4) until (= in-region? 1) do
      (loop for y from (- peasant-y 4) to (+ peasant-y 4) until (= in-region? 1) do
        (if (equalp (map-ref x y (global-state-map (strat-es-global (js-env-state s)))) #\T) ;; is tree
	    (setf in-region? 1))))
    in-region?))

;; these are linear features indicating the distance of the peasant's current coordinates from the target region, which is
;; specified by the particular goto(k) that was called

(defun agent_distance_x (s tid i)
  (let* ((peasant-state (get-unit-state (cdr tid) (js-env-state s)))
         (dest-loc (region-to-coordinate (get-goto-k i))))
    (abs (- (unit-x peasant-state) (car dest-loc)))))

(defun agent_distance_y (s tid i)
  (let* ((peasant-state (get-unit-state (cdr tid) (js-env-state s)))
         (dest-loc (region-to-coordinate (get-goto-k i))))
    (abs (- (unit-y peasant-state) (cdr dest-loc)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; UTIL FUNCTIONS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun get-goto-k (a)
  (cond ((equal a 'goto0) 0)
	  ((equal a 'goto1) 1)
	  ((equal a 'goto2) 2)
	  ((equal a 'goto3) 3)
	  ((equal a 'goto4) 4)
	  ((equal a 'goto5) 5)
	  ((equal a 'goto6) 6)
	  ((equal a 'goto7) 7)
	  ((equal a 'goto8) 8)
	  ((equal a 'goto9) 9)
	  ((equal a 'goto10) 10)
	  ((equal a 'goto11) 11)
	  ((equal a 'goto12) 12)
	  ((equal a 'goto13) 13)
	  ((equal a 'goto14) 14)
	  ((equal a 'goto15) 15)
	  (t 'unknown)))

;; references location (x, y) for two-dimensional map m (represented as list of letters)

(defun map-ref (x y m)
  (if (or (< x 0) (< y 0) (> x 31) (> y 31))
    0
    (if (not (symbolp (car (nth y m)))) 0 ; all grass line?
      (char (symbol-name (car (nth y m))) x))))

;; transforms a coordinate ([0,31], [0,31]) into its corresponding region in a 32x32 map divided into 16 regions
;; (done as specified in the Oregon State power point presentation)

(defun coordinate-to-region (x y)
  (if (and (< x 27) (< y 27))
    (+ (floor (/ x 9)) (* 3 (floor (/ y 9))))
    (if (< x 27)
      (+ (floor (/ x 9)) 9)
      (if (< y 27)
        (+ 12 (floor (/ y 9)))
        15))))

; old version
;(defun region-to-coordinate (k)
;  (cond ((= k 0) (cons 4 4))
;	((= k 1) (cons 13 4))
;	((= k 2) (cons 22 4))
;	((= k 3) (cons 4 13))
;	((= k 4) (cons 13 13))
;	((= k 5) (cons 22 13))
;	((= k 6) (cons 4 22))
;	((= k 7) (cons 13 22))
;	((= k 8) (cons 22 22))
;	((= k 9) (cons 4 29))
;       ((= k 10) (cons 13 29))
;       ((= k 11) (cons 22 29))
;       ((= k 12) (cons 29 4))
;	((= k 13) (cons 29 13))
;	((= k 14) (cons 29 22))
;	((= k 15) (cons 29 29))
;	(t 'unknown)))


; accesses the x and y coordinates of an unit-state

(defun unit-x (u)
  (car (unit-loc u)))

(defun unit-y (u)
  (cadr (unit-loc u)))

