;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; stratagus-env.lisp
;; Contains the top-level operations for Stratagus environments.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(in-package stratagus-env)







;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; global vars
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *socket* nil "The socket being used by the game.")
  


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; class definition
;; constructor takes single required initarg :game-socket, which must be a socket connected to a 
;; running instance of the game
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <stratagus-env> (<fully-observable-env>)
  ((game-socket :type socket-stream-internet-active :accessor stratagus-env-game-socket
		:initform (make-socket :remote-host 'localhost :remote-port 4870)
		:reader game-socket)
   (num-steps :type fixnum :accessor num-steps :initform 0)
   (world        :writer set-world :reader world :type array)
   (player-id    :writer set-pid :reader pid :type fixnum)
   (units-to-stop :accessor units-to-stop))
  (:documentation "<stratagus-env> class. Initargs
:game-socket - If provided, must be an active socket to an instance of the stratagus game.  If not provided, then at the time of creation of the instance, there must be a running instance of Stratagus listening for connections on port 4870, to which the instance will connect.

<stratagus-env> is a subclass of <env> that handles the details of the socket connection to the Stratagus executable, parsing of actions, etc.  Subclasses of <stratagus-env> will implement specific reward functions and termination conditions.

A legal action to a <stratagus-env> is an association list, where the keys are the unit IDs (which are integers), and the values are themselves lists, in which the first element is the action name (a symbol), and the remaining elements are arguments, as follows:
- (MOVE DIR) where DIR is one of the symbols n, s, e, or w.  This means try to move in the given direction until the next timestep.
- (MOVE-TOWARDS (C R)) where C and R are integers.  Moves towards the given location on the map for one step, then stop.
- (HARVEST RESOURCE) where RESOURCE is one of the symbols wood or gold.  Harvest this resource if within a certain radius of some location containing it.  Harvesting will continue until the unit is carrying the resource (but the unit will not automatically return to the base and drop off the resource).
- (DROPOFF): if near base, make unit drop off the resource (if any) it is carrying, otherwise do nothing.  In the first case, the action lasts until the resource has been dropped off and the unit is awaiting new orders.
- (BUILD BUILDING) means start building the given building at the unit's current location.
- (STOP) means stop doing the last given command.

Units not mentioned in the action continue or stop the previous command as per the description above.  Since the io-interface treats nil specially, the symbol 'noop may also be used instead.

All symbols in the above are exported by the stratagus-env package.  If using the io-interface on a stratagus-env, be sure to use the stratagus-env package, so that the strings you type in get interpreted correctly."))



;; after initializing, set value of state slot to be current game state
(defmethod initialize-instance :after ((e <stratagus-env>) &rest args)
  (setf *socket* (game-socket e))
  (set-state (get-game-state e) e)
  (set-pid (progn 
	     (format (game-socket e) "~a~%" *request-gameinfo-message*)
	     (force-output (game-socket e))
	     (gameinfo-player-id (read (game-socket e)))) e))


;; (set-world (generate-map e) e))
;; i commented this out temporarily because it caused an error in tactical-fc-env - BM

  


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; actions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defmethod do-action ((e <stratagus-env>) act &aux (s (game-socket e)))
  (when (eq act 'noop) (setf act nil))
  (check-type act list)

  ;; first, stop any one-step actions
  (awhile (pop (units-to-stop e))
    (format s "~a ~a ~a~%" *command-message* it *action-stop*))
	  
  (dolist (entry act)
    (dbind (id move-type &rest args) entry
      (force-format s "~a ~a ~a~%" *command-message* id 
		    (apply
		     (ecase move-type
		       (move #'get-move-action)
		       (move-towards #'get-move-towards-action)
		       (build #'get-build-action)
		       (dropoff #'get-dropoff-action)
		       (harvest #'get-harvest-action))
		     e id args))
      (let ((response (read s)))
	(assert (equal *ok-message* (symbol-name response)) ()
	  "Action ~a ~a for effector ~a failed.  Response was ~a." move-type args id response))))
  
  (send e *trans-message*)
  
  ;; receive and return new state
  (let* ((new-state (get-game-state e))
	 (r (compute-reward e (get-state e) act new-state)))
    (set-state new-state e)
    (values r new-state (at-terminal-state e))))


(defun get-harvest-action (e id res-type)
  (declare (ignore e id))
  (format nil "~a ~a" *action-harvest*
	  (ecase res-type
	    (gold 1)
	    (wood 2))))

(defun get-dropoff-action (e id)
  (declare (ignore e id))
  *action-dropoff*)

(defun get-build-action (e id building-type)
  (let ((current-loc (get-unit-loc id (get-state e))))
    (dbind (c r) current-loc
      (format nil "~a ~a ~a ~a"
	      *action-build* 
	      (ecase building-type
		(barracks *unit-human-barracks*))
	      c r))))

(defparameter *move-inc* 10)
(defconstant *action-delta* '((n 0 -1) (s 0 1) (e 1 0) (w -1 0)))

(defun get-move-action (e id dir)
  (push id (units-to-stop e))
  (let ((delta (mapping:evaluate *action-delta* dir))
	(current-loc (get-unit-loc id (get-state e))))
    (dbind (dc dr) delta
      (dbind (c r) current-loc
	(format nil "~a ~a ~a" *action-move* (+ (* *move-inc* dc) c) (+ (* *move-inc* dr) r))))))

(defun get-move-towards-action (e id dest)
  (push id (units-to-stop e))
  (dbind (c r) dest
    (format nil "~a ~a ~a" *action-move* c r)))
  

;; reset
;; e : <stratagus-env>
;;
;; reset state.
;; subclasses might want to define :after methods for this to randomize the initial state
(defmethod reset ((e <stratagus-env>) &optional (reset-to-nonterminal-state t) &aux (s (game-socket e)))
  (assert reset-to-nonterminal-state)
  (clear-input s)
  (send e *reset-message* t)
  (setf (units-to-stop e) nil)
  (set-state (get-game-state e) e))


(defun send (e m &optional check-ok)
  "send STRAT-ENV M &optional (CHECK-OK nil).  Send message M over the socket to the stratagus game and wait for, and return a response.  If CHECK-OK is t, then additionally verify that the response is *ok-message*.  Could hang if the connection is having issues and fails to respond."
  (send-socket (game-socket e) m check-ok))

(defun send-socket (s m &optional check-ok)
  "send-socket SOCKET M &optional (CHECK-OK nil).  Send message M over the socket to the stratagus game and wait for, and return a response.  If CHECK-OK is t, then additionally verify that the response is *ok-message*.  Could hang if the connection is having issues and fails to respond."
  (format s "~a~%" m)
  (force-output s)
  (let ((resp (read s)))
    (when check-ok
      (assert (equal (symbol-name resp) *ok-message*)))
    resp))
      






;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; internal operations
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;; get-game-state
;; e : <stratagus-env>
;;
;; receive the state from the stratagus game, parse and return it.
(defmethod get-game-state ((e <stratagus-env>) &aux (s (game-socket e)))
  
  ;; request state
  (format s "~a~%" *request-state-message*)
  (force-output s)
  ;; (format s "~a~%" *request-map-message*)
  ;; (force-output s)

  ;; flush all OKs
  (let ((a (read s)))
    (until (not (symbolp a))
      (assert (equal (symbol-name a) *ok-message*) () "Action for effector failed.  Response was ~a." a)
      (setf a (read s)))

    ;; receive and parse state. 
    (make-strat-es 
     :global (second a)
     :units (let* ((unit-pair-list (first a))
		   (max-id (reduce #'mymax unit-pair-list :key #'car)))
	      (assert (my> max-id '-infty) nil
		"Unit pair list ~a had max ~a" unit-pair-list max-id)
	      (let ((v (make-array (1+ max-id) :initial-element nil)))
		(dolist (p unit-pair-list v)
		  (dbind (id . us) p
		    (setf (unit-type us) (aref *unit-names* (unit-type us)))
		    ;; this might be unnecessary most of the time...
		  
		    (setf (aref v id) us))))))))


		     

  

(defgeneric compute-reward (e s a s-new)
  (:documentation "compute-reward (E <STRATAGUS-ENV>) (S strat-env-state) ACT (S-NEW strat-env-state).  Reward function of the environment.")
  (:method ((e <stratagus-env>) s a s-new)
	   (declare (ignore s a s-new))
	   0))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; old
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun old-prompt-for-choice (s choices e)
  (declare (ignore choices s e))
  
  (format t "~&Enter number of commands (or -1 to exit).")
  (let ((num-commands (read-from-string (read-line))))
    (when (eql num-commands -1)
      (error 'policy:choose-to-abort))
  
    (mapset 'list
	    #'(lambda (i)
		(format t "~&Command ~a.~&Unit to command? " i)
		(let ((id (read t)))
		  (format t "~&Command type? ")
		  (let* ((cmd-name (read t))
			 (arg-names (lookup-command cmd-name))
			 (args (mapcar #'(lambda (name)
					   (format t "~&Enter argument ~a. " name)
					   (read))
				       arg-names)))
		    (list* id cmd-name args))))
	    num-commands)))

(defun old-do-action (e act &aux (s nil))
  (declare (ignore fresh))
  (check-type act list)

  ;; send all the unit actions
  (dolist (entry act)
    (destructuring-bind (id move-type &rest args) entry
      (let ((a-num (action-num move-type)))
	(unless (equal a-num 0)
	  (format s "~a ~a ~a" *command-message* id a-num)
	  (dolist (a args)
	    (format s " ~a" a))
	  (terpri s)
	  (force-output s)
	  (let ((response (read s)))
	    (assert (equal *ok-message* (symbol-name response)) ()
	      "Action ~a ~a for effector ~a failed.  Response was ~a." move-type args id response))))))
  
  (send e *trans-message*)
  
  ;; receive and return new state
  (let* ((new-state (get-game-state e))
	 (r (compute-reward e (get-state e) act new-state)))
    (set-state new-state e)
    (values r new-state (at-terminal-state e))))

