(in-package maxq)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; generic operations on goal languages/goals
;; Goals are sets of states.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric empty-goal-description (goal-language)
  (:documentation "empty-goal-description GOAL-LANGUAGE.  Return a description, in this language, of the empty set."))

(defgeneric local-changes (goal-language desc)
  (:documentation "local-changes GOAL-LANGUAGE GOAL-DESC.  Return the set of descriptions obtained by making a local change to the current one."))

(defgeneric all-goals (goal-language)
  (:documentation "The set of goals that can be proposed for a new task.")
  (:method (goal-language)
	   (local-changes goal-language (empty-goal-description goal-language))))

(defgeneric goal-rel-vars (goal)
  (:documentation "Return list of variables relevant to this goal.")
  (:method ((g null)) nil)
  (:method ((g (eql t))) nil))


  

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; a simple goal language
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; goals of form VAR \in SET
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <var-val-set-goal> (<set>)
  ((var :accessor var :initarg :var)
   (val-set :accessor val-set :initarg :val-set)
   (vv-sname :accessor name :initarg :name)))

(defmacro make-var-val-set-goal (var val-set)
  "Macro make-var-val-set-goal VAR VAL-SET
VAR - a symbol (unevaluated) that is the name of a function
VAL-SET - evaluates to a set"
  `(make-instance '<var-val-set-goal> :var #',var :name ',var :val-set ,val-set))

(defmethod member? (x (s <var-val-set-goal>))
  (member? (funcall (var s) x) (val-set s)))

(defmethod goal-rel-vars ((g <var-val-set-goal>))
  (list (name g)))

(defmethod print-object ((s <var-val-set-goal>) str)
  (print-unreadable-object (s str :type t :identity nil)
    (format str "~a in ~a" (name s) (val-set s))))

(defmethod clone ((x <var-val-set-goal>)) x)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; goals of form VAR1=VAR2
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <var-var-goal> (<set>)
  ((var1 :accessor var1 :initarg :var1)
   (var2 :accessor var2 :initarg :var2)
   (name1 :accessor name1 :initarg :name1)
   (name2 :accessor name2 :initarg :name2)
   (test :accessor test :initform #'equalp)))

(defmacro make-var-var-goal (var1 var2)
  "Macro make-var-var-goal VAR1 VAR2.  VAR1 and VAR2 are unevaluated symbols.  See make-var-val-goal."
  `(make-instance '<var-var-goal> :var1 #',var1 :name1 ',var1 :var2 #',var2 :name2 ',var2))

(defmethod member? (x (s <var-var-goal>))
  (funcall (test s) (funcall (var1 s) x) (funcall (var2 s) x)))

(defmethod goal-rel-vars ((g <var-var-goal>))
  (list (name1 g) (name2 g)))

(defmethod print-object ((s <var-var-goal>) str)
  (print-unreadable-object (s str :type t :identity nil)
    (format str "~a = ~a" (name1 s) (name2 s))))

(defmethod set-eq ((s <var-var-goal>) (s2 t))
  nil)

(defmethod set-eq ((s <var-var-goal>) (s2 <var-var-goal>))
  (and (eq (name1 s) (name1 s2))
       (eq (name2 s) (name2 s2))
       (eq (test s) (test s2))))

(defmethod clone ((x <var-var-goal>)) x)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; The overall goal language
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <var-val-goal-language> ()
  ((var-specs :accessor var-specs :initarg :var-specs))
  (:documentation "Class <var-val-goal-language>.

Goal language for state spaces in which state consists of values of a fixed set of variables.  There are three types of goal descriptions:
1) VAR member of VAL-SET
2) VAR1=VAR2
3) the empty set (represented by nil)
4) The universal set t
For now, the description space is completely connected: any goal can be changed to any other.


Initargs
:var-specs - association list from var names to  objects of type var-spec"))

(defmethod empty-goal-description ((s <var-val-goal-language>))
  nil)

(defmethod local-changes ((s <var-val-goal-language>) goal &aux (specs (var-specs s)))
  (declare (ignore goal))
  (let ((vars (mapping:domain specs)))
    (disjoint-union

     (list nil t)
     
     ;; set of all var-var goals
     (ndlet ((var-set (subsets-of-size vars 2)))
       (dbind (name1 name2) var-set
	 (make-instance '<var-var-goal>
	   :var1 (vs-function (evaluate specs name1)) :var2 (vs-function (evaluate specs name2)) :name1 name1 :name2 name2)))
     
     ;; set of all var-val goals
     (ndunion ((var vars))
       (let* ((spec (evaluate specs var))
	      (fn (vs-function spec)))
	 (ndlet ((x (vs-allowed-goals spec)))
	   (make-instance '<var-val-set-goal> :name var :var fn :val-set x)))))))
	 

