;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Contains version of greedy hierarchy learning algorithm that
;; 1. Uses the true objective function when adding new tasks
;; 2. Incorporates horizon-length, solution cost in objective
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package maxq)




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Special variables
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Vars storing info about the current top-level
;; problem instance that doesn't change during a
;; particular call to greedy-hierarchy-learning
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *var-specs* nil "Association list from variable name to var-spec")
(defvar *goal-language* nil "Language of allowed goal and initiation sets")
(defvar *primitive-actions* nil "List of primitive actions in environment")
(defvar *trajectories* nil "Vector of observed trajectories")
(defvar *env-table* nil "Maps each environment to an integer id, for readability when debugging.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Vars that store top-level state of
;; greedy-hierarchy-learning
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *h* nil "The current hierarchy.")
(defvar *all-parses* nil "Vector of parses, one for each trajectory")
(defvar *task-parses* nil "Association list from task name to list of parses based on that task")
(defvar *task-rel-vars* nil "Association list from task name to relevant vars")
(defvar *task-horizons* nil "Association list from task name to average horizon of task")
(defvar *task-policies* nil "Association list from task name to policy for that task")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Other parameters that affect the algorithm's 
;; behaviour
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *goal-occurrence-threshold* .5 "A goal is considered if it occurs on this fraction of trajectories.")
(defparameter *init-occurrence-threshold* .5 "An initiation set is considered for a given goal if it occurs at some point before the goal this fraction of the time.")
(defparameter *var-cost* .1 "Determines how willing the relevance checker is to hypothesize that the value function depends on more variables.")
(defparameter *goal-values* '(1 10) "When proposing a task, we consider these as possible values for the goal states of the task.")
(defparameter *task-learning-num-steps* 50000 "Number of steps used when learning a task.")
(defparameter *task-learning-max-episode-length* 200 "Max number of steps per episode when learning a task") 
(defparameter *task-learning-rate* .5 "Learning rate when learning subtasks")
(defparameter *near-optimality-threshold* .1 "How close to optimal an action must be to be included in the greedy parse (see is-near-optimal)")



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Types
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defstruct (task-addition (:conc-name ta-) (:constructor make-task-addition (name task parents)) (:type vector))
  "task-addition represents a proposed addition to the hierarchy.  Create using make-task-addition NAME TASK PARENTS where

NAME is the name of the new task
TASK is of type extended-task 
PARENTS is a list of parent task names."
  name
  task 
  parents)

(defclass <extended-task> (<task>)
  ((goal-value :type real :accessor goal-value :initarg :goal-value :initform 0)
   (policy :accessor policy :initarg :policy))
  (:documentation "Class <extended-task> (<task>)

Extends maxq tasks with additional info that is used when learning a hierarchy.

Create using make-instance.  All the initargs from <task> apply, as well as
:goal-value - real number that represents value of exiting task at a goal state (value at nongoal states is 0)
:policy - policy being followed in this task."))

(defmethod print-object ((task <task>) str)
  (format str "<<Task> Children : ~a  Init-set : ~a  Goal : ~a  Goal-value : ~a>>" 
	  (mapcar #'car (children task)) (initiation-set task) (goal task) (goal-value task)))




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; top-level loop
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun greedy-hierarchy-learning (*goal-language* *trajectories* *primitive-actions* var-specs)
  "Todo document.

Returns 1) A <maxq-hierarchy> H 2) Sequence of valid parses of each trajectory w.r.t H"
  
  ;; Initialize top-level state by binding the relevant special variables
  (reset-env-table)
  (let ((*h* (initial-hierarchy))
	(*all-parses* (map 'vector #'init-parse *trajectories*))
	(*var-specs* (preprocess-var-specs var-specs)))
    (do ((*task-parses* (list (cons 'top *all-parses*))) ;; association list from task name to list of parses based on that task
	 (*task-rel-vars* (list (cons 'top (task-rel-vars *all-parses*)))) ;; association list from task name to relevant vars
	 (*task-policies* nil) ;; Association list from subtasks to policies (needed when learning what's relevant to the decomposed Q-function of the parent of that subtask)
	 (*task-horizons* (list (cons 'top (task-horizon *all-parses*)))))
	  
	;; Loop forever (for now)
	(nil (values (copy-hierarchy *h*) *all-parses*))

      ;; At each iteration,
      ;; 1. Get set of candidates
      
      (let ((candidates (candidate-tasks)))
	(debug-prompt 0 "Top-level loop of greedy learning~:@_Hierarchy is ~a~:@_Parses are ~a~&Task relevant vars are ~a~&Task horizons are ~a~&New candidate tasks are ~/set:pprint-set/"
		      *h* *all-parses* *task-rel-vars* (mapcar #'(lambda (entry) (cons (car entry) (round (cdr entry)))) *task-horizons*) candidates)
	
	;; 2. Pick the one that minimizes the objective function
	(let ((best-addition (minimizing-element candidates #'(lambda (addition) (evaluate-addition addition)))))
	  (debug-prompt 0 "Best addition : ~a" best-addition)

	  ;; 3. incorporate it into the top-level state
	  (incorporate-addition best-addition))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 0. Initialization helpers
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun init-parse (trajectory &aux (actions (trajectory-actions trajectory)) (n (num-transitions trajectory)))
  "Initial parse for a trajectory consists of the top task directly doing all the primitive actions"
  (make-parse-entry 
   trajectory 'top 0 n
   (mapset 'list #'(lambda (i) (make-parse-entry trajectory (elt actions i) i (1+ i) nil)) n)))


(defun initial-hierarchy ()
  "initial-hierarchy 
Return a fresh initial hierarchy used by the greedy search procedure.
Affected by: *primitive-actions*"
  (make-instance '<maxq-hierarchy>
    :high-level-actions (p2alist 'top (make-instance '<task> :children *primitive-actions* :goal nil))
    :primitive-actions *primitive-actions*))

(defun preprocess-var-specs (specs)
  "Modify var specs to deal with the fact we won't get states, but things of the form (Env . state)"
  (assert (not (member 'env-identifier specs :key #'car)) nil
    "Var specs ~a must not contain a variable named 'maxq::env-identifier")
  `((env-identifier . ,(make-var-spec #'car 'not-applicable))
    ,@(mapcar 
       #'(lambda (entry)
	   (cons (car entry)
		 (let ((spec (copy-var-spec (cdr entry))))
		   (_f compose (vs-function spec) #'cdr)
		   spec)))
       specs)))

(defun reset-env-table ()
  "Maintain a table of ids for environments.  Affected by: *trajectories*"
  (setf *env-table* (make-hash-table))
  (do-elements (tr *trajectories*)
    (let ((e (trajectory-env tr)))
      (set-unless-key-exists *env-table* e (hash-table-count *env-table*)))))
	  
	

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 1. generating set of candidate tasks
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun candidate-tasks ()
  "candidate-tasks
Return the set of candidate task additions (of type task-addition) given the current hierarchy and set of parses."
  (let ((spans (contiguous-spans)))
    
    ;; How to nondeterministically pick a task
    ;; First, pick a parent
    (ndunion ((parent (hla-names *h*)))
      (let* ((parent-spans (evaluate spans parent))
	     (num-spans (length parent-spans)))

	(unless (zerop num-spans)

	  ;; Pick a goal
	  (ndunion ((goal (all-goals *goal-language*)))

	    ;; When the goal occurs reasonably often...
	    (when (and (not (typep goal 'boolean))
		       (< *goal-occurrence-threshold*
			  (/ (sum-over
			      parent-spans
			      #'(lambda (span)
				  (indicator
				   (dbind (traj start end) span
				     (find-if 
				      #'(lambda (s) (member? s goal))
				      (trajectory-states traj)
				      :start (1+ start) :end (1+ end))))))
			     num-spans)))
	      
	      ;; Pick an initiation set
	      (ndunion ((init-set (all-goals *goal-language*)))
		
		;; When the initiation set often occurs before the goal, and is non-
		(when (< *init-occurrence-threshold*
			 (/ (sum-over parent-spans
				      #'(lambda (span) (indicator (contains-init-goal-pattern span init-set goal))))
			    num-spans))
		  
		  ;; ... return a bunch of instances of the task (for varying goal rewards)
		  (ndlet ((goal-value *goal-values*))
		    (make-task-addition (gensym "TASK-") (make-instance '<extended-task> :children *primitive-actions* :goal goal :init-set init-set :goal-value goal-value) (list parent))))))))))))


(defun contains-init-goal-pattern (span init goal)
  "contains-init-goal-pattern SPAN INIT GOAL
Does SPAN have an occurrence of INIT before an occurrence of GOAL?"
  (dbind (traj start end) span
    (let* ((states (trajectory-states traj))
	   (i (position-if #'(lambda (s) (and (member? s init) (not (member? s goal)))) states :start start :end end)))
      (aand i (position-if #'(lambda (s) (member? s goal)) states :start (1+ start) :end (1+ end) :from-end t) (< i it)))))
	    
  


(defun contiguous-spans ()
  "contiguous-spans 
Return an association list, where the keys are task names, and the values are list of triples of the form (TRAJECTORY S E), each of which denotes that in TRAJECTORY, states S through E inclusive occur in an instance of the task, and none of them is contained in any lower-level task "

  (mapcar 
   
   ;; 1. Function that finds contiguous spans for a particular task...
   #'(lambda (entry)
       (dsbind (name . parses) entry
	 (cons name
	       
	       ;; To find contiguous spans for a particular task, append together
	       ;; for each parse for that task, list of all contiguous spans in that parse
	       (mapcans
		#'(lambda (parse &aux (traj (pe-trajectory parse)))
		    
		    ;; To find contiguous spans in a particular parse, collect together
		    ;; all contiguous blocks of primitive subtasks, 
		    (let ((children (pe-children parse)))
		      (mapcan
		       #'(lambda (b)
			   (dbind (task-type start end) b
			     (when (eq task-type 'primitive)
			       (list (list traj start end)))))
		       (contiguous-block-states children :key #'(lambda (c) (task-type (pe-action c) *h*))))))
		parses))))
   
   ;; 2. ... is mapped to set of tasks
   *task-parses*))
	       

(defun contiguous-block-states (parses &rest args)
  (mapcar 
   #'(lambda (b)
       (list (first b)
	     (pe-start (nth (second b) parses))
	     (pe-end (nth (1- (third b)) parses))))
   (apply #'contiguous-blocks parses args)))
  




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 2. Objective function
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun evaluate-addition (addition)
  (debug-print 1 "Evaluating addition ~a" addition)
  (with-debug-indent
      (let* ((name (ta-name addition))
	     (task (ta-task addition))
	     (q (task-q-function task))
	     (complexity-change 0)
	     (new-task-parses
	   
	      ;; Go through the parents of this new task, and figure out how it would affect
	      ;; each of their parses, and therefore complexity.  Also collect (the mapcan) all parses 
	      ;; corresponding to this new task
	      (mapcan
	       #'(lambda (parent)
		   (let* ((old-parses (evaluate *task-parses* parent))
			  (modified-parses (mapset 'list #'(lambda (parse) (greedy-parse name parse task q t)) old-parses))
			  (old-complexity (evaluate-task-parses old-parses))
			  (new-complexity (evaluate-task-parses modified-parses)))
		     (debug-print 2 "Evaluating effect on parent ~a~&Modified parses are ~a~&Complexity of ~a went from ~:d to ~:d" 
				  parent modified-parses parent (round old-complexity) (round new-complexity))
		  
		     ;; Side effect: keep track of change in complexity
		     (incf complexity-change (- new-complexity old-complexity))
		  
		     ;; Return the parses corresponding to this new task
		     (child-parses-corresponding-to name modified-parses)
		     ))
	       (ta-parents addition))))

	;; Now use the new-task-parses to figure out the complexity of the new task and return the overall complexity change
	(incf complexity-change (evaluate-task-parses new-task-parses))
	(debug-print 1 "Overall complexity change : ~:d" (round complexity-change))
	(debug-prompt 2)
	complexity-change)))

(defun evaluate-task-parses (parses)
  "evaluate-task-parses PARSES  
Assuming PARSES is the set of all parses corresponding to some task, return AHS^2 where
A is the number of different subtasks called
H is the average length of the task (in terms of #subtasks)
S is the complexity of the value function, defined as the square of the total size of tabular representations of the Vr and Vc functions."

  (debug-print 2 "Evaluating task~:[ ~a~;~]" (is-empty parses) (unless (is-empty parses) (pe-action (elt parses 0))))
  (with-debug-indent
      (let ((vi-rel-vars (task-rel-vars parses))
	    (horizon (task-horizon parses))
	    (children (task-children parses)))
	(let ((complexity (task-complexity vi-rel-vars horizon children)))
	  (debug-print 2 "Children : ~a  Horizon : ~a  Relevant vars : ~a~&Estimated complexity: ~:d"
		       children horizon vi-rel-vars (round complexity))
	  complexity))))

(defun task-complexity (rel-vars horizon children)
  (* horizon (size children)
     (expt (reduce #'* rel-vars :key #'(lambda (v) (var-expected-size v *var-specs*))) 2)))
      
(defun get-env-number (e)
  (evaluate *env-table* e))

(defun task-horizon (parses)
  (iunless 
   (is-empty parses)
   (float
    (/ (reduce #'+ parses :key (fn (length pe-children)))
       (length parses)))))

(defun task-children (parses)
  (let ((children nil))
    (do-elements (p parses children)
      (dolist (c (pe-children p))
	(pushnew (pe-action c) children :test #'equal)))))


	   
    


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 3. Incorporating task additions 
;; This is the only place where the top-level state
;; is modified
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun incorporate-addition (addition)
  "Incorporate this new addition into the top-level state of the hierarchy learner, stored in *h*, *all-parses*, *task-parses*, *task-rel-vars*, *task-horizons*, and *task-policies*"
  (let ((name (ta-name addition))
	(task (ta-task addition))
	(parents (ta-parents addition)))
    
    ;; Reparse parent parses.  This works because 
    ;; 1) greedy parse modifies the parse it is given when the fourth arg is nil 
    ;; 2) *task-parses* contains pointers into the structures contained in *all-parses*
    (dolist (par parents)
      (do-elements (p (evaluate *task-parses* par))
	(greedy-parse name p task 'todo nil)))
    
    ;; Update task-parses to include the new task
    (push (cons name (mapcan #'(lambda (entry) (child-parses-corresponding-to name (cdr entry))) *task-parses*)) *task-parses*)
    
    ;; Add new task name to the lists and hierarchy
    (push (cons name nil) *task-rel-vars*)
    (push (cons name nil) *task-horizons*)
    (push (cons name nil) *task-policies*)
    (add-task task parents *h* name)
    
    ;; Update horizons, children, rel vars
    (dolist (task-name (cons name parents))
      (let ((parses (evaluate *task-parses* task-name))
	    (rel-vars-entry (assoc task-name *task-rel-vars* :test #'equal))
	    (horizon-entry (assoc task-name *task-horizons* :test #'equal))
	    (task (lookup-hla task-name *h*)))
	(setf (cdr rel-vars-entry) (task-rel-vars parses)
	      (cdr horizon-entry) (task-horizon parses)
	      (children task) (task-children parses))))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; A. relevant variable determination.  Used by 2)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; relevant variables for vr
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun vr-rel-vars (parses)
  (l1-binned-feature-select 
   (generate-vr-examples parses)
   *var-specs* :var-cost *var-cost*))

(defun generate-vr-examples (parses)
  (mapcans #'generate-vr-examples-for-parse parses))

(defun generate-vr-examples-for-parse (p)
  (mapcar #'generate-single-vr-example (pe-children p)))

(defun generate-single-vr-example (child-parse)
  (let ((traj (pe-trajectory child-parse))
	(pos (pe-start child-parse)))
    (cons (cons (trajectory-env traj) (aref (trajectory-states traj) pos)) (aref (trajectory-rewards traj) pos))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; relevant vars for vc
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun vc-rel-vars (parses)
  (l1-binned-feature-select
   (generate-vc-examples parses)
   *var-specs* :var-cost *var-cost*))


(defun generate-vc-examples (parses)
  (let ((examples (mapcans #'generate-vc-examples-for-parse parses)))
    examples))


(defun generate-vc-examples-for-parse (p)
  (let ((states (trajectory-states (pe-trajectory p)))
	(rewards (trajectory-rewards (pe-trajectory p)))
	(env (trajectory-env (pe-trajectory p)))
	(total-reward 0.0)
	(examples nil))
    (do ((i (1- (pe-end p)) (decf i)))
	((< i (pe-start p)) examples)
      (push (cons (cons env (aref states i)) total-reward) examples)
      (incf total-reward (aref rewards i)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; relevant vars for vi (= vr + vc)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun task-rel-vars (parses)
  (l1-binned-feature-select (generate-vi-examples parses) *var-specs* :var-cost *var-cost*))

(defun generate-vi-examples (parses)
  (mapcans #'generate-vi-examples-for-parse parses))

(defun generate-vi-examples-for-parse (p &aux (traj (pe-trajectory p)))
  (let ((states (trajectory-states traj))
	(rewards (trajectory-rewards traj))
	(env (trajectory-env traj))
	(total-reward 0.0)
	(examples nil))
    (do ((i (1- (pe-end p)) (decf i)))
	((< i (pe-start p)) examples)
      (incf total-reward (aref rewards i))
      (push (cons (cons env (aref states i)) total-reward) examples))))
      


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; B. Greedy parsing.  Used by 2 and 3.
;; In the second case, it modifies the top-level state.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *parse-debug-level* nil)
  

(defun greedy-parse (name parse task q copy? &aux (traj (pe-trajectory parse)))
  (with-debug-indent 
    (let* ((unexplained-blocks (remove 'high-level (contiguous-block-states (pe-children parse) :key #'(lambda (c) (task-type (pe-action c) *h*))) :key #'car))
	     ;; Blocks of primitive actions unexplained by higher-level actions
	     (new-parse (if copy? (copy-parse-entry parse) parse))
	     (explained-blocks
	      (mapcan
	       
	       ;; Function that takes in an unexplained block and returns list of explained blocks
	       #'(lambda (b)
		   (dbind (s e) (cdr b)
		     (mapcan
		      ;; 1. Function that looks for longest explained block within candidate subblock (if one exists)
		      #'(lambda (b2)
			  (dbind (s2 e2) b2
			    (awhen (get-explained-block traj s2 e2 q)
			      (list (cons it e2)))))
		      ;; 2. Applied to blocks of states in init \ goal which are followed by block where goal is true
		      (applicable-blocks (trajectory-states traj) (goal task) (initiation-set task) :start s :end e))))
	       unexplained-blocks))

	     (actions (trajectory-actions traj)))
	     
	(debug-print *parse-debug-level* "Greedy parse of ~a~&Unexplained blocks are ~a~&Explained blocks are ~a" parse unexplained-blocks explained-blocks)
	(setf (pe-children new-parse)
	  
	  ;; Go through the children.
	  (mapcan
	   #'(lambda (c)
	       (if (eq 'high-level (task-type (pe-action c) *h*))
		   
		   ;; Leave high-level children as is
		   (list c)
		 (let ((i (pe-start c)))
		   (aif (find-if #'(lambda (b) (between2 i (car b) (cdr b))) explained-blocks)
			
		       ;; Primitive explained children are replaced by an instance of the task that parses the corresponding explained block
		       (dsbind (s . e) it
			 (when (= i s)
			   (list (make-parse-entry  
				  traj name s e 
				  (mapset 'list #'(lambda (j) (make-parse-entry traj (aref actions j) j (1+ j) nil)) (int-range s e))))
				    
			   ))
			
		     ;; Leave primitive children that are not explained as is
		     (list c)))))
	   (pe-children new-parse)))

	;; return the modified parse
	new-parse)))



(defun get-explained-block (traj s e q)
  "get-explained-block TRAJECTORY START END Q-FUNCTION
For each i, the ith action (the action taken at the ith state) is said to be near-optimal if its Q-value is within *near-optimality-threshold* of the best action for the given Q-function.  If there exist i, such that start <= i < end, and for each i <= j < end, the jth action is near optimal, then return the minimal such i.  Otherwise, return nil."
  (assert (> e s))
  (with-debug-indent
      (let ((i (1- e))
	    (states (trajectory-states traj))
	    (actions (trajectory-actions traj))
	    (best nil))
	(debug-print *parse-debug-level* "Looking for explained block from ~a~&with actions ~a" (subseq states s e) (subseq actions s e))
	(debug-print 
	 *parse-debug-level* "Return value is ~a"
	 (loop
	   (if (< i s)
	       (return s)
	     (unless (is-near-optimal (aref actions i) (aref states i) q)
	       (return best)))
	   (setf best i)
	   (decf i))))))


(defun is-near-optimal (a s q)
  "is-near-optimal A S Q.
Let V be the best Q-value at S, and V-R the worst.  A is considered near-optimal if (V-Q(S,a))/R < *near-optimality-threshold*.  E.g., suppose the Q-values range from 10 to 20 at a given state.  A threshold of .1 would then mean that A has to have Q-value at least 19."
  (debug-print 
   *parse-debug-level* "~a is ~*~:[not ~;~]near-optimal in ~2:*~a?" 
   a s 
   (flet ((action-value (act) (q-fn:evaluate q s act)))
     (let* ((choices (q-fn:choices q s))
	    (qmax (reduce-set #'mymax choices :key #'action-value))
	    (qmin (reduce-set #'mymin choices :key #'action-value)))
       (debug-print *parse-debug-level* "Choices are ~a~&Best is ~a (~a)~&Worst is ~a~&Current is ~a" choices qmax (q-fn:best-choice q s) qmin (action-value a))
       (> (action-value a) (- qmax (* *near-optimality-threshold* (- qmax qmin))))))))
  


(defun child-parses-corresponding-to (name parses)
  (mapcans
   #'(lambda (parse)
       (remove-if #'(lambda (c) (not (equal name (pe-action c)))) (pe-children parse)))
   parses))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; C. Learning task q-function.  Used by 2
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun task-q-function (task &optional new-child-task)
  ;; Eventually should memoize this
  
  (let ((e (trajectory-env (sfirst *trajectories*))) ;; For now assuming we have a single, resettable environment
	(init-states (candidate-init-states task))
	(children (children task))
	(prim nil)
	(subtasks nil))
    
    ;; Go through the children and figure out if each one is primitive, an existing high-level task, or the new one
    (do-elements (c children)
      (if (and (consp c) (member (car c) *primitive-actions* :test #'equal))
	  (push (car c) prim)
	(let ((child-task
	       (handler-case
		   (lookup-hla c *h*)
		 (unknown-hla () new-child-task))))
	  (push (list c (let ((goal (goal child-task)))
			  (filter ':implicit (initiation-set child-task) #'(lambda (x) (not (member? x goal)))))
		      (policy child-task)) subtasks))))
    
    (debug-print *parse-debug-level* "Init states : ~a~&Calling subtask learner for ~a steps" init-states *task-learning-num-steps*)

    ;; Call the task learner
    (task-learner e init-states (initiation-set task) (goal task) (goal-value task) prim subtasks *task-learning-num-steps* 
		  :max-episode-length *task-learning-max-episode-length* :learning-rate *task-learning-rate*)))
  

(defun candidate-init-states (task)
  "candidate-init-states TASK
Look through the trajectories for states at which we would potentially care about Q-values for this task."
  (let ((init-states (make-array 0 :adjustable t :fill-pointer 0))
	(goal (goal task))
	(init (initiation-set task)))

    (do-elements (tr *trajectories* init-states)
      (let* ((states (trajectory-states tr))
	     (blocks (applicable-blocks states goal init)))
	
	(dolist (b blocks)
	  (for-loop (i (first b) (second b))
	    (let ((state (aref states i)))
	      (unless (find state init-states :test #'equalp)
		(vector-push-extend state init-states)))))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; General utility functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun applicable-blocks (states goal init &key (start 0) (end (length states)))
  "applicable-blocks STATES GOAL INIT &key (START 0) (END (length STATES))
Returns list of elements of form (S E), each of which means that the states numbered S through E-1 all belong to INIT\GOAL and state E belongs to GOAL."

  (let ((blocks (contiguous-blocks states :test #'equal :start start :end end :key #'(lambda (s) (cons (member? s goal) (member? s init)))))
	(l nil))
    (mapc #'(lambda (block1 block2)
	      ;; Pick out a block satisfying (init and not(goal)) followed by a block satisfying goal
	      (when (and (caar block2) (not (caar block1)) (cdar block2))
		(push (cdr block1) l)))
	  blocks (rest blocks))
    l))

	    

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Obsolete code
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun old-greedy-parse (name parse task copy? &aux (traj (pe-trajectory parse)))
  "greedy-parse NAME PARSE TASK
Try to parse long unexplained blocks in PARSE (i.e., that are currently just parsed as a sequence of primitive actions) using TASK. For now, in each unexplained block, we look greedily for longest sequences of the form FF*T w.r.t the goal and TT*{F|T} w.r.t the initiation set"

  ;; TODO clean this up
  (with-debug-indent

      (let ((unexplained-blocks (remove 'high-level (contiguous-block-states (pe-children parse) :key #'(lambda (c) (task-type (pe-action c) *h*))) :key #'car))
	    (new-parse (if copy? (copy-parse-entry parse) parse))
	    (goal (goal task))
	    (init (initiation-set task))
	    (states (trajectory-states traj))
	    (actions (trajectory-actions traj)))
	(let ((explained-blocks 
	       (mapcan
		#'(lambda (b)
		    ;; For each unexplained block, figure out set of subblocks explained by task
		    (dbind (s e) (cdr b)
		      ;; Classify the states into blocks w.r.t whether they satisfy the goal and initiation conditions
		      (let ((goal-blocks (contiguous-blocks states :test #'equal :key #'(lambda (s) (cons (member? s goal) (member? s init))) :start s :end (1+ e))))
			(mapcan 
			 #'(lambda (b1 b2)
			     (let ((t1 (first b1))
				   (t2 (first b2))
				   (s1 (second b1))
				   (s2 (second b2)))
			   
			       ;; An explained block is whenever you have a (nil . t) followed by a (t . *) w.r.t the goal and initiation
			       (when (and (equal t1 '(nil . t)) (equal (car t2) t))
				 (list (cons s1 s2)))))
			 goal-blocks (cdr goal-blocks)))))
		unexplained-blocks)))
	     

	
	  (debug-print *parse-debug-level* "Greedy parse of ~a~&Unexplained blocks are ~a~&Explained blocks are ~a" parse unexplained-blocks explained-blocks)
	  (setf (pe-children new-parse)
	  
	    ;; Go through the children.
	    (mapcan
	     #'(lambda (c)
		 (if (eq 'high-level (task-type (pe-action c) *h*))
		   
		     ;; Leave high-level children as is
		     (list c)
		   (let ((i (pe-start c)))
		     (aif (find-if #'(lambda (b) (between2 i (car b) (cdr b))) explained-blocks)
			
			  ;; Primitive explained children are replaced by an instance of the task that parses the corresponding explained block
			  (dsbind (s . e) it
			    (when (= i s)
			      (list (make-parse-entry  
				     traj name s e 
				     (mapset 'list #'(lambda (j) (make-parse-entry traj (aref actions j) j (1+ j) nil)) (int-range s e))))
				    
			      ))
			
			  ;; Leave primitive children that are not explained as is
			  (list c)))))
	     (pe-children new-parse)))

	  ;; return the modified parse
	  new-parse))))
