(defpackage maxq
  (:documentation "Package for code related to MAXQ hierarchies and specifically, for learning their structure.

Types
-----
<maxq-hierarchy>
task

Constructors
------------
make-maxq


Operations
----------
best-one-level-hierarchy
parse-trajectory 
pprint-maxq
pprint-parse

Parameters
----------
*task-policy-determinism-threshold*
*max-num-contexts*
*typical-neighbourhood-size*
*typical-cost-range*


Other symbols
-------------
top (top-level task of a hierarchy)

")
  (:use cl
	env-user
	create-env
	set
	local-search
	policy
	alisp-prog
	prob
	graph
	utils)
  (:export
   <maxq-hierarchy>
   task
   make-maxq
   best-one-level-hierarchy
   parse-trajectory
   pprint-maxq
   pprint-parse
   *task-policy-determinism-threshold*
   *max-num-contexts*
   *typical-neighbourhood-size*
   *typical-cost-range*
   top)
  (:import-from 
   mapping
   evaluate
   domain)
  (:import-from 
   rl-obs
   inform-env-step
   inform-start-episode
   knowledge-state
   <learning-algorithm>)
  )

(in-package maxq)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; type defs, basic accessors
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <maxq-hierarchy> (<alisp-program>)
  ((high-level-actions :accessor high-level-actions :initarg :high-level-actions :type list)
   (primitive-actions :accessor primitive-actions :initarg :primitive-actions :type [set])
   (descendants :accessor descendants :accessor descendants)
   (goal-stack :accessor goal-stack :type list))
  (:documentation 
   "Class <maxq-hierarchy> (<alisp-program>)

Initargs
:high-level-actions - an association list from names, which are symbols or lists, to objects of type task.  The high-level-actions are assumed to be consistent, i.e. they form a DAG (which is ordered so that the lowest level ones appear first in the list), there is a task called 'top, and all child tasks are either other high-level actions, or primitive.  
:primitive-actions - set of primitive actions"))


(defclass <task> ()
  ((initiation-set :accessor initiation-set)
   (goal :accessor goal)
   (children :accessor children))
  (:documentation "Class <task>

Create using make-instance with initargs
:goal - a set, or a membership predicate
:init-set - a set or membership predicate.  Can also be left out, in which case the complement of the goal is used.
:children - list of names of child tasks/actions"))

(defmethod initialize-instance :after ((task <task>) &rest args &key goal children (init-set nil init-supp))
  (declare (ignore args))
  (let ((goal-set
	 (etypecase goal
	   ([set] goal)
	   (function (filter ':implicit t #'(lambda (x) (funcall goal x)))))))
    
    (setf (goal task) goal-set
	  
	  (initiation-set task)
	  (if init-supp
	      (etypecase init-set
		([set] init-set)
		(function (filter ':implicit t #'(lambda (x) (funcall init-set x)))))
	    (complement-set goal-set))
	  
	  (children task) 
	  ;; This number we're consing with is the 'child cost'.  Not really used anymore.
	  (mapset 'list #'(lambda (c)
			    (if (and (consp c) (numberp (cdr c)))
				c
			      (cons c 0)))
		  children))))


(defmethod initialize-instance :after ((h <maxq-hierarchy>) &rest args)
  (declare (ignore args))
  (assert (assoc 'top (high-level-actions h)) nil "Hierarchy must contain a task called 'top")
  (sort-hlas h))

(defmacro make-maxq (high-level primitive)
  "Macro make-maxq HIGH-LEVEL PRIMITIVE

PRIMITIVE evaluates to set of primitive actions
HIGH-LEVEL (unevaluated) is list of elements of the form (NAME CHILDREN GOAL &optional (INITIATION-SET (complement GOAL)))
where NAME (unevaluated) is a symbol or list of atoms, and the remaining arguments are evaluated.

Expands to code that returns a new <maxq-hierarchy>."

  `(make-instance '<maxq-hierarchy>
     :primitive-actions ,primitive
     :high-level-actions
     (list ,@(mapcar #'(lambda (desc)
			 (dbind (name &rest args) desc
			   `(cons ',name
				  (make-instance '<task> 
				    :children ,(first args)
				    :goal ,(second args)
				    ,@(awhen (third args)
					`(:init-set ,it))))))
		     high-level))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Partial program that implements MAXQ semantics
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod start ((h <maxq-hierarchy>))
  (setf (goal-stack h) nil)
  (run-task h 'top))

(defun run-task (h name)
  (ecase (task-type name h)
    
    (high-level
     (flet ((name () (intern-compound-symbol (etypecase name (symbol name) (list (first name))) "-CHOICE")))
       ;; TODO stop doing string concatenation every time a task is called
       
       (let ((task (lookup-hla name h)))
		  
	 (push (goal task) (goal-stack h))
	 (until (some-goal-is-satisfied h)
	   (let ((s (env-state)))
	     (flet ((task-is-applicable (child)
		      ;; a task is applicable so long as the current state is not one of its goal states
		      (let ((c (car child)))
		      (not (and (eq (task-type c h) 'high-level)
				(member? s (goal (lookup-hla c h)))
				)))))
			
	       ;; choose an applicable subtask
	       (with-choice (name) 
		 (c (filter ':list (children task) #'task-is-applicable))
		 (run-task h (car c))))))
	 (pop (goal-stack h)))))
	      
    (primitive
     (action 'action name))))
	
(defun some-goal-is-satisfied (h)
  (let ((s (env-state)))
    (some #'(lambda (g) (member? s g)) (goal-stack h))))

    

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Other operations on MAXQ hierarchies
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun copy-hierarchy (h)
  "copy-hierarchy H.  Make an identical hierarchy to H that shares no structure with it except the set of primitive actions."
  (make-instance '<maxq-hierarchy>
    :primitive-actions (primitive-actions h)
    :high-level-actions (mapcar #'(lambda (hla-pair) 
				    (dsbind (name . task) hla-pair 
				      (cons name (make-instance '<task>
						   :children (clone (children task)) 
						   :goal (clone (goal task))
						   :init-set (clone (initiation-set task))))))
				(high-level-actions h))))

(defun add-task (task parents h &optional (name nil))
  "add-task TASK PARENTS H &optional (NAME).  Adds new task with the given parents.  If a name is not given, a gensym is used.  Destructively modifies H and returns it."
  (orf name (gensym "MAXQ-TASK-"))
  (push (cons name task) (high-level-actions h))
  (let ((hlas (high-level-actions h)))
    (do-elements (p parents h)
      (let ((parent-task (evaluate hlas p)))
	(push (cons name 0) (children parent-task))))))

(defun hla-names (h)
  "Return list of names of HLAs."
  ;; NOTE: Many functions rely on the returned list being safely modifiable, i.e., that it consists of cons cells that are
  ;; freshly created within this function
  (mapcar #'car (high-level-actions h)))

(defun non-ancestors (h name &aux (desc (descendants h)))
  (assert (member name (high-level-actions h) :key #'car :test #'equal) ()
    "HLA ~a not a member of ~a" name (hla-names h))
  (let ((l nil))
    (dolist (d desc l)
      (dsbind (dname . descendants) d
	(unless (or (equal name dname) (member name descendants :test #'equal))
	  (push name l))))))

(defun parents (h hla)
  (let ((l nil))
    (mapping:do-entries (name task (high-level-actions h) l)
      (when (member hla (children task) :test #'equal :key #'car)
	(push name l)))))
			
	    

(defun sort-hlas (h &aux (hlas (high-level-actions h)) (n (length hlas)))
  "This should be called whenever the HLA graph changes (TODO make this happen automatically)."
  (let ((parents (make-array n :initial-element nil)))
    (do-elements (hla hlas nil i)
      (do-elements (c (children (cdr hla)))
	(let ((pos (position (car c) hlas :key #'car :test #'equal)))
	  (awhen pos (adjoinf (aref parents it) i)))))
    (let ((m (descendant-matrix parents))
	  (order (nreverse (topological-sort parents))))
      (setf (descendants h)
	(let* ((names (hla-names h))
	       (l (mapcar #'list names)))
	  (do-elements (h1 names l i)
	    (do-elements (descs l nil j)
	      (when (aref m i j)
		(push h1 (cdr descs)))))))
      (setf (high-level-actions h) (mapcar #'(lambda (i) (elt hlas i)) order))))
  h)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; hierarchy isomorphism
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun isomorphic (h1 h2 &key (result-type 'function))
  "isomorphic H1 H2 &key (RESULT-TYPE 'function).  Returns true iff
1) h1 and h2 have the same set of primitive actions
2) there is a bijection f between the high-level actions of h1 and h2 such that:
 a) goal(f(h)) = goal(h) for all h
 b) children(f(h)) = f(children(h))
 c) if h is an action name occurring in both h1 and h2, then f(h) = h

If they are isomorphic, then the isomorphism is returned as a second value in a form determined by result-type, which can either be 'function or 'list (which causes the isomorphism to be represented as an association list)."

  
  (let ((hlas1 (high-level-actions h1))
	(hlas2 (high-level-actions h2))
	(names1 (hla-names h1))
	(names2 (hla-names h2))
	(prim1 (primitive-actions h1)))
    (and (set-eq prim1 (primitive-actions h2))
	 
	 
	 (when (equal (length hlas1) (length hlas2))
	   (let ((common (intersection names1 names2))
		 (diff1 (set-difference names1 names2))
		 (diff2 (set-difference names2 names1)))
	     
	     (flet ((extend-mapping (f)
		      #'(lambda (x)
			  (if (or (member x common) (member x prim1))
			      x
			    (funcall f x))))
		    
		    )
	       
	       ;; Main body: is there a bijection between diff1 and diff2 such that...
	       (mvbind (bijection exists?)
		   (find-element (bijections diff1 diff2)
				 #'(lambda (b)
			
				     ;; ...when extended to all of names1
				     (let ((f (extend-mapping b)))
			  
				       ;; ... for every hla1 in names1 and its image hla2 in names2
				       (every #'(lambda (name)
						  (let ((hla1 (lookup-hla name h1))
							(hla2 (lookup-hla (funcall f name) h2)))
				       
				       
						    (and
				      
						     ;; the goals are the same...
						     (set-eq (goal hla1) (goal hla2))
					
						     ;; ... and the children are isomorphic, i.e.
						     (let ((children1 (children hla1))
							   (children2 (children hla2)))
					  
						       ;; they have the same size and...
						       (and (eq (size children1) (size children2))
					       
							    ;; ... each element of children1 maps into children2
							    (every #'(lambda (child) (member (funcall f (car child)) children2 :key #'car)) children1))))))
				 
					      names1))))
		 
		 (let ((f (extend-mapping bijection)))
		   (values exists? 
			   (when exists?
			     (ecase result-type
			       (function f)
			       (list (mapcar #'(lambda (x) (cons x (funcall f x))) names1)))))))))))))
				     
	     
	      
  
  




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; looking up actions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-condition unknown-action (error)
  ((name :initarg :name :accessor name)
   (hierarchy :initarg :hierarchy :accessor hierarchy)))

(define-condition unknown-hla (error)
  ((name :initarg :name :accessor name)
   (hierarchy :initarg :hierarchy :accessor hierarchy)))

(defun task-type (name maxq)
  "Return 'high-level if name refers to a high-level task, 'primitive if it's primitive.  Signal an error of type unknown-action otherwise."
  (let ((entry (assoc name (high-level-actions maxq) :test #'equal)))
    (if entry
	'high-level
      (if (member? name (primitive-actions maxq))
	  'primitive
	(error 'unknown-action :name name :hierarchy maxq)))))

(defun lookup-hla (name maxq)
  "Return object of type task given name.  Signal an error of type unknown-hla if task not found."
  (let ((entry (assoc name (high-level-actions maxq) :test #'equal)))
    (unless entry
      (error 'unknown-hla :name name :hierarchy maxq))
    (cdr entry)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; debugging
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun pprint-maxq (str maxq)
  (pprint-logical-block (str nil)
    (format str "MAXQ Hierarchy")
    (format str "~:@_Primitive actions: ~w" (primitive-actions maxq))
    (format str "~:@_High-level actions")
    (dolist (hla (high-level-actions maxq))
      (format str "~2I~:@_~w" (car hla))
      (format str "~4I~:@_Goal: ~w~:@_Initiation set: ~w~:@_Children: ~w" 
	      (goal (cdr hla)) (initiation-set (cdr hla)) 
	      (children (cdr hla)))))
  (values))

(set-pprint-dispatch '<maxq-hierarchy> #'pprint-maxq)