;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Various ways of figuring out relevant variables to mappings and sets
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package maxq)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; feature selection for real-valued functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *l2-hist-debug-level* nil)
(defparameter *std-tol* .001)

(defun l1-binned-feature-select (examples var-specs &key (var-cost 1.0) &aux (all-vars (all-vars var-specs)) (num-examples (length examples)))
  "l1-binned-feature-select EXAMPLES VAR-SPECS &key (VAR-COST 1.0)

EXAMPLES - list of pairs (X . Y), where the Y values are real numbers.
VAR-SPECS - as usual
VAR-COST - a positive real

Return the subset S of the variables that minimizes VAR-COST*2^|S| + l where l is the maximum, with I running over all instantiations of all the variables, of the squared difference between the predicted Y-value of I if we use a histogram estimator based only on variables S versus the average Y value over observed examples of I.  Note that the Y values are all preprocessed by subtracting the sample mean and dividing by the sample std.

VAR-COST represents how willing we are to include more variables.  As this parameter approaches zero, the function tries to fit the fine details of the x-y mapping, at the expense of having a large set of variables, and vice versa as it approaches infinity."
  
  (unless (zerop num-examples)
    ;; Figure out the mean and variance beforehand
    (let* ((mean (/ (reduce #'+ examples :key #'cdr) num-examples))
	   (std (sqrt (max (- (/ (reduce #'+ examples :key #'(lambda (x) (expt (cdr x) 2))) num-examples) (* mean mean)) 0))))
    
      (when (> std *std-tol*) ;;; otherwise, it's almost a constant, and can be predicted using no variables
      
	(debug-print *l2-hist-debug-level* "Computing relevant variables.  Based on ~a samples, Y mean is ~$ and std is ~$." num-examples mean std)
	(with-debug-indent
      
	    ;; A helper function that figures out the best predictions given a particular set of variables
	    (flet ((make-predictions (vars)
		     (let ((f (var-set-evaluator vars var-specs))
			   (sums (make-hash-table :test #'equal))
			   (counts (make-hash-table :test #'equal))
			   (predictions (make-hash-table :test #'equal)))
		 
		 
		       ;; 1. Figure out the sum and count for each featurized x
		       (dolist (e examples)
			 (dsbind (x . y) e
			   (let ((fx (funcall f x)))
			     (setf (gethash fx sums) (+ (/ (- y mean) std) (or (gethash fx sums) 0.0))
				   (gethash fx counts) (1+ (or (gethash fx counts) 0))))))
		 
		       ;; 2. Use the sums and counts to generate predictions
		       (maphash #'(lambda (fx sum) (setf (gethash fx predictions) (/ sum (gethash fx counts)))) sums)
		 
		       ;; 3. Return the predictions
		       predictions)))
	
	
	      ;; Precompute the predictions given the entire set of variables
	      (let ((best-predictions (make-predictions all-vars))
		    (representative-states nil)
		    (f (var-set-evaluator all-vars var-specs)))
	  
		;; Also precompute an alist to go from instantiations stored as lists to state objects that the feature functions apply to
		(dolist (e examples)
		  (let* ((x (car e))
			 (fx (funcall f x)))
		    (unless (assoc fx representative-states :test #'equal)
		      (push (cons fx x) representative-states))))
	       
	  
		;; Next, minimize, over sets of variables...
		(minimizing-element
		 (make-instance '<powerset> :base-set all-vars :iteration-order ':increasing-size)
		 #'(lambda (vars)
		     (let ((predictions (make-predictions vars))
			   (f (var-set-evaluator vars var-specs))
			   (total-var-cost (* var-cost (expt 2 (length vars)))))

		       ;; ... sum of variable set size and
		       (+ total-var-cost
		       
			  ;; ... maximum over observed instantiations to all vars
			  (mvbind (ind max-val best)
			      (argmax
			       representative-states
			       :key #'(lambda (entry)
					(dsbind (fx . x) entry
				       
					  ;; ... of the absolute difference between
					  (abs-diff 
					   ;; 1. the best prediction for that overall instantiation
					   (gethash fx best-predictions)
					   ;; 2. The prediction given the current set of variables
					   (gethash (funcall f x) predictions)))))

			    (declare (ignore ind))
			    (setf max-val (mymax max-val 0))
			    (debug-print *l2-hist-debug-level* "Vars: ~a~80Tobj: ~$~95Tpred-loss: ~$~120Tinst: ~a" vars (+ max-val total-var-cost) max-val (cdar best))
			    max-val))))))))))))



(defun l2-histogram-feature-select (examples var-specs &key (var-cost 1.0) (combination-type ':average) &aux (num-examples (length examples)))
  ;; Obsolete
  "l2-histogram-feature-select EXAMPLES VAR-SPECS &key (VAR-COST 1.0) (COMBINATION-TYPE ':average)

EXAMPLES - list of pairs (X . Y), where the Y values are real numbers.
VAR-SPECS - as usual
VAR-COST - a positive real
COMBINATION-TYPE - either ':maximum or ':average or ':max-over-states

Return the subset S of the variables that minimizes VAR-COST*|S| + l where l is the squared loss paid on these examples by predicting Y using a piecewise constant approximator based on variables S.

The squared loss is computed from the examplewise squared losses depending on COMBINATION-TYPE:
- :average - average loss over examples
- :maximum - maximum loss over examples

VAR-COST represents how much the squared loss must go down by before we're willing to add a new variable.  As this parameter approaches zero, the function tries to fit the fine details of the x-y mapping, at the expense of having a large set of variables, and vice versa as it approaches infinity."

  (debug-print *l2-hist-debug-level* "Computing relevant variables")
  (with-debug-indent
      (minimizing-element 
       (make-instance '<powerset> :base-set (all-vars var-specs) :iteration-order ':increasing-size)
       #'(lambda (vars)
	   (let ((f (var-set-evaluator vars var-specs))
		 (sums (make-hash-table :test #'equal))
		 (counts (make-hash-table :test #'equal))
		 (predictions (make-hash-table :test #'equal)))
	 
	     ;; 1. figure out the sum and count for each featurized x
	     (dolist (e examples)
	       (dsbind (x . y) e
		 (let ((featurized-x (funcall f x)))
		   (setf (gethash featurized-x  sums) (+ y (or (gethash featurized-x  sums) 0.0))
			 (gethash featurized-x  counts) (+ 1 (or (gethash featurized-x  counts) 0))))))
	 
	     ;; 2. figure out the prediction (average y) for each featurized x
	     (maphash #'(lambda (fx sum) (setf (gethash fx  predictions) (/ sum (gethash fx  counts)))) sums)

	     ;; 3. Return the sum of
	     (let ((retval
		    (my+
		 
		     ;; a. the l2 loss
		     (flet ((loss (example)
			      (dsbind (x . y) example
				(expt (- y (gethash (funcall f x)  predictions)) 2))))
		       (case combination-type
			 (:average (/ (sum-over examples #'loss) num-examples))
			 (:maximum (reduce #'mymax examples :key #'loss))))
		 
		     ;; b. the num-features penalty
		     (* (size vars) var-cost))))
	   
	       (debug-print *l2-hist-debug-level* "Loss for var set ~a is ~a" vars retval)
	       retval)))))
  )
	 
	 
	       
	     
	 

  


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; conditional independence based criterion using
;; mutual information
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *mutual-info-indep-threshold* 0.1)



(defun cond-indep-relevant-vars (examples var-specs &key (test #'equalp) &aux (all-vars (all-vars var-specs)))
  "cond-indep-relevant-vars EXAMPLES VAR-SPECS &key (TEST #'equalp)

EXAMPLES - list of pairs (X . Y)
VAR-SPECS - mapping from var-name to var-spec objects
TEST - used to test equality on Y

Looks for a small  set W of the variables such that for each joint value w of W, the normalized mutual information of X\W and Y given W=w is less than *mutual-info-indep-threshold*"

  (declare (ignore test))
  (break  "This code no longer guaranteed to work due to changes in empirical-dist and set equality checking code.")
  (with-debug-indent 
      (debug-print 4 "Computing relevant vars using mutual information criterion")
    (let ((evaluated-examples (make-array (length examples))))
      (do-elements (w (make-instance '<powerset> :base-set all-vars :iteration-order ':increasing-size))
	(let* ((evaluator (compose (var-set-evaluator w var-specs) #'car))
	       (remaining-vars-evaluator (compose (var-set-evaluator (complement-set w :base-set all-vars) var-specs) #'car))
	       (insts (var-insts evaluator examples)))
	  (map-into evaluated-examples evaluator examples)
	  (debug-print 5 "Considering var set ~a, which has instantiations ~a" w insts)
	  (with-debug-indent
		
	      (when (every #'(lambda (inst)
			       (let* ((dist (empirical-dist
					     (loop
						 for evaluated-ex across evaluated-examples
						 for ex in examples
						 when (equal inst evaluated-ex)
						 collect ex)
					     :test #'equalp))
				      (info (mutual-information dist remaining-vars-evaluator #'cdr)))
				 (debug-print 6 "For instantiation ~a, empirical dist is~&~/prob::pprint-prob-dist/~&Mutual info is ~a" inst dist info)
				 (< info *mutual-info-indep-threshold*)))
			   insts)
		(debug-print 4 "Returning ~a" w)
		(return w))))))))
  




(defun var-insts (evaluator examples)
  (let ((l nil))
    (dolist (e examples l)
      (pushnew (funcall evaluator e) l :test #'equal))))

	


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; determinism criterion
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defparameter *determinism-threshold* .9
  "Used in the function determining-vars.  An X-Y dependency is deterministic if for each value of X, the most frequent value of Y occurs on at least this fraction of the examples.")


(defparameter *rel-debug-level* 4)


(defun determining-vars (examples var-specs &key (test #'equalp) &aux (all-vars (all-vars var-specs)))
  "determining-vars EXAMPLES VAR-SPECS &key (TEST #'equalp)

EXAMPLES - list of pairs (X . Y)
VAR-SPECS - mapping from var-name to var-spec objects
TEST - used to test equality on Y

Also affected by
*determinism-threshold*

Does a search over subsets of the variables, in increasing order of size, until one is found that makes Y deterministic at level *determinism-threshold*.  Asserts if no such subset is found.

Assumes X values can be compared using #'equal."
  
  (unless (is-empty examples)
    (flet ((determinism (vars)
	     (determinism-level examples (mapcar #'(lambda (v) (var-function v var-specs)) vars) test)))
      (let ((max-determinism (determinism all-vars)))
	(debug-print *rel-debug-level* "~&Max determinism is ~a" max-determinism)
	(mvbind (vars found?)
	    (find-element 
	     (make-instance '<powerset> :base-set all-vars :iteration-order ':increasing-size)
	     #'(lambda (s) 
		 (let ((det-level (determinism s)))
		   (debug-print *rel-debug-level* "~&Determinism level for ~a was ~a" s det-level)
		   (> det-level (* max-determinism *determinism-threshold*)))))
	  (assert found? nil "Was not able to find variables that achieved max determinism.  This shouldn't happen.  Table of values was ~/maxq::pprint-table/"
		  examples (make-table examples (mapset 'list #'(lambda (v) (var-function v var-specs)) all-vars) test))
	  vars)))))


(defun make-table (examples var-fns test)
  (let ((h (make-hash-table :test #'equal)))
    (dolist (ex examples h)
      (dsbind (x . y) ex
	(let* ((x-vals (mapcar #'(lambda (f) (funcall f x)) var-fns))
	       (x-table (orf (gethash x-vals h) (make-hash-table :test test))))
	  (setf (gethash y x-table) (1+ (or (gethash y x-table) 0))))))))

(defun pprint-table (&rest args)
  (with-pprint-dispatch ('hash-table #'pprint-hash)
    (pprint (second args) (first args))))


(defun determinism-level (examples var-fns test)
  ;; create a table mapping each (featurized) x example to a table of counts for all the observed y's for that x
  (let ((h (make-table examples var-fns test)))

    ;; determinism level for a particular x value
    (flet ((x-table-determinism (table)
	     (let ((sum (loop for count being each hash-value in table summing count)))
	       (assert (> sum 0) nil "Counts in table ~a were 0" table)
	       (/ (loop for count being each hash-value in table maximizing count) sum))))

      ;; minimize determinism level over all x-values
      (loop
	  for x-table being each hash-value in h using (hash-key x-value)
	  do (debug-print (1+ *rel-debug-level*) "~&X-vals: ~a~&Table : ~/utils:pprint-hash/" x-value x-table)
	  minimizing (x-table-determinism x-table)))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; learning of preconditions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun tightest-covering-set (positive negative sets)
  "tightest-covering-set POSITIVE-EXAMPLES NEGATIVE-EXAMPLES SETS
Return the member of SETS that covers as few of the NEGATIVE-EXAMPLES while still covering all the positive examples."
  (minimizing-element 
   
   ;; Subset of SETS that cover the positive examples
   (filter ':implicit sets #'(lambda (s) (subset positive s)))
   
   ;; Function that counts the intersection with the negative examples
   #'(lambda (s) (sum-over negative #'(lambda (example) (indicator (member? example s)))))))

(defun precond-relevant-vars (positive negative goal-language)
  "precond-relevant-vars POSITIVE-EXAMPLES NEGATIVE-EXAMPLES GOAL-LANGUAGE
Find the tightest covering set in the goal language for the examples and return the variables relevant to it."
  (goal-rel-vars (tightest-covering-set positive negative (all-goals goal-language))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; helper functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun var-set-evaluator (w var-specs)
  "var-set-evaluator W VAR-SPECS.  Return a function that takes in an object and returns the list of values of variables W for that object."
  (let ((functions (mapset 'list #'(lambda (v) (vs-function (evaluate var-specs v))) w)))
    #'(lambda (x)
	(mapcar #'(lambda (f) (funcall f x)) functions))))