(in-package maxq)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; The overall task learner.
;; Factored into:
;; 1) Subtask environment wrapper
;; 2) Hierarchical Q-learner
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun task-learner (env init-states app-set goal goal-value prim subtasks num-steps
		     &key (max-episode-length 'infty) (exploration-policy 'random) (learning-rate .1))
  "task-learner ENV INIT-STATES APPLICABLE-SET GOAL GOAL-VALUE PRIMITIVE-ACTIONS SUBTASKS NUM-STEPS &key (MAX-EPISODE-LENGTH 'infty) (EXPLORATION-POLICY ':random)

ENV - an environment that supports the reset-to-state operation
INIT-STATES - a set of initial states
APPLICABLE-SET - the set of applicable states for the task
GOAL - the goal set 
GOAL-VALUE - the value for reaching a goal state (the value for exiting at a nongoal state is 0)
PRIMITIVE-ACTIONS - set of actions that can be directly performed in environment
SUBTASKS - list of entries of form (NAME APPLICABLE-SET POLICY)
NUM-STEPS - number of steps of learning
MAX-EPISODE-LENGTH - Reset to a new state whenever we go for this many steps without finishing an episode.  
EXPLORATION-POLICY - Either an exploration policy, or the symbol ':random
LEARNING-RATE - Defaults to .1

Performs hierarchical Q-learning within this task.  Returns three values: the Qr, Qc, and Qe components of the Q-function within this task."
  (declare (ignore max-episode-length))
  (let* ((e (make-instance '<subtask-env> :parent-env env :init-states init-states 
			   :app-set (filter ':implicit app-set #'(lambda (x) (not (member? x goal))))
			   :subtasks subtasks :prim prim))
	 (alg (make-instance '<task-learner> :goal goal :goal-value goal-value :env e :lrate learning-rate)))
    (rl-user:learn e exploration-policy (list alg) num-steps :hist-length 1 :step-print-inc nil :ep-print-inc nil)
    (knowledge-state alg nil)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 1. A wrapper for the environment induced by a particular 
;; task: terminates as soon as we leave the task applicable
;; set, and makes the subtasks look like atomic actions
;; Does not include the exit values (these will be 
;; passed to the hierarchical Q-learner directly)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <subtask-env> (<fully-observable-env>)
  ((parent-env :accessor parent-env :initarg :parent-env)
   (init-states :accessor init-states :initarg :init-states)
   (subtasks :accessor subtasks :initarg :subtasks)
   (prim :accessor prim :initarg :prim)
   (applicable-set :accessor app-set :initarg :app-set)
   (actions :accessor actions)))


(defmethod initialize-instance :after ((e <subtask-env>) &rest args &key prim subtasks)
  (declare (ignore args))
  (setf (actions e) (append prim (domain subtasks))))

(defmethod sample-init ((e <subtask-env>))
  (sample-uniformly (init-states e)))

(defmethod avail-actions ((e <subtask-env>) s)
  (declare (ignore s))
  (actions e))

(defmethod sample-next ((e <subtask-env>) s a &aux (par (parent-env e)))
  (if (member? a (prim e))
      (sample-next par s a)
    (dbind (app-set pol) (evaluate (subtasks e) a)
      (let ((r 0)
	    (state s))
	(loop
	  (mvbind (s2 reward) (sample-next par state (policy:make-choice pol state))
	    (incf r reward)
	    (setf state s2)
	    (unless (member? state app-set)
	      (return (values state r)))))))))

(defmethod is-terminal-state ((e <subtask-env>) s)
  (not (member? s (app-set e))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 2. Hierarchical Q-learner
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *task-learner-debug-level* nil)

(defclass <task-learner> (rl-obs:<learning-algorithm>)
  ((qr :accessor qr)
   (qc :accessor qc)
   (qe :accessor qe)
   (q :accessor q)
   (learning-rate :accessor lrate :initarg :lrate)
   (prev-s :accessor prev-s)
   (goal :accessor task-goal :initarg :goal)
   (goal-value :accessor goal-value :initarg :goal-value)
   (env :initarg :env :accessor env)))

(defmethod knowledge-state ((alg <task-learner>) &optional (fresh t))
  (let ((q (q alg)))
    (if fresh (clone q) q)))

(defmethod rl-obs:reset ((alg <task-learner>) &aux (env (env alg)))
  (setf (qr alg) (make-instance 'q-fn:<env-q-function> :env env)
	(qc alg) (make-instance 'q-fn:<env-q-function> :env env)
	(qe alg) (make-instance 'q-fn:<env-q-function> :env env)
	(q alg) (make-instance 'q-fn:<sum-q-function> :q-fns (list (qr alg) (qc alg) (qe alg)) :strict nil)))

(defmethod inform-start-episode ((alg <task-learner>) s)
  (setf (prev-s alg) s))

(defmethod inform-env-step ((alg <task-learner>) a r s2 term)
  
  ;; We've just observed a transition s,a -> r,s2
  (let ((s (prev-s alg))
	(qr (qr alg))
	(qc (qc alg))
	(qe (qe alg))
	(lrate (lrate alg))
	(a2 (unless term 
	      (handler-case
		  (q-fn:best-choice (q alg) s2)
		(q-fn:unknown-state-action () 0)))))
    
    (debug-prompt *task-learner-debug-level* "Observed transition~4I~&s1 : ~a~&a  : ~a~&r  : ~a~&s2 : ~a~&Term : ~a" s a r s2 term)
    
    (flet ((backup (q target)
	     (debug-print *task-learner-debug-level* "~4TBackup~8Is : ~a~&a : ~a~&v : ~a" s a target)
	     (q-fn:update q s a target lrate))
	   (evaluate-at-dest (q) 
	     (handler-case
		 (q-fn:evaluate q s2 a2)
	       (q-fn:unknown-state-action () 0))))

	   
	   
      (backup qr r)
      (if term
	  (progn 
	    (backup qc 0)
	    (backup qe (iwhen (member? s2 (task-goal alg)) (goal-value alg))))
	(progn
	  (backup qc (+ (evaluate-at-dest qr) (evaluate-at-dest qc)))
	  (backup qe (evaluate-at-dest qe))))))
  
  (setf (prev-s alg) s2))





