(defpackage millennium-cluster
  (:documentation "
Package millennium-cluster - contains functions for setting up and running jobs on the millennium cluster.

Setup functions
---------------
setup-gexec-dir     create a staging area on the cluster machines from which the jobs can be launched

Functions callable from scripts
-------------------------------
init-run            call this at the beginning of a script
setup-random-state  components of init-run
create-subidr
get-run-id")
  (:nicknames cluster)
  (:use common-lisp
	#+allegro excl
           )
  (:export setup-gexec-dir
	   init-run
	   setup-random-state
	   get-run-id
	   create-subdir))

(in-package millennium-cluster)

	
;; setup-gexec-dir
;; 
;;
#+allegro (defun setup-gexec-dir (target-dir)
  "setup-gexec-dir TARGET-DIR.  Dumps the current lisp image to the given target directory and sets up the logical pathname translations, etc. To use this to set up an image on the millennium cluster :
- log in to a millennium machine
- start lisp and load hrl-system
- make sure the current directory is hrl/lisp
- call setup-gexec-dir with target-dir a subdirectory of /work, e.g. /work/bhaskara/hrl

Once the image is set up, it can be used to run multiple copies of a program on the cluster machines.  The program, e.g. test.lisp, should have no input files and may write output files in subdirectories of the current directory.  The first argument will be a number which can be used to choose parameter values/the output directory.  E.g. test.lisp might write its output to the directory test/$1
- copy test.lisp to the target directory
- chdir to the target directory
- make sure gexec-lisp.sh is executable
- run it using gexec, e.g. to run test.lisp on 10 machines : gexec -n 10 `pwd`/gexec-lisp.sh test.lisp"
    
  ;; make sure everything is compiled and loaded
  ;;(make-all)

  ;; if the directory already exists, prompt user about overwriting it
  (let ((current (first (directory target-dir))))
    (when current
      (if (y-or-n-p "Directory ~a already exists.  Overwrite? " current)
	  (delete-directory-and-files current)
	(assert () () "Exiting without doing anything."))))
  
  (format t "~&Creating directory")
  (ensure-directories-exist (format nil "~a/dummy" target-dir) :verbose t)
  
  (format t "~&Copying gexec script")
  (sys:copy-file "misc/gexec-lisp.sh" (format nil "~a/gexec-lisp.sh" target-dir))
  (sys:copy-file "misc/gexec-run" (format nil "~a/gexec-run" target-dir))

  ;; this part is hacky - instead of hardcoding the library file location, should just
  ;; copy to target-dir and modify hrl-hosts appropriately
  (format t "~&Copying pathname translation file")
  (sys:copy-file "misc/hrl-hosts" (format nil "~a/hrl-hosts" target-dir))
  (format t "~&Copying matlisp libraries")
  ;(sys:copy-file "external/matlisp/lib/libmatlisp.so" "/work/bhaskara/mysos/lib/libmatlisp.so" :overwrite t)
  
  
  (let ((host-file (format nil "~a/hrl-hosts" target-dir)))
    (unless (member host-file (logical-pathname-translations-database-pathnames) :test #'equal)
      (pushnew (format nil "~a/hrl-hosts" target-dir) (logical-pathname-translations-database-pathnames))))
  
  (format t "~&Dumping lisp image")
  ;; dump lisp image
  (dumplisp :name (format nil "~a/hrl.dxl" target-dir) :suppress-allegro-cl-banner t))



#+allegro (defun init-run (prefix)
  "cluster-init PREFIX.  Gets this run's id using get-run-id, then calls cluster-init-random-state and create-subdir.  Returns 1) name of output subdirectory 2) ID (as an integer)"
  (let ((id (get-run-id)))
    (setup-random-state id)
    (values (create-subdir prefix id) (read-from-string id))))


#+allegro (defun setup-random-state (id)
  "cluster-init-random-state ID.  Used on cluster runs to initialize random state based on which machine we're on."
  (sleep (read-from-string id))
  (setf *random-state* (make-random-state t)))


#+allegro (defun get-run-id ()
  "get-run-id.  Get a unique identifier for this run, which will be a string containing a single number.  In a situation where multiple versions of a script are running on different cluster machines using gexec-lisp.sh, each version will have a different number, passed in as a command-line argument to lisp.  If there is no command-line argument, then this will instead prompt for an ID."
  (if (> (length (sys:command-line-arguments)) 1)
      (sys:command-line-argument 1)
    (progn 
      (format t "~&This script is intended for running on the Millennium cluster.  If you would like to run it interactively, please enter an integer id.")
      (format nil "~a" (read)))))


#+allegro (defun create-subdir (prefix id)
  "create-subdir PREFIX.  Creates a subdirectory for this particular run and return its name.  So for example, if called with PREFIX being \"foo\" and id \"2\", will ensure that the subdirectory foo/2 of the current directory exists, and return the string \"foo/2\""
  (let ((dir-name (concatenate 'string prefix (format nil "/~a" id))))
    (ensure-directories-exist (concatenate 'string dir-name "/dummy"))
    dir-name))
  

(in-package cl-user)






