(in-package dmcmc)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; filter class
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <dmcmc-filter> ()
  ((observation-sequence :accessor obs :type vector)
   (ssm :initarg :ssm :accessor ssm)
   (inference-engine :accessor inf-engine)
   (burnin :initarg :burnin :accessor burnin :initform 0)
   (gap :initarg :gap :accessor gap :initform 1)
   (decay-rate :initarg :decay-rate :accessor decay-rate)
   (sampler :accessor sampler)
   (num-samples :initarg :num-samples :accessor num-samples))
  (:documentation "Class <dmcmc-filter>

Implements the basic filtering algorithm operations using decayed MCMC.

Initargs
:ssm - <state-space-model>
:decay-rate - as in <dmcmc>
:burnin - burnin for each mcmc run
:num-samples - number of samples per mcmc run
")) 

(defmethod initialize-instance :after ((f <dmcmc-filter>) &rest args)
  (declare (ignore args))
  (reset-filter f))

(setf (documentation 'num-samples 'function) "Accessor for number of samples used by <dmcmc-filter>.  Can be changed online, in between updates.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Filtering-algorithm operations
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod reset-filter ((d <dmcmc-filter>))
  (setf (obs d) (make-array 0 :adjustable t :fill-pointer 0)))


(defmethod incorporate-evidence ((d <dmcmc-filter>) obs)
  (vector-push-extend obs (obs d))
  (gen-samples d))

(defmethod incorporate-evidence-sequence ((d <dmcmc-filter>) obs)
  (map nil #'(lambda (o) (vector-push-extend o (obs d))) obs)
  (gen-samples d)
  (current-state-dist d))
  

(defun gen-samples (d)
  (let* ((burnin (burnin d))
	 (gap (gap d))
	 (num-samples (num-samples d))
	 (sample-gatherer (make-instance '<sample-log> :burnin (burnin d) :gap (gap d)))
	 (engine (make-instance '<dmcmc>
		   :ssm (ssm d)
		   :obs (obs d)
		   :observers (list sample-gatherer)
		   :decay-rate (decay-rate d))))
    (setf (inf-engine d) engine
	  (sampler d) sample-gatherer)
    (generate-samples engine (+ burnin (* gap num-samples)))))

(defmethod current-state-dist ((d <dmcmc-filter>))
  (if (zerop (length (obs d)))
      (init-dist (ssm d))
    (make-instance '<empirical-dist> :samples (samples (sampler d)) :hash-function #'equalp)))
    

