(defpackage dmcmc
  (:documentation "Package dmcmc

Types
-----
<dmcmc>
<dmcmc-filter>
<monte-carlo-observer>
<sample-log>
<debug-observer>

Operations
----------
reset-sample
draw-sample
mc-estimate
reset-observer
num-samples

Decay rates
-----------
polynomial-decay
exponential-decay
fixed-lag
fixed-window

Running experiments
-------------------
convergence-curve
single-timestep-estimates

Special vars
------------
*dmcmc-print-inc*
")
  (:export
   <dmcmc>
   <dmcmc-filter>
   <monte-carlo-observer>
   <sample-log>
   <debug-observer>
   
   reset-sample
   draw-sample
   mc-estimate
   reset-observer
   num-samples
   
   polynomial-decay
   exponential-decay
   fixed-lag
   fixed-window

   convergence-curve
   single-timestep-estimates

   *dmcmc-print-inc*
)

  (:use
   cl
   prob
   utils
   ssm
   set
   mapping))


(in-package dmcmc)

(defvar *dmcmc-print-inc* 500)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; inference engine class
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <dmcmc> ()
  ((state-space-model :initarg :ssm :reader ssm :type <state-space-model>)
   (state-sequence :accessor states :type vector)
   (observation-sequence :initarg :obs :accessor obs :type vector)
   (observers :initarg :observers :accessor observers :initform nil)
   (time-dist :accessor time-dist :type [prob-dist]))
  (:documentation "Class <dmcmc>

An instance of this class represents a decayed MCMC 'inference engine', i.e., it holds the computational state of decayed MCMC running on some particular evidence sequence and model.

Initargs
:ssm - the <state-space-model>
:obs - the observation sequence, which must be a vector.  
:decay-rate - a [mapping] (e.g., function, vector) from 0,...,T-1 to the positive reals, where the decay rate at k represents the (unnormalized) weight given to time T-1-k.  Defaults to uniform.
:observers - a list of observer objects.  Whenever the engine is reset, or a sampling step is done, inform-sample will be called on each observer.

:pause-every - defaults to nil.  Either nil or a fixnum.
:print-samples - Boolean.  defaults to nil.
:print-progress-inc - defaults to nil.  If it's a fixnum, then a '.' is printed every so many samples.
"))

(defmethod initialize-instance :after ((d <dmcmc>) &rest args &key (decay-rate nil d-supp) obs (pause-every nil) (print-progress-inc nil) (print-samples nil) &aux (n (length obs)))
  (declare (ignore args))
  (unless d-supp (setf decay-rate (constantly (/ 1 n))))
  (let ((total (sum-over n #'(lambda (i) (evaluate decay-rate i)))))
    (setf (time-dist d) (mapset 'vector #'(lambda (i) (float (/ (evaluate decay-rate (- n 1 i)) total))) n))
    (push (make-instance '<debug-observer> :pause-every pause-every :print-samples print-samples :step-print-inc print-progress-inc) (observers d)))
  (debug-print 3 "Initializing dmcmc engine with:~& Timestep dist ~/prob:pprint-prob-dist/~& Evidence ~a" (time-dist d) obs))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; basic accessors
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun num-ts (d)
  (length (obs d)))

(defmacro notify-observers (d op &rest args)
  (let ((o (gensym))
	(arg-vars (mapcar #'(lambda (x) (declare (ignore x)) (gensym)) args)))
    `(let ,(mapcar #'list arg-vars args)
       (dolist (,o (observers ,d))
	 (,op ,o ,@arg-vars)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; basic operations on observers
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric inform-sample (obs states diffs)
  (:documentation "inform-sample DMCMC-OBS STATES DIFFS.  Inform the observer that a sampling step happened.  STATES is the new state sequence.  DIFFS is the diffs from the old sample.  Note that each of the last two arguments is redundant given the other, but both are useful in different situations."))

(defgeneric reset-observer (obs)
  (:documentation "Reset this observer to the state before any sampling was done."))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Sampling operations
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun generate-samples (d n)
  "generate-samples DMCMC N.  Do N samples of decayed mcmc."
  (debug-print 2 "Beginning dmcmc sampling")
  (with-debug-indent 
      (reset-sample d)
    (dotimes (i (1- n))
      (when (divisible-by i *dmcmc-print-inc*)
	(debug-print 2 "~a samples done" i))
      (draw-sample d))))

(defun reset-sample (d)
  (setf (states d) (state-prefix (sample (ssm d)) (num-ts d)))
  (debug-print 4 "Initializing state sequence to ~a" (states d))
  (notify-observers d inform-sample (states d) 'initial-sample))

(defun draw-sample (d)
  "draw-sample DMCMC.  Do one step of decayed Gibbs sampling."
  (let ((ts (sample (time-dist d)))
	(states (states d)))
    (setf (aref states ts)
      (let ((dist (timestep-gibbs-sampling-distribution d ts)))
	(debug-print 4 "Sampling timestep ~a from distribution ~/prob:pprint-prob-dist/" ts dist)
	(sample dist)))
    (debug-print 4 "State sequence is now ~a" (states d))
    (notify-observers d inform-sample states (list 'flip-timestep ts (aref states ts)))))

(defun timestep-gibbs-sampling-distribution (d ts)
  (let* ((ssm (ssm d))
	 (trans-dist (trans-dist ssm))
	 (states (states d))
	 
	 ;; start with the prior distribution given the previous timestep
	 (dist (if (zerop ts) (init-dist ssm) (cond-dist trans-dist (aref states (1- ts))))))
    
    ;; condition on next state if necessary
    (unless (= ts (1- (num-ts d)))
      (_f condition-on-dist dist (trans-dist ssm) (aref states (1+ ts))))
    
    ;; condition on observations and return
    (condition-on-dist dist (obs-dist ssm) (aref (obs d) ts))))
