(in-package dmcmc)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Monte Carlo estimation
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <monte-carlo-observer> ()
  ((rv :initarg :rv :reader rv)
   (estimate :accessor estimate :reader mc-estimate :initform 0.0)
   (num-steps :accessor num-steps :initform 0)
   (burnin :accessor burnin :initarg :burnin :initform 0)
   (gap :accessor gap :initarg :gap :initform 1))
  (:documentation "An MCMC observer that implements Monte Carlo estimation, i.e., it maintains the sample average of some real-valued random variable.

Initargs
:rv - a random variable
:burnin - the number of samples to skip until we start keeping an estimate.  Defaults to 0 (start immediately).
:gap - interval between samples used for estimate.  Defaults to 1 (use every sample)."))

(defmethod inform-sample ((obs <monte-carlo-observer>) states diffs)
  (declare (ignore diffs))
  (let ((num-steps (incf (num-steps obs))))
    (mvbind (div rem) (floor (- num-steps (burnin obs)) (gap obs))
      (when (zerop rem)
	(let ((x (evaluate-rv (rv obs) states)))
	  (check-type x number)
	  (avgf (estimate obs) x (/ 1 div)))))))

(defmethod reset-observer ((obs <monte-carlo-observer>))
  (setf (estimate obs) 0.0
	(num-steps obs) 0))
			   

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Log of samples (of state at final step)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <sample-log> ()
  ((samples :accessor samples :initform (make-array 0 :adjustable t :fill-pointer 0 :element-type 'vector))
   (burnin :accessor burnin :initarg :burnin :initform 0)
   (gap :accessor gap :initarg :gap :initform 1)
   (num-steps :initform 0 :accessor num-steps)))

(defmethod inform-sample ((obs <sample-log>) states diffs)
  (declare (ignore diffs))
  (let ((samples (samples obs))
	(ns (incf (num-steps obs))))
    (mvbind (div rem) (floor (- ns (burnin obs)) (gap obs))
      (declare (ignore div))
      (when (zerop rem)
	(vector-push-extend (clone (slast states)) samples)))))

(defmethod reset-observer ((obs <sample-log>))
  (setf (num-steps obs) 0))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Debugging
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <debug-observer> ()
  ((num-samples :accessor num-samples :initform 0)
   (step-print-inc :accessor step-print-inc :initarg :step-print-inc :initform nil)
   (output-stream :accessor str :initarg :str :initform t)
   (print-samples :initform nil :initarg :print-samples :accessor print-samples)
   (pause-every :initform nil :initarg :pause-every :accessor pause-every)))

(defmethod inform-sample ((obs <debug-observer>) states diffs &aux (str (str obs)) (ns (incf (num-samples obs))))
  (if (print-samples obs)
      (progn
	(fresh-line str)
	(pprint-logical-block (str nil)
	  (format str "Step ~a.~4I~:@_State: ~w~:@_Diffs: ~w" ns states diffs)))
    (when (aand (step-print-inc obs) (divisible-by ns it))
      (write-char #\. str)))
  (when (aand (pause-every obs) (divisible-by ns it))
    (format str "~&Press enter to continue. ")
    (read-line *query-io*)))
  
  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; History gatherer
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <estimate-history> (<sample-log>)
  ((fn :accessor estimate-fn :initarg :fn)
   (estimate-timesteps :accessor estimate-timesteps :initarg :estimate-timesteps)
   (estimates :accessor estimate-history)))
  

(defmethod reset-observer ((obs <estimate-history>))
  (setf (estimate-history obs) (make-adjustable-array))
  (call-next-method))

(defmethod inform-sample ((obs <estimate-history>) states diffs)
  (declare (ignorable states diffs))
  (call-next-method)
  (when (member (num-steps obs) (estimate-timesteps obs))
    (vector-push-extend (funcall (estimate-fn obs) (make-instance '<empirical-dist> :samples (samples obs)))
			(estimate-history obs))))


  


  
   
  
  