(defpackage dmcmc-scaling
  (:use cl 
	utils
	set
	prob
	lin-alg
	dmcmc
	exact-filter
	particle-filter
	ssm))

(in-package dmcmc-scaling)




(defvar *flip-probs* #(.6 .3))
(defvar *obs-det* .95)
(defvar *hist-length* 10)
(defvar *expt* 2)

(defclass <bitflip-trans-dist> (<prob-dist>)
  ((num-states :accessor num-states :reader sample-space)
   (m :accessor m :initarg :m)
   (n :accessor n :initarg :n)
   (flip-probs :accessor flip-probs)))

(defmethod initialize-instance :after ((d <bitflip-trans-dist>) &rest args &key n m)
  (declare (ignore args))
  (setf (num-states d) (expt 2 n)
	(flip-probs d) (mapset '(vector float)
			 #'(lambda (i)
			     (let ((neighbour (if (zerop i) (1- n) (1- i))))
			       (aref *flip-probs* (indicator (logbitp neighbour m)))))
			 n)))
  

(defmethod prob ((d <bitflip-trans-dist>) k)
  (with-slots (m n flip-probs) d
    (reduce-set
     #'* n
     :key #'(lambda (i)
	      (let ((p (aref flip-probs i)))
		(if (eq (logbitp i m) (logbitp i k))
		    (- 1 p)
		  p))))))
  
  

(defclass <bitflip-obs-dist> (<prob-dist>)
  ((m :initarg :m :accessor m)
   (n :initarg :n :accessor n)
   (num-obs :accessor num-obs :reader sample-space)))

(defmethod initialize-instance :after ((d <bitflip-obs-dist>) &rest args &key n)
  (declare (ignore args))
  (setf (num-obs d) (expt 2 n)))
  

(defmethod prob ((d <bitflip-obs-dist>) k)
  (with-slots (m n num-obs) d
    (iwhen (between2 k 0 num-obs)
      (reduce-set #'* n :key #'(lambda (i) 
				 (if (eq (logbitp i m) (logbitp i k))
				     *obs-det*
				   (- 1 *obs-det*)))))))
  





(defun make-model (n)
  "Make a bitflip ssm
Depends also on *flip-probs* and *obs-det*

The model works as follows: you have n bits, initially iid uniform.  For the transitions, let the neighbour of bit i be bit i-1 (or n-1 for bit 0).  Flip bit i with probability (aref *flip-probs* b) where b is the value of the neighbouring bit.  For the observations, independently flip each bit with probability 1- *obs-det*."

  (let ((num-states (expt 2 n)))
    (make-instance '<state-space-model>
      :init-dist (make-unif-dist-over-set num-states)
      :trans-dist #'(lambda (m)
		      (make-instance '<bitflip-trans-dist> :m m :n n))
      :obs-dist #'(lambda (m)
		    (make-instance '<bitflip-obs-dist> :m m :n n)))))


(defun generate-point (n l)
  (let* ((ssm (make-model n))
	 (ef (make-instance '<exact-filter> :ssm ssm))
	 (decay (polynomial-decay *expt*))
	 ed)
    (mvbind (states obs) (sample-finite-trajectory ssm *hist-length*)
      (declare (ignore states))
      (format t "~&Observation sequence ~a~&Exact filtering -> ~/prob:pprint-prob-dist/" obs ed)
      (time (setf ed (incorporate-evidence-sequence ef obs)))
    
      (dolist (ns l)
	(let ((df (make-instance '<dmcmc-filter> :ssm ssm :decay-rate decay :num-samples ns))
	      dd)
	  (format t "~&~&~&~a samples~&" ns)
	  (time (setf dd (incorporate-evidence-sequence df obs)))
	  (format t "~&Dist ~/prob:pprint-prob-dist/~&with error ~a" dd (total-variation-distance dd ed)))))))

	
	  