(defpackage particle-filter
  (:documentation "Package particle-filter

Types
-----
<particle-filter>
")
  (:export
   <particle-filter>)
  (:use
   cl
   utils
   set
   prob
   ssm)
  (:nicknames pf))

(in-package particle-filter)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Class def
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <particle-filter> ()
  ((timestep :accessor ts :type fixnum)
   (particles :type vector :accessor particles)
   (num-particles :accessor num-particles :type fixnum :initarg :num-particles)
   (ssm :initarg :ssm :accessor ssm)
   (trans-strategy :initform ':prior :initarg :trans-strategy :accessor trans-strategy)
   (resampling-strategy :initform ':correlated :initarg :resampling-strategy :accessor resampling-strategy)
   (new-particles :accessor new-particles)
   (weights :accessor weights :type vector)
   (first-step? :accessor first-step?))
  (:documentation "Class <particle-filter>

Represents a vanilla particle filter: use the transition distribution to sample, and the observation distribution to reweight, then resample.

Initargs
:num-particles - number of particles to use.
:ssm - state space model.
:resampling-strategy - either ':simple or ':correlated (default).
:trans-strategy - how to transition particles forward.  Either ':prior (use the ssm transition distribution) or ':posterior (use the transition distribution conditioned on the next observation).
"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Filter methods
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod initialize-instance :after ((pf <particle-filter>) &rest args &key num-particles)
  (declare (ignore args))
  (setf (particles pf) (make-array num-particles)
	(new-particles pf) (make-array num-particles)
	(weights pf) (make-array num-particles))
  (reset-filter pf))

(defmethod reset-filter ((pf <particle-filter>))
  (let ((init (init-dist (ssm pf)))
	(particles (particles pf)))
    (dotimes (i (num-particles pf))
      (setf (aref particles i) (sample init)))
    (setf (first-step? pf) t)
    (debug-print 3 "Reset particle filter.  Particles are ~a" particles)))


(defmethod incorporate-evidence ((pf <particle-filter>) o)
  (if (first-step? pf)
      (setf (first-step? pf) nil)
    (step-particles-forward pf (particles pf) o))
  (weight-particles pf o (weights pf))
  (funcall (ecase (resampling-strategy pf)
	     (:simple #'simple-resample)
	     (:correlated #'correlated-resample))
	   pf (new-particles pf))
  (when-debugging 3
    (debug-print 0 "Incorporating evidence ~a.~&Weighted particles ~a~&Resampled particles: ~a" 
		 o (map 'list #'cons (particles pf) (weights pf)) (new-particles pf)))
  (copy-new-particles pf))

(defmethod current-state-dist ((pf <particle-filter>))
  (make-instance '<empirical-dist> :samples (particles pf)))
  

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Internal
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun step-particles-forward (pf v o)
  (let ((trans (sampling-dist pf o)))
    (map-into v #'(lambda (x) (cond-sample trans x)) (particles pf))))

(defun sampling-dist (pf o)
  (let ((trans (trans-dist (ssm pf)))
	(obs (obs-dist (ssm pf))))
    (ecase (trans-strategy pf)
      (:prior trans)
      (:posterior #'(lambda (x) (condition-on-dist (cond-dist trans x) obs o))))))


(defun weight-particles (pf obs w &aux (obs-dist (obs-dist (ssm pf))))
  (ecase (trans-strategy pf)
    (:prior (let ((total 0))
	      (map-into w
			#'(lambda (particle)
			    (let ((p (cond-prob obs-dist obs particle)))
			      (incf total p)
			      p))
			(particles pf))
	      (assert (> total 0) nil "Particle weights ~a for ~a had nonpositive sum ~a"
		      w (particles pf) total)
	      (map-into w #'(lambda (x) (/ x total)) w)))
    (:posterior (fill w (/ 1 (length w))))))

(defun simple-resample (pf new-particles)
  ;; just draw the samples one at a time
  (with-slots (weights particles) pf
    (dotimes (i (num-particles pf))
      (let ((np (sample weights)))
	(setf (aref new-particles i) (aref particles np))))))

(defun correlated-resample (pf new-particles)
  ;; correlated resample: a bit more efficient and lower variance
  (let ((np (num-particles pf))
	(particles (particles pf))
	(weights (weights pf))
	(next 0))
    (map nil
      #'(lambda (p w) (fill new-particles p :start next :end (incf next (floor (* np w)))))
      particles weights)
    (for-loop (i next np)
      (setf (aref new-particles i) (aref particles (sample weights))))))
	      

(defun copy-new-particles (pf)
  (rotatef (particles pf) (new-particles pf)))

            
    
  
	    
      
  
   
   