(defpackage ssm
  (:documentation "Package ssm

Creation
--------
<state-space-model>
make-hmm
make-lds

Operations on ssms
------------------
init-dist
trans-dist
obs-dist
trajectory-dist
state-prefix
obs-prefix
sample-finite-trajectory
trajectory-states
trajectory-observations
same-state

Filtering operations 
--------------------
reset-filter
incorporate-evidence
current-state-dist
filter-observations
incorporate-evidence-sequence
")
  (:export
   <state-space-model>
   make-hmm
   make-lds
   
   trajectory-dist
   init-dist
   trans-dist
   obs-dist
   state-prefix
   obs-prefix
   sample-finite-trajectory
   trajectory-states
   trajectory-observations
   same-state
   
   reset-filter
   incorporate-evidence
   current-state-dist
   filter-observations
   incorporate-evidence-sequence

   )
  (:use
   cl
   utils
   set
   mapping
   prob))

(in-package ssm)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; general ssm
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <state-space-model> (<prob-dist>)
  ((init-dist :initarg :init-dist :reader init-dist :type [prob-dist])
   (trans-dist :initarg :trans-dist :reader trans-dist :type [cond-prob-dist])
   (obs-dist :initarg :obs-dist :reader obs-dist :type [cond-prob-dist]))
  (:documentation "Class <state-space-model>

Represents a general state-space model consisting of (hidden) states and outputs (observations).  The required initargs are:
:init-dist - of type [prob-dist]
:trans-dist - of type [cond-prob-dist]
:obs-dist - of type [cond-prob-dist]

A generative description of the distribution is that the initial state s_0 is sampled from init-dist, then o_0 is sampled from obs-dist conditional on s_0, then s_1 is sampled from trans-dist conditioned on s_0, and so on.

The fact that a state-space-model inherits from <prob-dist> means that the standard operations should in principle be allowed.  For now, sample will return a random (infinitely long) ssm trajectory.  An ssm trajectory is represented as a pair consisting of a state function and observation function.  Each function takes nonnegative integers.  For example, calling the obs function on 0 returns o_0.    The evaluation is lazy and randomized, but stable.  In other words, if you ask it a question, its answer will correctly obey the conditional distribution given previous answers to any questions you asked it.  As a special case, if you ask the same question twice, you will get the same answer.  Prob and expectation are not implemented at the moment."))

		  

(defmethod sample ((ssm <state-space-model>) &aux (unassigned (gensym)))
  (let ((states (make-array 0 :adjustable t :fill-pointer 0))
	(observations (make-array 0 :adjustable t :fill-pointer 0))
	(trans (trans-dist ssm))
	(obs (obs-dist ssm)))
    (vector-push-extend (sample (init-dist ssm)) states)
    (labels ((state (i)
	       ;; do any sampling necessary
	       (for-loop (k (length states) (1+ i))
		 (vector-push-extend (cond-sample trans (aref states (1- k))) states))
	       
	       ;; return
	       (aref states i))
	     
	     (obs (i)
	       ;; grow the array as necessary
	       (for-loop (k (length observations) (1+ i))
		 (vector-push-extend unassigned observations))
	       
	       ;; sample if necessary
	       (when (eql (aref observations i) unassigned)
		 (setf (aref observations i) (cond-sample obs (state i))))
	       
	       ;; return
	       (aref observations i)))
      
      (cons #'state #'obs))))
		 
	     

;; TODO just have a defstruct for traj
(defun trajectory-observations (traj)
  (cdr traj))

(defun trajectory-states (traj)
  (car traj))

(defun obs-prefix (traj n)
  "observation-prefix SSM-TRAJECTORY N.  Return the vector consisting of observations 0:N-1."
  (let ((obs (trajectory-observations traj)))
    (mapset 'vector #'(lambda (i) (evaluate obs i)) n)))

(defun state-prefix (traj n)
  (let ((states (trajectory-states traj)))
    (mapset 'vector #'(lambda (i) (evaluate states i)) n)))

(defun sample-finite-trajectory (ssm n)
  "sample-finite-trajectory SSM N.  Return two values STATES and OBSERVATIONS, each of length N."
  (let ((tr (sample ssm)))
    (values (state-prefix tr n) (obs-prefix tr n))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Special cases
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun make-hmm (init trans obs &key (states (length init)) (observations (array-dimension obs 1)))
  "make-hmm INIT TRANS OBS &key STATES OBSERVATIONS

INIT - an n-d vector where INIT(i) = prob of starting at state i
TRANS - an nxn array where TRANS(i,j) = prob of going to state j from state i
OBS - an nxm array where TRANS(i,j) = prob of generating obs j from state i

STATES and OBSERVATIONS should be sets.  They default to 0:n-1 and 0:m-1 respectively."
  (let ((n (size states))
	(m (size observations)))
    (make-instance '<state-space-model>
      :init-dist (make-multinomial-dist init states)
      :trans-dist #'(lambda (s) (make-multinomial-dist 
				 (make-array n :displaced-to trans :displaced-index-offset (* (item-number s states) n))
				 states))
      :obs-dist #'(lambda (s) (make-multinomial-dist
			       (make-array m :displaced-to obs :displaced-index-offset (* (item-number s states) m))
			       observations)))))



(defun make-lds (&key init-mean init-cov trans trans-bias trans-cov obs obs-bias obs-cov)
  "make-lds INIT-MEAN INIT-COV TRANS-BIAS TRANS-COV OBS-BIAS OBS-COV

Return an object of type <state-space-model> that represents linear dynamical system with the given parameters."
  
  (make-instance '<state-space-model>
    :init-dist (make-instance '<gaussian> :mean init-mean :cov init-cov)
    :trans-dist (make-instance '<linear-gaussian> :trans trans :bias trans-bias :cov trans-cov)
    :obs-dist (make-instance '<linear-gaussian> :trans obs :bias obs-bias :cov obs-cov)))
  
  





;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; printing
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod print-object ((s <state-space-model>) str)
  (print-unreadable-object (s str :type t :identity nil)))

(defun pprint-ssm (&rest args)
  (bind-pprint-args 
   (str ssm) args
   (write ssm :stream str :pretty nil)))

(set-pprint-dispatch '<state-space-model> #'pprint-ssm 1)