(defpackage advantage-function
  (:documentation "Package advantage-function (adv-fn)

Types
-----
<advantage-function>
<tabular-advantage-function>
<advantage-maximizing-policy>
<sum-advantage-function>
sum-advantage-functions

Operations
----------
advantage
max-advantage
best-choice
update
normalize-state
evaluate-comps
adv-components
best-advantages
pprint-advantage-fn 

Conditions
----------
unknown-state
unknown-action
unknown-state-action

Tabular-advantage-functions
---------------------------
add-to-table (restart function)
")
  (:export
   
   <advantage-function>
   <tabular-advantage-function>
   <advantage-maximizing-policy>
   <sum-advantage-function>
   sum-advantage-functions
   
   advantage
   max-advantage
   best-choice
   update
   normalize-state
   evaluate-comps
   adv-components
   best-advantages
   pprint-advantage-fn

   
   unknown-state
   unknown-state-action
   unknown-action
   
   add-to-table
   )
  
  (:use
   cl
   utils
   set)
  
  (:nicknames adv-fn)
  )


(in-package advantage-function)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; advantage function class
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <advantage-function> ()
  ((avail-actions-function :type function :initarg :avail-actions-fn :reader avail-actions-fn :writer set-avail-actions-fn)
   (avail-actions-specifier :initarg :avail-actions-specifier :accessor avail-actions-specifier))
  (:documentation "Class <advantage-function>

Instances of this class should not be created.  Instead, create an instance of one of its children.

Initargs for this class and its subclasses must be according to one of two cases.

Case 1
:avail-actions-specifier - Must be one of 1) A function of one argument that is applied to a state and returns the set of available actions 2) An object of type <env>, whose avail-actions method is then used
plus any initargs for the subclass

Case 2
:clone-from - must be of type <advantage-function>
"))

(defmethod initialize-instance :after ((adv <advantage-function>) &rest args &key avail-actions-specifier (clone-from nil))
  (declare (ignore args))
  (let ((spec (aif clone-from (avail-actions-specifier it) avail-actions-specifier)))
    (when clone-from
      (setf (avail-actions-specifier adv) spec))
    (set-avail-actions-fn 
     (etypecase spec
       (function spec)
       (env-user:<env> #'(lambda (s) (env-user:avail-actions spec s))))
     adv)))

(defmethod clone ((adv <advantage-function>))
  (make-instance (class-of adv) :clone-from adv))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; condition types
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-condition unknown-state (error)
  ((adv-fn :initarg :adv-fn :reader adv-fn)
   (state :initarg :state :reader state))
  (:report (lambda (c s) (format s "Unknown state ~a for advantage function ~a." (state c) (adv-fn c)))))

(define-condition unknown-action (error)
  ((adv-fn :initarg :adv-fn :reader adv-fn)
   (action :initarg :action :reader action))
  (:report (lambda (c s) (format s "Unknown action ~a for advantage function ~a." (action c) (adv-fn c)))))

(define-condition unknown-state-action (error)
  ((adv-fn :initarg :adv-fn :reader adv-fn)
   (state :initarg :state :reader state)
   (action :initarg :action :reader action))
  (:report (lambda (c s) (format s "Unknown state-action pair (~a, ~a) for advantage function ~a." (state c) (action c) (adv-fn c)))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; functions for use by subclasses
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun avail-actions (adv-fn s)
  (funcall (avail-actions-fn adv-fn) s))

(defun best-choice (adv-fn s)
  (mvbind (ind v best)
      (argmax (avail-actions adv-fn s)
	      :key #'(lambda (a)
		       (handler-case
			   (advantage adv-fn s a)
			 (unknown-action () '-infty))))
    (declare (ignore ind v))
    best))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; exported operations on advantage functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric advantage (adv-fn s a)
  (:documentation "advantage ADVANTAGE-FUNCTION STATE ACTION.  Advantage of doing this action in this state.  For the optimal advantage function, this should be 0 if a is optimal, negative otherwise.  May signal an unknown-state-action, unknown-state, or unknown-action error.  A use-value restart is provided that allows a default value to be used.")
  (:method :around (adv-fn s a)
	   (declare (ignore adv-fn s a))
	   (restart-case
	       (call-next-method)
	     (use-value (x) x))))
	       

(defgeneric max-advantage (adv-fn s)
  (:documentation "max-advantage ADVANTAGE-FUNCTION STATE.  Maximum advantage of any action in this STATE.  If advantages are normalized, this must be zero, but at intermediate points during learning, this may not be the case.  Unknown actions are treated as having advantage -infty.  If the state is unknown, the max advantage is assumed to be 0.")
  (:method (adv-fn s)
	   (handler-case
	       (let ((max '-infty))
		 (do-elements (a (avail-actions adv-fn s) 
				 (if (eq max '-infty) 0.0 max))
		   (_f mymax max 
		       (handler-case
			   (advantage adv-fn s a)
			 (unknown-action () '-infty)))))
	     (unknown-state () 0.0))))

(defgeneric update (adv-fn s a target eta)
  (:documentation "update ADVANTAGE-FUNCTION STATE ACTION TARGET ETA.  Make advantage of action in state be more like target.  If the state or action is unknown, proceed as if old advantage had been 0."))

(defgeneric normalize-state (adv-fn s &optional d)
  (:documentation "normalize-state ADV-FN STATE &optional DIFF.  Subtract DIFF (which defaults to the maximum advantage at S) from the advantage of each action at S.  May be inexact due to function approximation."))


(defmethod policy:print-advice ((adv <advantage-function>) omega choices str)
  (handler-bind (((or unknown-state unknown-action)
		  #'(lambda (c)
		      (use-value (class-of c)))))
    (pprint-logical-block (str nil)
      (format str "~:@_Advantages : ~a~:@_Recommendation : ~a"
	      (let ((advs (make-array 0 :fill-pointer 0 :adjustable t)))
	    
		(set:do-elements (u choices advs)
		  (vector-push-extend (cons u 
					    (let ((val
						   (advantage adv omega u)))
					      (if (numberp val)
						  (round-decimal val 2)
						val)))
				      advs)))
	      (maximizing-element choices #'(lambda (u) (advantage adv omega u)))))))

(defun pprint-advantage-fn (&rest args)
  "pprint-advantage-fn [STREAM t] ADV-FN"
  (bind-pprint-args 
   (str f) args
   (pprint-advantages str f)))

(defgeneric pprint-advantages (str f)
  (:documentation "pprint-advantages STREAM ADV-FN"))
		    


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; greedy policies of advantage functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <advantage-maximizing-policy> (policy:<policy>)
  ((adv-fn :reader adv-fn :initarg :adv-fn)
   (unknown :accessor unknown :initarg :unknown-state-behaviour :initform ':choose-randomly))
  (:documentation "<advantage-maximizing-policy>
Initargs
:adv-fn - advantage function to maximize
:unknown-state-behaviour - one of ':choose-randomly or ':signal-error.  Defaults to ':choose-randomly.  If it's :signal-error, may signal a 'policy:unknown-state error.
 
If the state is known but some action is unknown, its advantage is assumed to be -infty."))

(defmethod policy:make-choice ((pol <advantage-maximizing-policy>) s &optional choices &aux (adv-fn (adv-fn pol)))
  (declare (ignore choices))
  (handler-case

      ;; main form: maximize advantage
      (best-choice adv-fn s)
      
    ;; behaviour at unknown states
    (unknown-state ()
      (ecase (unknown pol)
	(:signal-error (error 'policy:unknown-state :state s :policy pol))
	(:choose-randomly (prob:sample-uniformly (avail-actions adv-fn s)))))))
	
(defmethod clone ((pol <advantage-maximizing-policy>))
  (make-instance '<advantage-maximizing-policy> :adv-fn (clone (adv-fn pol))))
  
  
