(in-package advantage-function)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; class def
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defclass <sum-advantage-function> (<advantage-function>)
  ((advantage-functions :type <numbered-set> :accessor adv-fns :initarg :adv-fns))
  (:documentation "Class <sum-advantage-function> (<advantage-function>)

Initargs

Case 1
:adv-fns - <numbered-set> of <advantage-function>s, which must have the same set of available actions at any state.

Case 2
:clone-from - object of type <sum-advantage-function>
"))

(defmethod initialize-instance :after ((adv <sum-advantage-function>) &rest args &key (clone-from nil))
  (declare (ignore args))
  (awhen clone-from
    (setf (adv-fns adv) (clone (adv-fns it))))
  (assert (not (is-empty (adv-fns adv)))))

(defun sum-advantage-functions (&rest adv-fns)
  "sum-advantage-functions &rest ADV-FNS.  Return a <sum-advantage-function> that is the sum of ADV-FNs.  Assumes they all have the same set of available actions at each state."
  (assert adv-fns)
  (make-instance '<sum-advantage-function> :adv-fns adv-fns 
		 :avail-actions-specifier (avail-actions-specifier (first adv-fns))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; general operations for advantage functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defmethod advantage ((adv-fn <sum-advantage-function>) s a)
  "Evaluate just sums componentwise"
  (sum-over (adv-fns adv-fn)
	    #'(lambda (f)
		(advantage f s a))))


;; max-advantage need not be overridden from <advantage-function>
;; update and normalize-state can't be called on sum-advantage-functions
(defmethod update ((adv-fn <sum-advantage-function>) s a target eta)
  (declare (ignore s a target eta))
  (error "Can't call update on ~a" adv-fn))

(defmethod normalize-state ((adv-fn <sum-advantage-function>) s &optional best)
  (declare (ignore s best))
  (error "Can't call normalize-state on ~a" adv-fn))



(defmethod pprint-advantages (str (adv-fn <sum-advantage-function>))
  (pprint-logical-block (str nil :prefix "[" :suffix "]")
    (format str "Sum advantage function~2I")
    (do-elements (f (adv-fns adv-fn) nil i)
      (format str "~:@_Component ~a~4I~:@_" i)
      (pprint-advantages str f))))
	      

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; operations specific to sums
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun evaluate-comps (adv-fn s a)
  "evaluate-comps SUM-ADV-FN S A.  Return list of advantages of each component of SUM-ADV-FN at S,A."
  (mapset 'list #'(lambda (f) (advantage f s a)) (adv-fns adv-fn)))

(defun adv-components (f)
  "adv-components SUM-ADVANTAGE-FUNCTION.  Return the set of components."
  (adv-fns f))

(defun num-components (f)
  (length (adv-fns f)))

(defun best-advantages (adv-fn s)
  "best-advantages SUM-ADV-FN S.  Return advantages of best action at S."
  (handler-case
      (evaluate-comps adv-fn s (best-choice adv-fn s))
    (unknown-state ()
      (make-list (num-components adv-fn) :initial-element 0.0))))


(defparameter *sum-adv-fn-field-width* 10)

(defmethod policy:print-advice ((adv <sum-advantage-function>) omega choices str &aux (adv-fns (adv-fns adv)) (n (length adv-fns)))
  (flet ((make-tabs (s) (mapset 'list #'(lambda (x) (list *sum-adv-fn-field-width* x)) s)))
    (pprint-logical-block (str nil)
      (format str "Advantages are~:@_~:@_ ~:{~0,v:@T  ~a~}" (make-tabs n))
      (format str "~0,v:@TTotal" *sum-adv-fn-field-width*)
    
      (unwind-protect 
	  (do-elements (u choices)
	    (pprint-pop)
	    (format str "~:@_~W~:{~0,v:@T ~5,2f~}" u (make-tabs (append (mapset 'list #'(lambda (f) (advantage f omega u)) adv-fns)
								      (list (advantage adv omega u))))))
	(format str "~:@_Recommendation : ~a" (maximizing-element choices #'(lambda (u) (advantage adv omega u))))))))
    

      