(in-package advantage-function)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; class
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <tabular-advantage-function> (<advantage-function>)
  ((states :accessor states)
   (actions :accessor actions)
   (state-featurizer :accessor state-featurizer :initarg :state-featurizer :initform #'identity)
   (action-featurizer :accessor action-featurizer :initarg :action-featurizer :initform #'identity)
   (table :accessor table)
   (test :accessor test :initarg :test))
  (:documentation "<tabular-advantage-function> (<advantage-function>)

Initargs:

Case 1
:test - test used in the hash tables that store states and actions.  #'equal by default.
:state-featurizer - function that maps a state into the object which is actually stored in the table.  Used to specify what the relevant features of the state are.  Defaults to #'identity
:action-featurizer - likewise for actions.

Case 2
:clone-from - of type <tabular-advantage-function>
"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; 
;; constructor
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod initialize-instance :after ((adv-fn <tabular-advantage-function>) &rest args &key (test #'equal) (clone-from nil))
  (declare (ignore args))
  (condlet
   
   ;; Bindings for creating object from another one
   ((clone-from (states (clone (states clone-from))) 
		(actions (clone (actions clone-from))) 
		(table (clone (table clone-from))) 
		(actual-test (test clone-from))
		(state-featurizer (state-featurizer clone-from))
		(action-featurizer (action-featurizer clone-from)))
    
    ;; Bindings for creating object from scratch
    (t (states (make-hash-table :test test))
       (actions (make-hash-table :test test))
       (table (make-array '(0 0) :element-type '(or symbol float) :initial-element nil))
       (actual-test test)
       (state-featurizer (state-featurizer adv-fn))
       (action-featurizer (action-featurizer adv-fn))))
   
   ;; Set the slots
   (setf (states adv-fn) states
	 (actions adv-fn) actions
	 (table adv-fn) table
	 (state-featurizer adv-fn) state-featurizer
	 (action-featurizer adv-fn) action-featurizer
	 (test adv-fn) actual-test)))
  

   
   
   
   


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; looking up states and actions in the table
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun lookup-state (adv-fn s &aux (states (states adv-fn)))
  (let ((features (funcall (state-featurizer adv-fn) s)))
    (restart-case
      
	;; by default, if the item is not present, signal error
	(mvbind (ind present?) 
	    (gethash features states)
	  (if present? ind
	    (error 'unknown-state :adv-fn adv-fn :state s)))
    
      ;; a restart is provided that allows callers to add the new state to the table instaed
      (add-to-table ()
	(let ((n (array-dimension (table adv-fn) 0))
	      (m (array-dimension (table adv-fn) 1))
	      (k (hash-table-count states)))
	  (when (= k n)
	    (adjustf (table adv-fn) (list (1+ (* 2 n)) m) :initial-element nil :element-type '(or symbol float)))
	  (setf (gethash features states) k)))
      )))

(defun lookup-action (adv-fn a &aux (features (funcall (action-featurizer adv-fn) a)) (actions (actions adv-fn)))
  (restart-case
      (handler-case
	  (lookup-action-features adv-fn features)
	(unknown-action ()
	  (error 'unknown-action :adv-fn adv-fn :action a)))

    ;; a restart is provided that allows callers to add the new action to the table instaed
    (add-to-table ()
      (let ((n (array-dimension (table adv-fn) 0))
	    (m (array-dimension (table adv-fn) 1))
	    (k (hash-table-count actions)))
	(when (= k m)
	  (adjustf (table adv-fn) (list n (1+ (* 2 m))) :initial-element nil :element-type '(or symbol float)))
	(setf (gethash features actions) k)))))

(defun lookup-action-features (adv-fn features &aux (actions (actions adv-fn)))
  (mvbind (ind present?) 
      (gethash features actions)
    (if present? ind
      (error 'unknown-action :adv-fn adv-fn :action features))))


(defun add-to-table (c)
  "Can be used as a handler by code that uses tabular advantage functions.  It causes an unknown state or action to be added to the table."
  (declare (ignore c))
  (invoke-restart 'add-to-table))


	 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; exported operations
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defmethod advantage ((adv-fn <tabular-advantage-function>) s a)
  (let ((i (lookup-state adv-fn s))
	(j (lookup-action adv-fn a)))
    (or (aref (table adv-fn) i j) (error 'unknown-action :adv-fn adv-fn :action a))))
   

(defmethod normalize-state ((adv-fn <tabular-advantage-function>) s &optional (m (max-advantage adv-fn s)))
  (let ((i (lookup-state adv-fn s))
	(table (table adv-fn)))
    (maphash 
     #'(lambda (k v)
	 (declare (ignore k))
	 (when (aref table i v) (decf (aref table i v) m)))
     (actions adv-fn))
    ))

(defmethod update ((adv-fn <tabular-advantage-function>) s a target eta)
  ;; if we see unknown states or actions, we just add them to the table
  (handler-bind 
      (((or unknown-state unknown-action) #'add-to-table))
    
    (let ((i (lookup-state adv-fn s))
	  (j (lookup-action adv-fn a)))
      (symbol-macrolet
	  ((advantage (aref (table adv-fn) i j)))
	(if advantage
	    (avgf advantage target eta)
	  (setf advantage (* target eta)))))))
  

      
(defmethod pprint-advantages (str (f <tabular-advantage-function>) &aux (table (table f)))
  (pprint-logical-block (str nil :prefix "[" :suffix "]")
    (format str "Tabular advantage function~2I")
    (do-elements (s (states f) nil i)
      (format str "~:@_State ~a~4I" s)
      (do-elements (a (actions f) nil j)
	(format str "~:@_~a : ~a" a (aref table i j)))
      (format str "~2I"))))
      





