;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; rl-functions/policy/prompt-policy.lisp
;; TODO : this class is getting unwieldy
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package policy)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; class def, constructor
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <prompt-policy> (<policy>)
  ((advisors :type list :reader advisors :initarg :advisors :initform nil)
   (choice-fn :type function :reader choice-fn :initarg :choice-fn)
   (print-prefix :reader print-prefix :initarg :print-prefix :initform t)
   (io-specifier :initarg :io-specifier :reader io-specifier :initform nil))
  
  (:documentation "A <prompt-policy> is one which makes choices by asking the user.  It also has a list of 'advisors', which can print their advice before the user enters a choice.

Initargs
:choice-fn - Function that is called on a state and returns set of available choices
:advisors - list of advisors whose advice will be printed at each choice (see print-advice).  nil by default.
:print-prefix - If this equals ':state, then the state, available choices and advisor output is printed to standard output before prompting for a choice.  If it is any other true value, then just the available choices and advisor output are printed.  
:io-specifier - An object that governs how the IO is done when asking the user to enter a choice.  Nil by default, which amounts to just printing the state and set of choices, and using the Lisp reader to interpret what the user enters as an action.  This will work fine in almost all situations.  Modifying this behaviour requires providing a method for prompt-for-choice"))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; special variables
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *max-num-choices-to-print* 10)
(defparameter *print-sparse-feature-vectors* t "If true (default), feature vectors are printed sparsely by the prompt policy.")



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; implemented methods
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod make-choice ((pol <prompt-policy>) s &optional choices)
  (declare (ignore choices))
  (let ((choices (funcall (choice-fn pol) s))
	(str t))
    (when (print-prefix pol)
      (format str "~&~%")
      (pprint-logical-block (str nil)
	(format str "Prompt policy")
	(when (eq (print-prefix pol) ':state)
	      (format str "~5I~:@_State:~15T~W" s))
	(format str "~5I~:@_Choice set:~15T~W~:@_~50@{~A~:*~}" choices #\-)
	(set:do-elements (a (advisors pol) nil i)
	  (format str "~:@_Advisor ~W:~5I~:@_" i)
	  (print-advice a s choices str)
	  (format str "~:@_~50@{~A~:*~}" #\-))))
    
    (prompt-for-choice s choices (io-specifier pol))
    ))

(defgeneric prompt-for-choice (s choices io-spec)
  (:documentation "prompt-for-choice STATE CHOICES IO-SPEC.  A way to customize the IO behaviour of prompt for choice.  Methods for this generic function are expected to prompt the user for a choice, parse the entered choice, and return it.  The process may involve a back-and-forth interaction with the user.  The method may also signal a 'choose-to-abort error if the user enters some value that signifies that the io-interface should finish.  

The default method 
- asks the user to enter the choice
- parses it using the lisp read function
- If it equals nil, aborts
- If it's not a member of the choice set, and if the set of choices has size 1, then prints a warning message and returns that single member.
- If it's not a member of the choice set, and if the set of choices has size greater than 1, asserts and allows the user to enter another choice.
- Finally returns the choice.

IO-SPEC receives standard values in various settings:
- rl:io-interface uses a prompt policy with io-spec being the environment
- alisp:io-interface and calisp:io-interface use a prompt policy with io-spec being the partial program.

")
  (:method (s choices io-spec)
	   (declare (ignore s io-spec))
	   (format t "~&Please enter choice, or nil to terminate. ")
	   ;; if one of the choices is nil, you're out of luck...
	   (let ((ch (read))
		 is-member reason)
	     (assert
		 (progn
		   (unless ch
		     (error 'choose-to-abort))
		   (if (eql (set:size choices) 1)
		       (progn
			 (unless (set:member? ch choices)
			   (warn "~a is not a valid choice, but using default value ~a instead."
				 ch (set:item 0 choices))
			   (setf ch (set:item 0 choices)))
			 (setf is-member t))
		     (multiple-value-setq (is-member reason)
		       (set:member? ch choices)))
		   is-member)
		 (ch)
	       (set:get-full-explanation-string reason))
	     ch)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; advice generic function and some common methods
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric print-advice (advisor omega choices str)
  (:documentation "print-advice ADVISOR STATE CHOICES STREAM.  Given that we are at STATE and have choices CHOICES, print ADVISOR's opinion to STREAM."))

(defmethod print-advice ((adv <policy>) omega choices str)
  (declare (ignore choices))
  (format str "Policy recommends ~a" (make-choice adv omega)))




(defmethod print-advice ((adv mdp:<mdp>) omega choices str)
  ;; a mdp prints its transition distribution
  (format t "Transition distributions are ")
  (set:do-elements (u choices)
    (format str "~&~a" (mdp:trans-dist adv omega u))))


(defmethod print-advice ((adv function) omega choices str)
  (let ((l (get-lambda-list adv)))
    (ecase (length l)
      (2 (print-state-action-feature-advice adv omega choices str))
      (1 (let ((result (funcall adv omega)))
	   (if (set:member? result choices)
	       ;; treat adv as a policy
	       (format str "Policy recommends ~a" result)
	     (if (typep result '[cond-prob-dist])
		 (print-cond-dist-advice adv omega choices str)
	       (format str "~a" result))))))))

(defun print-cond-dist-advice (adv omega choices str)
  (format str "Function of one argument being treated as a conditional distribution by prompt-policy.")
  (set:do-elements (u choices)
    (format str "~&When making choice ~a" u)
    (prob:do-sample-points (omega2 prob (cond-dist adv (cons omega u)))
      (format str "~&Item ~a with probability ~a" omega2 prob))))


(defun print-state-action-feature-advice (adv omega choices str)
  (pprint-logical-block (str nil)
    (format str "Choice~30:TFeatures")
    (set:do-elements (u choices)
      (pprint-pop)
      (format str "~:@_~W~30:T~W" u (funcall (if *print-sparse-feature-vectors* #'sparsify #'identity) (funcall adv omega u))))))
		