;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; rl-functions/value/tabular-value-function.lisp
;; Defines the <tabular-value-fn> class
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package value-fn)


(defclass <tabular-value-fn> (<value-fn>)
  ((vals :initarg :vals
	 :accessor vals
	 :type (array float 1))
   (featurizer :initarg :featurizer
	       :initform #'identity
	       :accessor featurizer)	       
   (state-set :initarg :state-set
	      :accessor state-set
	      :initform nil
	      :type set:[numbered-set]
	      ))
  (:documentation "A <tabular-value-fn> consists of a vector VALS of values, and a numbered set S.  The value of X is gotten by looking up the number of X in S, and getting the corresponding item from the vector VALS.

Initargs
:STATE-SET - the state set.  If this is not given, starts out being the empty set.
:VALS - the initial values.  If not given, defaults to a vector of the same length as STATE-SET, all of whose entries are uninitialized.
:FEATURIZER - function of state that returns the key used to look up the state.  Defaults to #'identity."))


(defun make-tabular-value-fn (v &optional (s (length v)))
  "make-tabular-value-fn VALS &optional STATE-SET.
Makes a <tabular-value-fn> having the given values.  If SET is not given, it is assumed to be the set 0,1,...,N-1 where N is the length of VALS."
  (make-instance '<tabular-value-fn> :vals v :state-set s))

(defmethod initialize-instance :after ((v <tabular-value-fn>) &rest args &key (clone-from nil))
    (declare (ignore args))
    (aif clone-from
       (setf (state-set v) (state-set it)
	     (vals  v) (clone (vals it))
	     (featurizer v) (featurizer it))
			      
       (set-if-unbound 'vals v (make-array (size (state-set v))))))


(defun lookup (v s &aux (featurized-state (funcall (featurizer v) s)))
  (symbol-macrolet
      ((states (state-set v))
       (table (vals v)))
    
    (restart-case
      
	;; by default, if the item is not present, signal error
	(handler-case
	    (item-number featurized-state states)
	  (item-not-in-set ()
	    (error 'unknown-state :value-fn v :state featurized-state)))
    
      ;; a restart is provided that allows callers to add the new state to the table instaed
      (add-to-table ()
	(let ((n (array-dimension table 0))
	      (k (set:size states)))
	  (when (= n k)
	    (adjustf table (list (1+ (* 2 n))) :initial-element nil :element-type '(or symbol float)))
	  (addf states featurized-state k)
	  k)))))

(defun add-to-table (c)
  (declare (ignore c))
  (invoke-restart 'add-to-table))

(defmethod value ((v <tabular-value-fn>) x)
  (let ((i (lookup v x)))
    (or (aref (vals v) i) 0.0)))

(defmethod update-value ((v <tabular-value-fn>) s target eta)
  (handler-bind
      ((unknown-state #'add-to-table))
    (let ((i (lookup v s)))
      (symbol-macrolet 
	  ((val (aref (vals v) i)))
	
	(if val
	    (avgf val target eta)
	    
	  ;; if value is nil, ignore step size and set it equal to target
	  (setf val target)
	  )))))



(defmethod print-object ((v <tabular-value-fn>) str)
  (format str "Tabular value function on set ~a with values ~a" (state-set v) (vals v)))

(defmethod pprint-values (str  (v <tabular-value-fn>) &aux (vals (vals v)))
  (pprint-logical-block (str nil :prefix "[" :suffix "]")
    (format str "Tabular value function~2I")
    (set:do-elements (x (state-set v) nil i)
      (pprint-pop)
      (format str "~:@_~W~10@T~@_~W" x (aref vals i)))))

(set-pprint-dispatch '<tabular-value-fn> #'pprint-values)

(defmethod lin-alg:lp-dist ((v <tabular-value-fn>) (v2 <tabular-value-fn>) &optional (p 1))
  "TODO Should do error checking to see if state sets are equal."
  (lin-alg:lp-dist (vals v) (vals v2) p))