(defpackage value-function
  (:documentation "Package for value functions.

Types
-----
<value-fn>
<tabular-value-fn>
<sum-value-fn>

Constructors
------------
make-tabular-value-fn

Operations
----------
value
pprint-value-fn
value-components
update-value

Conditions, restarts
--------------------
unknown-state
add-to-table
")
  (:nicknames value-fn)
  (:use common-lisp 
	utils
	set)
  (:import-from policy
		unknown-state)
  (:export 
   <value-fn>
   <tabular-value-fn>
   <sum-value-fn>
   make-tabular-value-fn
   value
   pprint-value-fn
   value-components
   update-value
   unknown-state
   add-to-table))

(in-package value-function)


(defclass <value-fn> () 
  ()
  (:documentation "Class for value functions.  Only operation is value.  Note that <q-function> is a subclass of this."))


(defgeneric value (val-fn omega)
  (:documentation "value VALUE-FUNCTION STATE.  Return optimal value of this state.  Might signal an unknown-state error.")
  (:method :around (val-fn omega)
	   (declare (ignore val-fn))
	   (restart-case (call-next-method)
	     (use-value (val)
		 :report "Provide a value to use this time as the value."
		 :interactive (lambda () 
				(format t "~&Enter a value for state ~a.  " omega)
				(list (read)))
	       val))))


(defmethod policy:print-advice ((v <value-fn>) omega choices str)
  (declare (ignore choices))
  (format str "Value of state is ~a" (value v omega)))

(defgeneric update-value (val-fn s target eta)
  (:documentation "update-value VALUE-FUNCTION STATE TARGET ETA.  Make value at STATE be closer to TARGET by a stepsize of ETA."))

(defun pprint-value-fn (&rest args)
  (bind-pprint-args 
   (str v) args
   (pprint-values str v)))

(defgeneric pprint-values (str v)
  (:documentation "pprint-values STREAM VALUE-FUNCTION."))

(defmethod clone ((f <value-fn>))
  (make-instance (class-of f) :clone-from f))

(defparameter *sum-value-fn-advice-field-width* 11)

      
    
      
    


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; sum-value-function
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <sum-value-fn> (<value-fn>)
  ((value-fns :accessor value-fns :initarg :value-fns))
  (:documentation "Class <sum-value-fn>

Initargs
:value-fns - set of value functions")
  )

(defmethod initialize-instance :after ((v <sum-value-fn>) &rest args &key (clone-from nil))
  (declare (ignore args))
  (awhen clone-from
    (setf (value-fns v) (clone (value-fns it))))
  (assert (not (is-empty (value-fns v)))))

    

(defmethod value ((v <sum-value-fn>) s)
  (sum-over (value-fns v) #'(lambda (f) (value f s))))

(defun value-components (v)
  "value-components SUM-VALUE-FUNCTION.  Return the set of component value functions."
  (value-fns v))



(defmethod pprint-values (str (v <sum-value-fn>))
  (pprint-logical-block (str nil :prefix "[" :suffix "]")
    (format str "Sum of value functions~2I")
    (do-elements (f (value-fns v) nil i)
      (pprint-pop)
      (format str "~:@_Component ~a~4I~:@_" i)
      (pprint-values str f)
      (format str "~2I")))
  )

(defmethod policy:print-advice ((v <sum-value-fn>) omega choices str &aux (n (length (value-fns v))))
  (declare (ignore choices))
  (flet ((make-tabs (s) (mapset 'list #'(lambda (i) (list *sum-value-fn-advice-field-width* i)) s)))
    (pprint-logical-block (str nil)
      
      ;; Column headings
      (format str "~:@_Component~:{~0,v:@T   ~a~}" (make-tabs n))
      (format str "~0,v:@TTotal" *sum-value-fn-advice-field-width*)
      
      ;; Values
      (format str "~:@_Value~:{~0,v:@T~5,2f~}" (make-tabs (append (mapcar #'(lambda (v) (value v omega)) (value-fns v))
							       (list (value v omega))))))))