(defpackage advantage-updating
  (:documentation "Package advantage-updating (adv-up)

Types
-----
<advantage-updating>
<decomposed-advantage-updating>

Operations (apart from general ops on learning algorithms)
----------
set-reward-components
")
  (:export
   <advantage-updating>
   <decomposed-advantage-updating>
   
   set-reward-components)
  (:nicknames adv-up)
  (:use
   cl
   rl-obs
   utils
   adv-fn
   set))

(in-package advantage-updating)

(defclass <advantage-updating> (<advantage-learning-algorithm> <value-learning-algorithm>)
  ((prev-s :accessor prev-s)
   (spec :initarg :avail-actions-specifier :accessor avail-actions-specifier)
   (learning-rate :initarg :lrate :initform .01 :accessor lrate)
   ;; advantage learning typically needs a higher learning rate since adv functions
   ;; have a smaller dynamic range than q-functions
   (advantage-function :accessor adv-fn :initarg :adv-fn)
   (initial-advantage-function :accessor init-adv-fn)
   (initial-value-function :accessor init-value-fn)
   (test :initarg :test :reader test :initform #'equal)
   (value-function :accessor value-fn :initarg :value-fn)
   )
  (:documentation "Class <advantage-updating> (<advantage-learning-algorithm>).

Initargs
:avail-actions-specifier - see argument with same name for <advantage-function>
:test - used for hashtable by tabular advantage function.  Defaults to #'equal.
:lrate - learning rate.  Defaults to .01
:adv-fn - Initial advantage function.  Defaults to a tabular one.
:value-fn - Initial value function.   Defaults to tabular rep."))


(defmethod initialize-instance :around ((alg <advantage-updating>) 
					&rest args &key (adv-fn nil adv-supp)
							(value-fn nil val-supp))
  (apply #'call-next-method args)
  (unless (slot-boundp alg 'initial-advantage-function)
    (setf (init-adv-fn alg)
      (if adv-supp (clone adv-fn)
	(make-instance '<tabular-advantage-function> :test (test alg) :avail-actions-specifier (avail-actions-specifier alg)))
      
      (init-value-fn alg)
      (if val-supp (clone value-fn)
	(make-instance 'value-fn:<tabular-value-fn>)))
    (reset alg)))

(defmethod reset ((alg <advantage-updating>))
  (setf (adv-fn alg) (clone (init-adv-fn alg))
	(value-fn alg) (clone (init-value-fn alg))))
  


(defmethod inform-start-execution ((alg <advantage-updating>))
  )

(defmethod inform-start-episode ((alg <advantage-updating>) s)
  (setf (prev-s alg) s))

    

(defmethod inform-env-step ((alg <advantage-updating>) a r s2 term)
  (let ((s (prev-s alg))
	(adv-fn (adv-fn alg))
	(value-fn (value-fn alg)))
    
    
    (let ((eta (lrate:get-rate (lrate alg) (cons s a)))
	  (max-adv (max-advantage adv-fn s))
	  (value (handler-case (value-fn:value value-fn s) (value-fn:unknown-state ()  0.0)))
	  (next-value (handler-case (value-fn:value value-fn s2) (value-fn:unknown-state ()  0.0))))
      (condlet
       ((term (target r) (prev-s nil))
	(t (target (+ r next-value)) (prev-s s2)))
       
       ;; advantage update
       (update adv-fn s a (+ max-adv (- target value)) eta)
       (when-debugging (format t "~%~%~&Advantage updating.~&Observed ~a ~a -> ~a, ~a, ~a." s a r s2 term))
       
       ;; value function update
       (let ((adv-increase (- (max-advantage adv-fn s) max-adv)))
	 (when-debugging
	  (format t "~&Updating advantage towards ~a.~&Advantages for state are ~a.~&Advantage increase is ~a" 
		    (+ max-adv (- target value)) 
		    (mapset 'list #'(lambda (a) (cons a (handler-case (advantage adv-fn s a) (unknown-action () 'unknown))))
			    (adv-fn::avail-actions adv-fn s))
		    adv-increase))
	 (value-fn:update-value value-fn s (+ value adv-increase) 1))

       (when-debugging
	(format t "~&Advantage function is now~&")
	(pprint-advantage-fn adv-fn)
	(format t "~&Value function is now~&")
	(value-fn:pprint-value-fn value-fn))
       
       ;; Normalize
       (adv-fn:normalize-state adv-fn s)
       
       (when-debugging
	(format t "~&After normalizing, advantage function is now~&")
	(pprint-advantage-fn adv-fn))
       
       (setf (prev-s alg) prev-s)))))





(defmethod knowledge-state ((alg <advantage-updating>) &optional (fresh t))
  (if fresh
      (clone (list (adv-fn alg) (value-fn alg)))
    (list (adv-fn alg) (value-fn alg))))

(defmethod get-advantage-fn ((alg <advantage-updating>) ks)
  (first ks))

(defmethod get-value-fn ((alg <advantage-updating>) ks)
  (second ks))


  
   