(defpackage certainty-equivalence
  (:documentation "Package certainty-equivalence

Exports
-------
<certainty-equivalence>")
  (:use
   cl
   rl-obs
   utils
   set)
  (:export <certainty-equivalence>))

(in-package certainty-equivalence)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; class
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <certainty-equivalence> (<q-learning-algorithm>)
  ((discount :type float :initarg :discount :initform 1.0 :accessor discount)
   (prev-state :accessor prev-state)
   (state-descs :accessor state-descs)
   (change-flag :accessor change-flag :initform t)
   (ks :accessor ks))
  (:documentation "Class <certainty-equivalence> (<q-learning-algorithm>)
Like <gold-standard>, except it uses sparse representations of transition and reward models.

Initargs
:discount.  Defaults to 1.0
:unknown-state-behavior (see <q-learning-algorithm>"))

(defmethod shared-initialize :after ((alg <certainty-equivalence>) names &rest args)
  (declare (ignore names args))
  (setf (state-descs alg) (make-hash-table :test #'equalp)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; methods from <rl-observer>
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod reset ((alg <certainty-equivalence>))
  (reinitialize-instance alg))

(defmethod inform-start-episode ((alg <certainty-equivalence>) s)
  (notice-state alg s nil)
  (setf (prev-state alg) s))

(defmethod inform-env-step ((alg <certainty-equivalence>) a r s2 term &aux (s (prev-state alg)))
  (notice-state-action alg s a s2 term)
  (let* ((sa-desc (mapping:evaluate (second (mapping:evaluate (state-descs alg) s)) a))
	 (entry 
	  (or (assoc s2 (first sa-desc) :test #'equalp)
	      (let ((new-entry (cons s2 0)))
		(push new-entry (first sa-desc))
		(push (cons s2 r) (second sa-desc))
		new-entry))))
    (incf (cdr entry)))
  (setf (prev-state alg) (if term 'no-prev-state s2)
	(change-flag alg) t)
  (values))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; methods from q-learning-algorithm
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *mpi-k* 6)
(defmethod knowledge-state ((alg <certainty-equivalence>) &optional (fresh t))
  (declare (ignore fresh)) ;; always returns a fresh object
  (if (change-flag alg)
      (setf (change-flag alg) nil
	    
	    (ks alg)
	    (let ((m (get-sparse-mdp alg)))
	      (multiple-value-bind (pol val)
		  (dp:policy-iteration m :k *mpi-k* :discount (discount alg))
		(declare (ignore pol))
		(list m (dp:q-from-v m val (discount alg))))))
    
    (ks alg)))

(defmethod get-q-fn ((alg <certainty-equivalence>) ks)
  (second ks))

(defmethod get-mdp ((alg <certainty-equivalence>) ks)
  (first ks))
      
	 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; internal
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun notice-state (alg s term &aux (descs (state-descs alg)))
  (mvbind (val present?) (gethash s descs)
    (if present?
	(assert (eq (first val) term) nil
	  "~a appeared as both a terminal and nonterminal state." s)
      (setf (gethash s descs)
	(list term (make-hash-table :test #'equal))))))

(defun notice-state-action (alg s a s2 term &aux (descs (state-descs alg)))
  (notice-state alg s2 term)
  (mvbind (s-desc present?) (gethash s descs)
    (assert present?)
    (unless (hash-table-has-key (second s-desc) a)
      (setf (gethash a (second s-desc)) (list nil nil)))))


(defun get-sparse-mdp (alg)
  (let ((h (make-hash-table :test #'equalp)))
    (maphash
     #'(lambda (s sd)
	 (setf (gethash s h)
	   (list
	    (first sd)
	    (let ((h2 (make-hash-table :test #'equal)))
	      (maphash 
	       #'(lambda (a ad)
		   (dbind (trans reward) ad
		     (let ((total (sum-over trans #'cdr)))
		       (setf (gethash a h2)
			 (list
			  (mapcar #'(lambda (x) (cons (car x) (/ (cdr x) total))) trans)
			  (clone reward))))))
	       (second sd))
	      h2
	      ))))
     (state-descs alg))
    (make-instance 'mdp:<sparse-mdp> :state-descs h)))
