(in-package advantage-updating)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; initialization
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;; An alternate, perhaps more clean, way to implement would be to have, in addition to the 
;; <decomposed-advantage-updating> object, a set of component objects of some type
;; <component-learner>, one for each reward component, such that the central object
;; forwards the appropriate information to the components
(defclass <decomposed-advantage-updating> (<advantage-updating>)
  ((reward-components :accessor reward-components :initarg :reward-components :writer set-reward-components))
  (:documentation "Class <decomposed-advantage-updating> (<advantage-updating>)

:adv-fns - list of advantage functions
:value-fns - list of value functions
:reward-components - list in which each element is a real-valued function of four arguments, S, A, R, S2.  They are guaranteed to always be called in sequence.

Initargs for <advantage-updating>
:avail-actions-specifier
:lrate
"))


(defmethod initialize-instance :after ((alg <decomposed-advantage-updating>) &rest args 
				       &key adv-fns value-fns avail-actions-specifier)
  (declare (ignore args))
  (setf (init-adv-fn alg) (make-instance '<sum-advantage-function> :adv-fns adv-fns :avail-actions-specifier avail-actions-specifier)
	(init-value-fn alg) (make-instance 'value-fn:<sum-value-fn> :value-fns value-fns))
  (reset alg))

(defmethod reset ((alg <decomposed-advantage-updating>))
  (when (slot-boundp alg 'initial-advantage-function)
    (setf (adv-fn alg) (clone (init-adv-fn alg))
	  (value-fn alg) (clone (init-value-fn alg)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; responses to messages
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod inform-env-step ((alg <decomposed-advantage-updating>) a r s2 term)
  (handler-bind ((value-fn:unknown-state #'(lambda (c) (declare (ignore c)) (use-value 0.0))))
    (with-accessors ((s prev-s) (adv-fn adv-fn) (rcomps reward-components)
		     (value-fn value-fn)) alg
      (let ((eta (lrate:get-rate (lrate alg) (cons s a)))
	    (adv-fns (adv-components adv-fn))
	    (value-fns (value-fn:value-components value-fn))
	    (rewards (mapcar #'(lambda (rc) (funcall rc s a r s2)) rcomps)))
	(let ((vals (mapcar #'(lambda (f) (value-fn:value f s)) value-fns))
	      (next-vals (mapcar #'(lambda (f) (value-fn:value f s2)) value-fns))
	      (best-advs (best-advantages adv-fn s)))
	  (condlet
	   ((term (targets rewards) (new-prev-s nil))
	    (t (targets (mapcar #'+ rewards next-vals)) (new-prev-s s2)))
	   
	   (when-debugging
	    (format t "~%~%~&Observed~&State: ~a~&Action: ~a~&Next state: ~a~&Terminal: ~a" s a s2 term)
	    (format t "~&Reward is ~a, decomposed as ~a" r rewards)
	    (format t "~&Value of current state is ~a, decomposed as ~a" (reduce #'+ vals) vals)
	    (format t "~&Value of new state is ~a, decomposed as ~a" (reduce #'+ next-vals) next-vals)
	    (format t "~&Best advantages at previous state ~a are ~a" s best-advs))

	  
	   ;; advantage updates
	   (do-sets ((f adv-fns) (target targets) (value vals) (best best-advs)) ()
	     (when-debugging
	      (format t "~&Updating advantage of ~a, ~a towards ~a by ~a" s a (+ best (- target value)) eta))
	     (update f s a (+ best (- target value)) eta))
	  
	   (let ((new-best-advs (best-advantages adv-fn s)))
	    
	     ;; value updates
	     (let ((diffs (mapcar #'- new-best-advs best-advs)))
	       (debug-print 0 "~&Best advantages are now ~a~&Componentwise advantage changes are ~a" new-best-advs diffs)
	       (do-sets ((f value-fns) (v vals) (d diffs)) ()
		 (value-fn:update-value f s (+ v d) 1)
		 (when-debugging
		  (format t "~&Increasing value of ~a by ~a" s d))))
	    
	     (when-debugging
	      (format t "~&Advantage function is now~&")
	      (pprint-advantage-fn adv-fn)
	      (format t "~&Value function is now~&")
	      (value-fn:pprint-value-fn value-fn)
	      )
	    
	     ;; normalization
	     (do-sets ((f adv-fns) (best new-best-advs)) (nil i)
	       (debug-print 0 "~&Normalizing adv fn ~a at ~a by ~a" i s best)
	       (normalize-state f s best)))
	  
	   (setf (prev-s alg) new-prev-s)
	  
	   (when-debugging
	    (format t "~&After normalizing, advantage function is~&")
	    (pprint-advantage-fn adv-fn)
	    (debug-prompt))))))))