(defpackage q-learning
  (:documentation "Defines the <q-learning> learning algorithm.

Exports
-------
<q-learning>
set-update-frequency
<decomposed-q-learning>
")
  (:import-from 
   rl
   )
  (:export 
   set-update-frequency
   <decomposed-q-learning>
   q-fn
   <q-learning>)
  (:use 
   rl-obs
   cl
   utils))

(in-package q-learning)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Class def
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defclass <q-learning> ( <q-learning-algorithm>)
  ((discount :type float
	     :initarg :discount
	     :initform 1.0
	     :reader discount)
   (learning-rate :type lrate:[learning-rate]
		  :initarg :lrate
		  :initform .01
		  :reader lrate)
   (q-function :type (or null q-fn:<q-function>)
	       :reader q-fn
	       :initarg :q-fn
	       :initform nil
	       :writer set-q-fn)
   (batched-updates :type array
		    :reader batched-updates
		    :writer set-batched-updates)
   (update-frequency :type fixnum
		     :reader update-frequency
		     :writer set-update-frequency
		     :initarg :update-frequency
		     :initform 1)
   (init-q-function :type q-fn:<q-function>
		    :reader init-q-fn
		    :writer set-init-q-fn)
   (prev-s :accessor prev-s)
   (prev-a :accessor prev-a))
  (:documentation "The <q-learning> class.  Inherits from <learning-algorithm>.  Create using make-instance. Initargs

:lrate - <learning-rate>
:q-fn - <q-function> (defaults to a tabular q-fn - see the env initarg below)
:env - if q-fn initarg not provided, must provide this instead.
:discount (optional, default 1.0)
:update-frequency - a fixnum that defaults to 1.  Q-function updates are accumulated and when this many have collected, they're all done.
as well as the initargs from <learning-algorithm>

When asked for its knowledge state, a <q-learning> object returns the currently estimated Q-function.  Its convert-to-policy method turns a Q-function into the corresponding greedy policy.
"))


(defmethod initialize-instance :after ((alg <q-learning>) &rest args &key env)
  (declare (ignore args))
  (unless (q-fn alg)
    (set-q-fn (make-instance 'q-fn:<env-q-function> :env env) alg))
  (set-init-q-fn (clone (q-fn alg)) alg))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Operations from <rl-observer>
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod inform-start-execution ((alg <q-learning>))
)

(defmethod inform-start-episode ((alg <q-learning>) s)
  (setf (prev-s alg) s))
  

(defmethod inform-env-step ((alg <q-learning>) a r s2 term)
  (let* ((prev-s (prev-s alg))
	 (eta (lrate:get-rate (lrate alg) (cons prev-s a))))
    (debug-msg alg "~&Env step happened in Q-learning.  Q-function is:~&~W" (q-fn alg))
    (if term
	(progn
	  (do-update alg prev-s a r eta)
	  (setf (prev-s alg) nil))
      (progn
	(do-update alg prev-s a
		     (+ r 
			(* (discount alg) 
			   (handler-case
			       (q-fn:value (q-fn alg) s2)
			     (q-fn:unknown-state-action () 0))))
		     eta)
	(setf (prev-s alg) s2)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Operations from <q-learning-algorithm>
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod reset ((alg <q-learning>))
  (set-q-fn (clone (init-q-fn alg)) alg)
  (set-batched-updates (make-array (update-frequency alg) :adjustable t :fill-pointer 0)
		       alg)
  (lrate:reset-counts (lrate alg)))

(defmethod knowledge-state ((alg <q-learning>) &optional (fresh t))
  "The state of knowledge is just the Q-function."
  (let ((q (q-fn alg)))
    (if fresh (clone q) q)))

(defmethod get-q-fn ((alg <q-learning>) ks)
  ks)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Internal
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun do-update (alg s a target eta)
  (let ((v (batched-updates alg))
	(q (q-fn alg)))
    (if (= (length v) (1- (update-frequency alg)))
	(progn
	  (q-fn:update q s a target eta)
	  (debug-msg alg "~&Batched q-function updates being done")
	  (map nil #'(lambda (x) (debug-msg alg ".") (apply #'q-fn:update q x)) v)
	  (setf (fill-pointer v) 0))
      (progn
	(vector-push (list s a target eta) v)
	(debug-msg alg "~&Q-function update batched.")))))
      
		  



(in-package cl-user)

   