(defpackage ltk-observer
  (:documentation "Package ltk-observer (ltk-obs)

Defines an observer that uses the LTK package to graphically display the state of an environment, and can also be used to choose actions.

Types
-----
<ltk-observer>

Ops
---
ltk-interface

Ops for envs to implement
-------------------------
display-init-state
update-state

Special-vars
------------
*canvas*
*print-avail-actions*


")
  (:export
   
   <ltk-observer>
   
   ltk-interface
   
   display-init-state
   update-state
   
   *canvas*
   *print-avail-actions*
   )
  
  (:nicknames ltk-obs)
  (:use cl
	rl-obs
	policy
	ltk
	utils))

(in-package ltk-obs)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Class
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defclass <ltk-observer> (<rl-observer> <policy>)
  ((canvas :accessor canvas)
   (env :initarg :env :accessor env)
   (action :accessor action)
   (exit :accessor exit :initform nil)
   (prev-s :accessor prev-s))
  (:documentation "<ltk-observer>

Observer that uses the LTK package to graphically display the state of an environment.

Initargs
:env - the environment

The environment must have methods defined for display-init-state and update-state"))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Top level interface
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ltk-interface (env)
  (let ((obs (make-instance '<ltk-observer> :env env)))
    (rl-user:reinforcement-learning 
     (make-instance 'rl-user:<rl-control>
       :env env :policy obs :obs (list obs)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; RL observer methods
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *canvas* nil)

(defmethod inform-start-execution ((obs <ltk-observer>))
  (exit-wish)
  (start-wish)
  (let* ((sc (make-instance 'scrolled-canvas))
	 (c (setf (canvas obs) (canvas sc)))
	 (e (make-instance 'entry))
	 (b (make-instance 'button :text "exit"
			   :command #'(lambda () (setf (exit obs) t)))))
    (pack sc :side :top :fill :both :expand 1) 
    (pack e :side :top :anchor :w)
    (pack b :side :top :anchor :center)
    (setf *canvas* c)
    (bind e "<Return>" 
	  #'(lambda (event) 
	      (declare (ignore event))
	      (setf (action obs) (text e)
		    (text e) "")))
    (ltk:scrollregion c 0 0 800 800)))

(defmethod inform-finish-execution ((obs <ltk-observer>))
  (exit-wish))

(defmethod inform-start-episode ((obs <ltk-observer>) s)
  (clear (canvas obs))
  (display-init-state (env obs) (canvas obs) s)
  (print-avail-actions (env obs) s)
  (setf (prev-s obs) s))

(defmethod inform-env-step ((obs <ltk-observer>) a r s2 term)
  (declare (ignore a r term))
  (update-state (env obs) (canvas obs) (prev-s obs) s2)
  (print-avail-actions (env obs) s2)
  (setf (prev-s obs) s2))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Policy method
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod make-choice ((obs <ltk-observer>) s &optional choices)
  (let ((x (gensym)))
    (with-slots (action exit) obs
      (setf action x)
      (loop
	(process-events)
	(unless (and (eq action x) (not exit))
	  (if exit
	      (error 'choose-to-abort)
	    (handler-case (return-from make-choice (env-user:translate-action (env obs) (read-from-string action) s choices))
	      (env-user:untranslatable-action ()
		(setf action x)
		(print "Unknown action"))))
	  (sleep .1)))))) ;; Ugly, but LTK does not support true multithreading for all platforms
    

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Methods for environments to implement
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric display-init-state (e c s)
  (:documentation "display-init-state ENV CANVAS STATE
Display background for the environment (e.g. a grid), as well as the given state, on CANVAS, using the ltk operations.  A before method clears the canvas")
  (:method :before (e c s)
	   (declare (ignore e s))
	   (clear c)))

(defgeneric update-state (e c s1 s2)
  (:documentation "update-state ENV CANVAS PREV-STATE CURR-STATE

Update the state displayed on CANVAS to CURR-STATE using the ltk operations."))
			    

   
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Helpers
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *print-avail-actions* t)

(defun print-avail-actions (e s)
  (when *print-avail-actions*
    (format t "~%~%Available actions are:~2I~&~w" (env:avail-actions e s))))