;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; rl/rl-observer.lisp
;; Defines the <rl-observer> abstract class and some generic functions.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package rl)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; class def
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <rl-observer> ()
  ((name :accessor name :initarg :name))
  (:documentation "Class <rl-observer>.    An rl-observer is any object that observes execution during reinforcement learning.  This can include learning algorithms (see subdirectory learn/), and also objects that maintain logs or statistics of execution (see subdirectory obs/). 
  
Subclasses may implement any subset of the following methods.  If a subclass doesn't implement a method, that just means the corresponding event is ignored by that type of observer.
- inform-start-episode
- inform-env-step
- inform-start-execution
- inform-finish-execution
"))

(defmethod initialize-instance :after ((o <rl-observer>) &rest args)
  (declare (ignore args))
  (set-if-unbound 'name o (cons (class-of o) (gensym))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; definition macro
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro defmessage (name lambda-list &optional (doc-string ""))
  "defmessage NAME LAMBDA-LIST DOC-STRING.  Used to define messages to a rl-observer (as generic functions).  As an example, (defmessage inform-reset (s) doc-string) expands to code that creates a generic function inform-reset with two arguments - observer and s, documentation field doc-string, and a single method defined for the case when observer is of type <observer> that does nothing.  See examples in this file."
  (with-gensyms (observer)
    `(defgeneric ,name ,(cons observer lambda-list)
		 (:documentation ,doc-string)
       (:method ,(cons `(,observer <rl-observer>) lambda-list) (declare (ignore ,@lambda-list)) ()))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; kinds of messages that can be sent to an observer
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmessage inform-start-episode (s)
  "inform-start-episode OBSERVER NEW-STATE.  Inform observer that environment has been reset to state NEW-STATE.")

(defmessage inform-env-step (act rew to term)
  "inform-env-step OBSERVER ACTION REWARD NEXT-STATE TERM.  Inform observer of that ACTION was done in the environment, resulting in new state NEXT-STATE, REWARD units of reward.  TERM is true iff the environment has now terminated. (Note that the observer will always know the previous state - because of a previous inform-start-episode or inform-env-step message.")

(defmessage inform-finish-execution ()
  "inform-finish-execution OBSERVER.  Inform observer that the current run is about to finish (i.e. we may not have reached the end of an episode, but we're going to stop anyway).")

(defmessage inform-start-execution ()
  "inform-start-execution OBSERVER.  Inform observer that we've just started a new run.")

  
(defgeneric reset (alg)
  (:documentation "reset ALG.  Reset the state of this algorithm, so it forgets everything that has been learnt.  There is a top-level :before method that resets the history.  There is an around method that makes ALG be returned.")
  (:method :before (alg)
	   
	   (awhen (hist-out-dir alg)
	     (dotimes (i (num-hist-files alg))
	       (delete-file (format nil "~a/~a" it i))
	       ))
	   (setf (alg-hist alg) (make-array 0 :adjustable t :fill-pointer 0)
		 (num-hist-files alg) 0
		 (env-steps alg) 0
		 (episode-steps alg) 0))

  ;; This is hacky
  (:method :around (alg)
	   (progn
	     (call-next-method)
	     alg)))

